#!/bin/sh

. ./fsx.conf

echo " ////////////////////////////////////////////////////////////////////"
echo "      FreyaSX installer                            "
echo "                               FreyaSX ver. $SXVER "
echo "                               DeleGate ver. $DGVER"
echo "                               August 18, 2014, Yutaka Sato"
echo " ////////////////////////////////////////////////////////////////////"
echo " ADMIN=$ADMIN (you should set this in fsx.conf)"

export DGROOT=`pwd`/DGROOT
export FSXHOME=`pwd`/FSXHOME
echo " DGROOT=$DGROOT"
echo " FSXHOME=$FSXHOME"
echo ""

if [ ! -d "$DGROOT" ]; then
  mkdir "$DGROOT"
fi
if [ ! -d "$FSXHOME" ]; then
  mkdir "$FSXHOME"
fi

LOG=./INSTALL.log
echo 'Detailed log is put into "'$LOG'"'
echo ""

if [ -d c:/ ]; then
  EXE=".exe"
else
  EXE=""
fi

######################################
echo "*** extracting FreyaSX-$SXVER ..."
gzip -d < freyasx-$SXVER.tar.gz | tar xf -
(
cd freyasx-$SXVER
echo "*** making FreyaSX-$SXVER ... (it may take a minute)"
if [ -d c:/ ]; then
  (cd freya/src; makevc.sh) >> ../$LOG 2>&1
  makewin.sh >> ../$LOG 2>&1
else
  (cd freya/src; make) >> ../$LOG 2>&1
  make >> ../$LOG 2>&1
  testsx >> ../$LOG 2>&1
fi
)

######################################
echo "*** extracting DeleGate/$DGVER ..."
gzip -d < delegate$DGVER.tar.gz | tar xf -
(
cd delegate$DGVER
echo "*** making DeleGate/$DGVER ... (it may take several minutes)"
if [ -d c:/ ]; then
  cp -p ../../../dg$DGVER/src/delegated.exe src/
  export ADMIN=$ADMIN
  if [ ! -f make-vs8_win8.bat ]; then
    echo  "!!! So sorry but you need the full edition of DeleGate to build from the source." > /dev/tty
    echo  "!!! Or you can get the executable of DeleGate (as any2fdif and fhttpd) for Windows." > /dev/tty
    exit 1
  else
    make-vs8_win8.bat >> ../$LOG 2>&1
  fi
else
  make ADMIN=$ADMIN >> ../$LOG 2>&1
fi
echo "*** getting character conversion table ..."
src/delegated$EXE LIBDIR=$FSXHOME/etc CHARSET=UTF-8 -Fccx < /dev/null >> ../$LOG 2>&1
)
if [ $? != 0 ]; then
   echo "--- Sorry again."
   exit 1
fi

######################################
echo ""
echo "Installing to: $FSXHOME"
cp -rp freyasx-$SXVER/bin     $FSXHOME
cp -rp freyasx-$SXVER/etc     $FSXHOME
cp -rp freyasx-$SXVER/cgi-bin $FSXHOME
cp -rp freyasx-$SXVER/bank    $FSXHOME
cp -rp freyasx-$SXVER/SXwww   $FSXHOME
cp -rp freyasx-$SXVER/test    $FSXHOME
cp -p delegate$DGVER/src/delegated$EXE $FSXHOME/etc
if [ -a test/fhttpd.conf ]; then cp -p test/fhttpd.conf $FSXHOME/etc; fi
ln -s ../etc/delegated$EXE $FSXHOME/bin/any2fdif$EXE
ln -s ../etc/delegated$EXE $FSXHOME/bin/fhttpd$EXE
ln -s ../etc/delegated$EXE $FSXHOME/bin/fnntpd$EXE
echo "  "$FSXHOME/bin
echo "  "$FSXHOME/etc
echo "  "$FSXHOME/cgi-bin
echo "  "$FSXHOME/bank
echo ""
echo "0) You can remove following unnecessary files:"
echo '---------------------------------------------------------------------'
echo "  "rm -r `pwd`/delegate$DGVER
echo "  "rm -r `pwd`/freyasx-$SXVER
echo "  "rm `pwd`/$LOG

echo ''
echo '1) Complete the installation and setup your environmnet as follows:'
echo '---------------------------------------------------------------------'
if [ "$SHELL" = "/bin/csh" -o "$SHELL" = "/bin/tcsh" ]; then
  PROMPT='%'
  echo '  setenv FSXHOME '"$FSXHOME"
  echo '  setenv DGROOT '"$DGROOT"
  echo '  set path=($FSXHOME/bin $path)'
  echo '  rehash'
else
  PROMPT='$'
  echo '  export FSXHOME='"$FSXHOME"
  echo '  export DGROOT='"$DGROOT"
  echo '  export PATH="$FSXHOME/bin:$PATH"'
fi
echo ''
echo '2) Then, you can test it as follows:'
echo '---------------------------------------------------------------------'
echo '  findex'$EXE' sx'
echo '  fsearch'$EXE' sx "freya*"'
echo '  any2fdif'$EXE' -r $HOME/Mail | findex mail'
echo ''
echo '3) Also you can start HTTP server for this FreyaSX and test it:'
echo '---------------------------------------------------------------------'
echo '  fhttpd'$EXE' -P8880'
echo '  ... then open URL "http://'`hostname`':8880/test/" in your WWW browser'
echo '  ... you can stop the HTTP server with a command "fhttpd -P8880 -Fkill"'
echo ''
echo 'See http://www.delegate.org/freyasx/ for more details'
echo " ////////////////////////////////////////////////////////////////////"
