// Copyright (c) 2004 Yutaka Sato. All rights reserved.
//
// text pointer in IndexFile = {docid:31bits, offset:26bits, type:6bits}
// "type" should be after offset, to let pos_t values increase monotonously
// and it sould be shifted out when searching adjacent word by pos_t value
//
typedef struct {
	int c_docid; // document ID [0..]
	int c_type;  // element type {URL, TITLE, AUTHOR, ...}
	int c_words; // word number
/*
	int c_words_total;
*/
	off_t c_words_total;
	char c_delch; // delimiter character
} ElemCtx;

INLINE int pos2did(pos_t pos){
	return 0x7FFFFFFF & (pos >> 32);
}
INLINE int pos2off(pos_t pos){
	return 0x03FFFFFF & (pos >> 6);
}
INLINE int pos2type(pos_t pos){
	return 0x0000003F & (pos >> 0);
}

INLINE int _frompos(pos_t pos, urlid_t *urlid, size_t *offset, int *type){
	*urlid =  0x7FFFFFFF & (pos >> 32);
	*offset = 0x03FFFFFF & (pos >> 6);
	*type =   0x0000003F & (pos >> 0);

#ifdef DEBUG
	fprintf(stderr,"#<< pos=[%08X %08X] did=%X typ=%X off=%04X\n",
	  (int)(pos>>32),(int)pos,urlid,type,(int)offset);
#endif
	return 0;
}
/*
INLINE pos_t topos(int docid, int type, int offset){
*/
INLINE pos_t topos(unsigned int docid, unsigned int type, unsigned int offset){
        pos_t pos;
	pos = ((pos_t)docid<<32)|((0x3FFFFFF & offset)<<6)|((0x3F & type)<<0);

#ifdef DEBUG
	fprintf(stderr,"#>> pos=[%08X %08X] did=%X typ=%X off=%04X\n",
	  (int)(pos>>32),(int)pos,docid,type,offset);
#endif
        return pos;
}
INLINE pos_t setposdid(pos_t pos, int docid){
	return (pos & ~(((pos_t)0x7FFFFFFF) << 32))
             | ((docid & (pos_t)0x7FFFFFFF) << 32);
}

//
// shuffle before diffpack() / diffunpack()
// {did:31bits, off:26bits, typ:6bits}
// {did:16,typ:2,off:13 .. did:15,typ:4,off:13}
//
INLINE pos_t EnPACK(pos_t pos){
	pos_t epos;
	urlid_t did; size_t off; int typ;
	_frompos(pos,&did,&off,&typ);
	epos =
	(pos_t)(0x7FFF8000 & did)<<( 2+13+32-15) | (0x7FFF & did) <<(13+4)|
	(pos_t)(0x00000030 & typ)<<(   13+32- 4) | (0x000F & typ) <<(  13)|
	(pos_t)(0x03FFE000 & off)<<(    0+32-13) | (0x1FFF & off) <<(   0);

#ifdef DEBUG
	fprintf(stderr,"##ENC## [%08X %08X] -> [%08X %08X]\n",
		(int)(pos>>32),(int)pos,int(epos>>32),(int)epos);
#endif
	return epos;
}
INLINE pos_t DePACK(pos_t pos){
	pos_t dpos;
	unsigned int did,typ,off;
	did = (0x7FFF8000 & (pos >>( 2+13+32-15))) | (0x7FFF & (pos >>(13+4)));
	typ = (0x00000030 & (pos >>(   13+32- 4))) | (0x000F & (pos >>(  13)));
	off = (0x03FFE000 & (pos >>(    0+32-13))) | (0x1FFF & (pos >>(   0)));
	dpos = topos(did,typ,off);

#ifdef DEBUG
	fprintf(stderr,"##DEC## [%08X %08X] <- [%08X %08X]\n",
		(int)(dpos>>32),(int)dpos,(int)(pos>>32),(int)pos);
#endif
	return dpos;
}

INLINE pos_t toPos(ElemCtx *ectx, pos_t offset){
	ectx->c_words_total++;
	return topos(ectx->c_docid, ectx->c_type, ectx->c_words);
}

INLINE void
	poscpy(pos_t *dst, const pos_t *src, int len){
	if(16 < len){
		memcpy(dst,src,len*sizeof(pos_t));
	}else{
		int i;
		for(i = 0; i < len; i++)
			dst[i] = src[i];
	}
}
