// Copyright (c) 2004 Yutaka Sato. All rights reserved.

#include "ConText.h"

#define O_ISLIT  0x80000000 // direct (1-4 bytes of characters)
#define O_BOE    0x40000000 // start of a element
#define O_PREFIX 0x3E000000 // prefixing delimiter symbol (1-31)
#define O_LEXOFF 0x01FFFFFF // offset in lexicon table
#define toPrefix(del)  ((0x1F & del)<<25)
#define getPrefix(off) (0x1F & (off>>25))

#define S_BOD	0x01  // begin of document

typedef struct {
    int kid;
} TmpCtx;

static int prevDocid = -1;
static int prevType = -1;
static char *delchs = "!\"#$%&'()*+,-./:;<=>?@[\\]^_{|}~";
/*
"delch" should be a string like "://"
thus delchs should be char *delchs[]
*/
static int Nlit;

/*
static FILE* fwsmap;
static FILE* flomap;
*/

void ConText::putwsmap(int mi,int li,int off){
	wsmap[mi][li] = off;
}
void ConText::putlomap(int mi,int li,int off){
	lomap[mi][li] = off;
}
int ConText::getwsmap(int mi,int li){
	return wsmap[mi][li];
}
int ConText::getlomap(int mi,int li){
	return lomap[mi][li];
}

ConText::ConText(const char *filename, const char *mode){
    insF = 0; insN = insX = 0;
    open(filename,mode);
    LexBase = 0;
    LexSer = 0;
    lser = 0;
    Idxid = 0;
    lexf = NULL;
/*
    for(int i = 0; i < 256; i++){
        lmap[i] = 0;
        newidx[i] = 0;
    }
*/
	for(int i = 0; i < LMAPSIZ; i++){
		wsmap[i] = 0;
		newidx[i] = 0;
	}
/*
LapReport(0,"--- ConText1 %s %X %X",filename,(int)fwsmap,(int)flomap);
	if( fwsmap == NULL ){
		char path[128];
		char *dp;
		strcpy(path,filename);
		dp = strrchr(path,'.');
		if( dp ) *dp = 0;
		strcat(dp,"mmp");
		fwsmap = fopen(path,"w+");
		strcat(path,"-2");
		flomap = fopen(path,"w+");
LapReport(0,"--- ConText1 %s %X %X",path,(int)fwsmap,(int)flomap);
	}
*/
}
ConText::ConText(){
    insF = 0; insN = insX = 0;
    ctxf = 0;
    LexBase = 0;
    LexSer = 0;
    lser = 0;
    Idxid = 0;
    lexf = NULL;
/*
    for(int i = 0; i < 256; i++){
        lmap[i] = 0;
        newidx[i] = 0;
    }
*/
	for(int i = 0; i < LMAPSIZ; i++){
		wsmap[i] = 0;
		newidx[i] = 0;
	}
/*
	if( fwsmap == NULL ){
		fwsmap = fopen("/tmp/freyasx-lmap","w+");
		flomap = fopen("/tmp/freyasx-lomap","w+");
LapReport(0,"--- ConText2 %d",(int)sizeof(off_t));
	}
*/
}
ConText::~ConText(){
    if(ctxf != NULL){
       delete ctxf;
       ctxf = 0;
    }
}

int
ConText::open(const char *filename, const char *mode){
    ctxf = new Fstream(filename,mode);
    if(ctxf == NULL || !ctxf->good()){
        cerr << "Cannot open: " << filename << "(" << mode << ")" << endl;
        return -1;
    }
    return 0;
}

int
ConText::getctx(char *buf, int size){
    int rcc;
    if(ctxf != NULL){
        rcc = ctxf->read(buf,size);
        if(0 < rcc)
            return rcc;
    }
    if(insX < insN){
        if(ctxf != NULL)
        {
            delete ctxf;
            ctxf = 0;
        }
        open(insF[insX++],"r");
        if(!ctxf->good()){
            cerr << "Cannot open: " << insF[insX-1] << "(r)" << endl;
            return 0;
	}
        rcc = ctxf->read(buf,size);
        return rcc;
    }
    return 0;
}

void
ConText::CtxPutpos(Word &term, ElemCtx &ectx, int offset){
    TmpCtx tc;
    int isBOE = (ectx.c_type != prevType);

    char delch;
    int pfx = 0;
    tc.kid = term.kid;
    prevDocid = ectx.c_docid;
    prevType = ectx.c_type;

    char *dp;
    if(ectx.c_delch && (dp = strchr(delchs,ectx.c_delch))){
        delch = ectx.c_delch;
        pfx = 1 + (dp-delchs);
    }else{
        delch = 0;
        pfx = 0;
    }
/*
printf("PUTPOS [%c][%s]\n",delch?delch:' ',term.toString());
*/
    // ISLIT -- Japanese word of 4bytes
    if(ectx.c_type == TYPE_META){
        if(ectx.c_words == 1){
            tc.kid = O_ISLIT | (S_BOD << 8);
        }
    }else
    if(!isBOE){
        unsigned char *key = (unsigned char*)term.toString();
        int len = term.length();
        if(len == 4 && ectx.c_delch == 0 ){
            if(0x80 & key[0]){
                tc.kid = 0;
                for(unsigned char *s = key; *s; s++)
                    tc.kid = (tc.kid << 8) | *s;
                tc.kid |= O_ISLIT;
                Nlit++;
            } 
        }else{
            int cc = 0;
            if(ectx.c_delch) cc++;
            cc += len;
            if(cc <= 3){
                tc.kid = 0;
                if(ectx.c_delch) tc.kid = (0xFF & ectx.c_delch);
                for(unsigned char *s = key; *s; s++)
                    tc.kid = (tc.kid << 8) | *s;
                tc.kid |= O_ISLIT;
                Nlit++;
            }
        }
    }
    if((tc.kid & O_ISLIT) == 0){
        if(isBOE) tc.kid |= O_BOE;
        if(pfx) tc.kid |= toPrefix(pfx);
    }


/*
fprintf(stderr,"#### %3d %08X [%c]%s\n",LexSer,tc.kid,delch?delch:' ',term.toString());
*/

    ctxf->write((char*)&tc,sizeof(tc));
/*
    Fwrite(ctxf,(char*)&tc,sizeof(tc));
*/
    LexSer++;
    ectx.c_delch = 0;
}

void
ConText::adjustLex(ConText *octx, IMask *mask){
    LapReport(0,"adjusting .ctx ...",0,0,0,0);
    int docid = -1;
    int ndeldoc = 0;
    int ndelpos = 0;
    int ignore = 0;
    int mi = 0;
/*
    int mt = smap[0];
*/
    off_t mt = smap[0];
    int isnew = newidx[0];
    Nlit = 0;
/*
    int i;
*/
    off_t i;

    for(i = 0; ; i++){
        if((i%100) == 0)
/*
        LapReport(1,"adjusting .ctx ... %d occurrences",i,0,0,0);
*/
/*
        LapReport(1,"adjusting .ctx ... %lld occurrences",i);
*/
        LapReport(1,"adjusting .ctx ... %lld occurrences",(CODEINT)i);

        TmpCtx tc;
        int rcc = getctx((char*)&tc,sizeof(tc));
        if(rcc < (int)sizeof(tc)){
            break;
        }
        if(!isnew){
            tc.kid = ntohl(tc.kid);
        }

        if(mt <= i){
            mt = smap[++mi];
            isnew = newidx[mi];
        }
        int li,xid,xmi,xli;
        int off = 0;

        // ISLIT
        if(tc.kid & O_ISLIT){
            Nlit++;
            off = tc.kid;
            if((tc.kid & 0xFF) == 0){
                docid++;
                if(mask){
                    ignore = mask->get(docid) == -1;
                    if(ignore)
                        ndeldoc++;
               }
            }
            if(ignore){
                 ndelpos++;
                 continue;
            }
        }else{
            if(ignore){
                 ndelpos++;
                 continue;
            }
            li = tc.kid & 0x1FFFFFF;
            if(!isnew){
                if(lomap[mi] == 0){
                    fprintf(stderr,"adjustLex: no lomap[%d]\n",mi);
                    break;
                }
		/*
                off = lomap[mi][li];
		*/
                off = getlomap(mi,li);
            }else{
		/*
                if(lmap[mi] == 0){
		*/
                if(wsmap[mi] == 0){
                    fprintf(stderr,"adjustLex: no lmap[%d]\n",mi);
                    break;
                }
		/*
                xid = lmap[mi][li];
		*/
                xid = getwsmap(mi,li);
                xmi = 0xFF & (xid>>24);
                xli = xid & 0xFFFFFF;
		/*
                off = lomap[xmi][xli];
		*/
                off = getlomap(xmi,xli);
            }
            if(tc.kid & O_BOE) off |= O_BOE;
            if(int pfx = getPrefix(tc.kid)) off |= toPrefix(pfx);
        }
        off = htonl(off);
/*
        Fwrite(octx->getfile(),(char*)&off,sizeof(off));
*/
        octx->ctxf->write((char*)&off,sizeof(off));
    }
/*
    LapReport(0,"adjusting .ctx ... %d - %d / %d - %d docs DONE",
        i,ndelpos,docid+1,ndeldoc);
    LapReport(0,"Ctx literal=%d / position=%d (%d%%)",Nlit,LexSer,
        (int)((Nlit*100.0)/LexSer),0);
*/
    LapReport(0,"adjusting .ctx ... %lld (%d) / %d (%d) docs DONE",
        (CODEINT)i,-(ndelpos),docid+1,-(ndeldoc));
/*
        i,-(ndelpos),docid+1,-(ndeldoc));
    LapReport(0,"Ctx literal=%d / position=%lld (%.1f%%)",Nlit,LexSer,
*/
    LapReport(0,"Ctx literal=%d / position=%lld (%.1f%%)",Nlit,(CODEINT)LexSer,
        (Nlit*100.0)/LexSer);
}

#define POSSIZ 4
/*
 * a word serial number (in Lex and Idx), poiting into ConText,
 * begins with #1 where the first word #1 is a META set in
 * Indexer::addMETA(" begin")
 */
int ConText::getword(char *word,int size,off_t docbase,int wid){
        int lexpos = 0;
	int rcc;

	*word = 0;
	if( wid < 1 ){
		return -1;
	}
	if( ctxf == NULL ){
		return -1;
	}
	ctxf->seekg((docbase+wid-1)*POSSIZ);
	rcc = ctxf->read((char*)&lexpos,POSSIZ);

	if(rcc != POSSIZ){
		return -1;
	}
	lexpos = ntohl(lexpos);

        LexiconEntry entry;
	int isLIT = lexpos & O_ISLIT;
	int lexoff = isLIT?0: lexpos & O_LEXOFF;

	if(isLIT && (lexpos & 0xFF) == 0 ){ // META
		return 0;
        }
        if(isLIT){
		char *kp = word;
		for(int i = 3; 0 <= i; i--){
		    char ch;
		    if((ch = (lexpos >> i*8)))
		    if((ch & 0x7F) != 0 ) // ignore empty byte like 0x80XXYYZZ
			*kp++ = ch;
		}
		*kp = 0;
		if(strchr(delchs,word[0])){
			memmove(word,word+1,strlen(word));
		}
		return strlen(word);
	}else{
		if( lexf ){
			lexf->getlexstr(lexoff,entry);
			Strncpy(word,entry.getkey(),size);
			return strlen(word);
		}
	}
	return -1;
}
static void getlit(int lexpos,char *word,int size,char *delch){
	char *kp = word;
	for(int i = 3; 0 <= i; i--){
	    char ch;
	    if((ch = (lexpos >> i*8)))
	    if((ch & 0x7F) != 0 ) // ignore empty byte like 0x80XXYYZZ
		*kp++ = ch;
	}
	*kp = 0;
	if(strchr(delchs,word[0])){
		delch[0] = word[0];
		delch[1] = 0;
		memmove(word,word+1,strlen(word));
	}else{
		delch[0] = 0;
	}
}
void ConText::dump(){
	int ci;
	int lexpos;
	int rcc;
	char word[1024];
	char delch[8];
	int nlex = 0;
	int ndel = 0;
	int nlit = 0;
	int docid = 0;

	for( ci = 0; ; ci++ ){
		rcc = ctxf->read((char*)&lexpos,POSSIZ);
		if( rcc <= 0 )
			break;
		int isLIT = lexpos & O_ISLIT;
		int lexid = isLIT?0: lexpos & O_LEXOFF;
		int isBOD = isLIT? ((lexpos&~O_ISLIT)==(S_BOD<<8)) : 0;

		printf("%6d %08X ",ci,lexpos);
		if( isBOD ){
			printf("------ -------- begin document #%d\n",docid);
			docid++;
		}else
		if( isLIT ){
			getlit(lexpos,word,sizeof(word),delch);
			if( delch[0] )
				printf("LIT[%s] %s\n",delch,word);
			else	printf("LIT    %s\n",word);
			nlit++;
			if( delch )
				ndel++;
		}else{
			nlex++;
			if( lexf ){
				LexiconEntry entry;
				lexf->getlexstr(lexid,entry);
				Strncpy(word,entry.getkey(),sizeof(word));
			}else{
				word[0] = 0;
			}
			printf("%-6d %s\n",lexid,word);
		}
	}
	printf("%d docs, %d Literals[%d], %d Lexid\n",
		docid,nlit,ndel,nlex);
}

char *Strncpy(char *s1, const char *s2, int n);
const char *html_entitize(char *buf, const char *src);

void
ConText::getContext(LexiconFile *Lex, pos_t *ctx1sts,int nelem,
                    char *out,int osiz, int WINSIZE, const char *delimiter){
  char *buf = out;
  *buf = 0;
  int catenate = 0;
/*
  int last = 0;
*/
  off_t last = 0;
  int delput = 1;

  for(int ie = 0; ie < nelem; ie++){
    buf += strlen(buf);
    if(out+osiz < buf+64)
        break;
/*
    int ctx1st = ctx1sts[ie];
    int next;
*/
    off_t ctx1st = ctx1sts[ie];
    off_t next;
    if((ie+1) < nelem){
        next = ctx1sts[ie+1];
    }else{
        next = -1;
    }

    *buf = 0;
    char *bp = buf;
    const char *key;
    char keybuf[8];
    int putBend = 0;
    int num;
    int wordcat = catenate;

/*
sprintf(bp,"[%d]",(int)ctx1st);
bp += strlen(bp);
*/

/*
    clearerr(ctxf);
*/
    ctxf->clear();
    int postN = 1;
    if(catenate){
        num = WINSIZE * postN;
    }else
    if(ctx1st < WINSIZE){
        num = (ctx1st + 1 + WINSIZE * postN);
/*
        fseek(ctxf,0,0);
*/
        ctxf->seekg(0);
    }else{
        num = WINSIZE + 1 + WINSIZE * postN;
	/*
        int from = (ctx1st - WINSIZE)*POSSIZ;
	*/
        off_t from = (ctx1st - WINSIZE)*POSSIZ;

/*
        if(0 < ie && from <= ftell(ctxf)){
*/
        if(0 < ie && from <= ctxf->tellg()){
            wordcat = 1;
        }else{
/*
            fseek(ctxf,from,0);
*/
            ctxf->seekg(from);
        }
    }

    if(ie == 0 || catenate || wordcat){
    }else{
        strcpy(bp,delimiter);
        bp += strlen(bp);
        delput = 1;
    }

/*
    int cpos = ftell(ctxf)/POSSIZ;
*/
/*
    int cpos = ctxf->tellg()/POSSIZ;
*/
    off_t cpos = ctxf->tellg()/POSSIZ;
    int ii;
    for(ii = 0; ii < num; ii++){
        if(cpos+1 == next){
            catenate = 1;
            break;
        }

        int lexpos = 0;
/*
        int rcc = Fread(ctxf,(char*)&lexpos,POSSIZ);
*/
        int rcc = ctxf->read((char*)&lexpos,POSSIZ);
        if(rcc != POSSIZ)
            break;
        lexpos = ntohl(lexpos);

        cpos++;
        if(cpos <= last){
            wordcat = 1;
            continue;
        }

        LexiconEntry entry;
        int isLIT = lexpos & O_ISLIT;
        int isBOE = isLIT?0: lexpos & O_BOE;
        char delch = isLIT?0: getPrefix(lexpos); 
        int lexoff = isLIT?0: lexpos & O_LEXOFF;
        char del[32];
        del[0] = 0;

        if(isLIT && (lexpos & 0xFF) == 0 ){ // META
            // should not be counted into the number of displayed words
            continue;
        }
        if(!isLIT && isBOE){
            if(cpos <= ctx1st){
               bp = buf;
               *bp = 0;
            }else{
                break;
            }
        }
        if(isLIT){
            key = keybuf;
            char *kp = keybuf;
            for(int i = 3; 0 <= i; i--){
                 char ch;
                 if((ch = (lexpos >> i*8)))
                     if((ch & 0x7F) != 0 ) // ignore empty byte like 0x80XXYYZZ
                         *kp++ = ch;
            }
            *kp = 0;

            if(strchr(delchs,keybuf[0])){
                char eb[2];
                eb[0] = keybuf[0];
                memmove(keybuf,keybuf+1,strlen(keybuf));
                if(!delput){
                    eb[1] = 0;
                    html_entitize(del,eb);
                }
            }
        }else{
            Lex->getlexstr(lexoff, entry);
            key = entry.getkey();
        }

        if(del[0] == 0){
            if(delch){
             if(!delput){ /* not the top word, shoud insert delimiter */
              char eb[2];
              eb[0] = delchs[delch-1];
              eb[1] = 0;
              html_entitize(del,eb);
             }
            }else
            if(wordcat || 2 <= bp-buf)
            if(key[0] & 0x80){
               if(bp[-2] == key[0] && bp[-1] == key[1]){
                   if(cpos == ctx1st){
                       bp -= 2;
                       *bp = 0;
                   }else{
                       key += 2;
                   }
               }
            }else
            if((key[0] & 0x80) == 0){
               if((bp[-1] & 0x80) == 0){
                   if(strchr(delchs,bp[-1])==0 && strchr(delchs,key[0])==0){ 
                       del[0] = ' ';
                       del[1] = 0;
                   }
               }
            }
        }

        if(putBend){
           putBend = 0;
           strcpy(bp,"</B>");
           bp += strlen(bp);
        }
        if(del[0]){
           strcpy(bp,del);
           bp += strlen(bp);
        }
        if(cpos == ctx1st){
           strcpy(bp,"<B>");
           bp += strlen(bp);
           putBend = 1;
        }
        html_entitize(bp,key);
        bp += strlen(bp);
        catenate = 0;
        delput = 0;
/*
fprintf(stderr,"%2d/%2d [%s] %d[%s][%s]\n",ie,ii,out,delch,del,key);
*/
    }
    if(putBend){
        strcpy(bp,"</B>");
    }
    last = cpos;
  }
}

/*
int SX_newMMap(int fd,int off,int size,void **addrp);
int SX_unmap(void *addr,int len);
*/
static off_t mapsize;

void
/*
ConText::newMap(int ix, int siz){
*/
ConText::newMap(int ix, off_t siz){
	int size;
	int *ptr;
/*
    if(lmap[ix] != 0){
        fprintf(stderr,"OVERWRIT lmap[%d]=%X\n",ix,(int)lmap[ix]);
*/
    if(wsmap[ix] != 0){
        /*
        fprintf(stderr,"OVERWRIT lmap[%d]=%X\n",ix,(int)wsmap[ix]);
        */
        fprintf(stderr,"OVERWRIT lmap[%d]=%X\n",ix,(ArgInt)(PtrInt)wsmap[ix]);
        return;
    }
/*
    lmap[ix] = new int[siz];
    lomap[ix] = new int[siz]; 
*/

	if( 0 < ix ){
		int i,wsfill,lofill;
		wsfill = 0;
		for( i = 0; i < wsmap_size[ix-1]; i++ )
			if( wsmap[ix-1][i] != 0 )
				wsfill++;
		lofill = 0;
		for( i = 0; i < lomap_size[ix-1]; i++ )
			if( lomap[ix-1][i] != 0 )
				lofill++;
		/*
		fprintf(stderr,"--- unmapped %d/%X %d/%X\n",
			wsfill,(int)wsmap[ix-1],lofill,(int)lomap[ix-1]);
		SX_unmap(wsmap[ix-1],4*siz);
		SX_unmap(lomap[ix-1],4*siz);
		wsmap[ix-1] = 0;
		lomap[ix-1] = 0;
		*/
	}
#if 1 || defined(_MSC_VER)
	size = 4*siz;
	ptr = (int*)malloc(size);
	wsmap[ix] = ptr;
	wsmap_size[ix] = siz;

	ptr = (int*)malloc(size);
	lomap[ix] = ptr;
	lomap_size[ix] = siz;
#else
	size = SX_newMMap(fileno(fwsmap),ix*4*siz,4*siz,(void**)&ptr);
	wsmap[ix] = (int*)ptr;
	wsmap_size[ix] = siz;

	size = SX_newMMap(fileno(flomap),ix*4*siz,4*siz,(void**)&ptr);
	lomap[ix] = (int*)ptr;
	lomap_size[ix] = siz;
#endif

	mapsize += size;
	LapReport(0,"... mapped %X %X, %d bytes (TOTAL %lld)",
                /*
		(int)wsmap[ix],(int)lomap[ix],size,(CODEINT)mapsize);
                */
		(ArgInt)(PtrInt)wsmap[ix],(ArgInt)(PtrInt)lomap[ix],size,(CODEINT)mapsize);
/*
		(int)wsmap[ix],(int)lomap[ix],size,mapsize);
*/
}
void
/*
ConText::startMap(int siz){
*/
ConText::startMap(off_t siz){
    lser = 0;
    newMap(LexBase,siz);
    newidx[LexBase] += 1;
}
void
ConText::doneMap(){
    smap[LexBase] = LexSer;
    LexBase++;
}
void
/*
ConText::setMap(int size){
*/
ConText::setMap(off_t size){
    LexSer += size;
    smap[LexBase] = LexSer;
    LexBase++;
}
