// Copyright (c) 2004 Yutaka Sato. All rights reserved.
//
// .ctx -- condensed text for contextual display
// Condensed text representation (not so much condensed)
// - to be used for displaying Context of a searched word
// - consists of a series of Lexicon-IDs

#ifndef CONTEXT_H_INCLUDED
#define CONTEXT_H_INCLUDED

#include "SX.h"
#include "Word.h"
#include "LexiconEntry.h"
#include "LexiconFile.h"
#define LMAPSIZ 2048

class ConText {
private:
    Fstream *ctxf;   // .ctx file
    char **insF;     // .ctx files (input to be merged)
    int insN;        //
    int insX;
    int LexBase;     // serial number of sub-LexiconFile
    int lser;        // serial number of Lexicon in a ConText file
/*
    int LexSer;      // serial number of Lexicon in ConText files (cummulative)
*/
    off_t LexSer;    // serial number of Lexicon in ConText files (cummulative)
/*
    int smap[256];   // LexSer at the end of each map (base in next map)
    int *lmap[256];  // mapping from WordSerialNum (recorded by WordBreaker)
                     // to LexOrderNumber
    int *lomap[256]; // mapping from LexOrderNumber to LexOffset
    char newidx[256];// not existing
*/
	off_t smap[LMAPSIZ]; // LexSer at the end of each map (base in next map)
	char newidx[LMAPSIZ];// not existing

	int *wsmap[LMAPSIZ]; // mapping from WordSerialNum (recorded by
			     // WordBreaker) to LexOrderNumber
	int wsmap_off[LMAPSIZ];
	int wsmap_size[LMAPSIZ];
	void putwsmap(int mi,int li,int off);
	int getwsmap(int mi,int li);

	int *lomap[LMAPSIZ]; // mapping from LexOrderNumber to LexOffset
	int lomap_off[LMAPSIZ];
	int lomap_size[LMAPSIZ];
	void putlomap(int mi,int li,int off);
	int getlomap(int mi,int li);

public:
    ConText();
    ConText(const char *filename, const char *mode);
    ~ConText();
    void close(){
        if(ctxf){
            delete ctxf;
            ctxf = 0;
        }
    }
    int open(const char *path, const char *mode);
    void CtxPutpos(Word &term, ElemCtx &ectx, int off);
    void adjustLex(ConText *octx, IMask *mask);
    void getContext(LexiconFile *Lex, pos_t *pos,int npos, char *buf,int siz, int winsize,const char *delimiter);
    void inputs(char **filenames, int num){ insF = filenames; insN = num; }
    int getctx(char *buf, int size);
    int good(){ return ctxf != NULL && ctxf->good(); }
    int isnewidx(int ix){ return newidx[ix]; }
/*
    void newMap(int ix, int size);
    void startMap(int size);
*/
    void newMap(int ix, off_t size);
    void startMap(off_t size);
    void doneMap();
/*
    void setMap(int size);
*/
    void setMap(off_t size);
    void addkey(const char *key, int off);
    void addpos(int pos);
    void addKeyX(const char *key, int mi, int li, int off);

    int Idxid;
    LexiconFile *lexf;
    int getword(char *word,int size,off_t docbase,int wid);
    void dump();
};

inline void
ConText::addKeyX(const char *key,int mi,int li,int off){
	if( mi < 0 || LMAPSIZ <= mi ){
		fprintf(stderr,"--- addKeyX mi=%d li=%d off=%d\n",mi,li,off);
		exit(-1);
	}
	if( li < 0 || lomap_size[mi] <= li ){
		fprintf(stderr,"--- addKeyX mi=%d li=%d off=%d\n",mi,li,off);
		exit(-1);
	}
	putlomap(mi,li,off);
/*
    lomap[mi][li] = off;
*/
}

// add the mapping for [sorted-order -> offset-in-LexiconFile]
//
inline void
ConText::addkey(const char *key, int off){
    addKeyX(key,LexBase,lser,off);
}

// add the mapping for [created-oder -> sorted-order]
// the initial .ctx is recorded with created-order
//
inline void
ConText::addpos(int kid){
	putwsmap(LexBase,kid,(LexBase<<24)|lser);
/*
    lmap[LexBase][kid] = (LexBase<<24)|lser;
*/
/*
fprintf(stderr,"#### %8X %8X addpos(%d,%d) %d\n",
(int)lmap[LexBase],(int)&lmap[LexBase][kid], LexBase,kid,lser);
*/
    lser++;
}

#endif
