// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Description.cc,v 0.10 1998-01-08 01:45:36+09 harada Exp harada $
//

#include "SX.h"
#include "DescFile.h"
#include "htmlutil.h"

int
DescFile::getRecord(DescRecord &rec, int nth){
    char buf[DSCBLOCKSIZE*2];
    file.clear();
    file.seekg(blksize * nth);
    file.read(buf, blksize);
/*
    if(file.gcount() != (size_t)blksize){
*/
    if((size_t)file.gcount() != (size_t)blksize){
        PRINTERRMSG("read error occured while reading .dsc record.");
        return 0;
    }
    rec.init(buf);
    return 1;
}

int
DescFile::getInts(int nth, int &size, int &date, int &lmod){
    if( file.eof() )
        file.clear();
    file.seekg(blksize * nth);
    file.read((char*)&size,sizeof(size));
    file.read((char*)&date,sizeof(date));
    file.read((char*)&lmod,sizeof(lmod));
    size = ntohl(size);
    date = ntohl(date);
    lmod = ntohl(lmod);
    return 1;
}
int
DescFile::GetLastModified(int docid){
    int lmod = 0;
    if(docid < 0 || (int)((siz*4)/sizeof(FileDescRecord)) < docid){
        return 0;
    }
    int *mm = getmmap();
    if( mm != (int*)-1 ){
/*
        FileDescRecord *desc = (FileDescRecord*)mm;
        if((lmod = desc[docid].d_lastmod) == 0)
            lmod = desc[docid].d_date;
*/
        lmod = ntohl(lmod);
        int *base = &mm[(256/sizeof(int)) * docid];
        lmod = ntohl(base[2]);
    }
    return lmod;
}
char*
DescFile::GetUrl(int docid){
    FileDescRecord *desc = (FileDescRecord*)getmmap();
    if( desc != (FileDescRecord*)-1 ){
        return desc[docid].d_url;
    }
    return "";
}
char*
DescFile::GetAuthor(int docid){
    FileDescRecord *desc = (FileDescRecord*)getmmap();
    if( desc != (FileDescRecord*)-1 ){
        char *u = desc[docid].d_url;
        return u + strlen(u) + 1;
    }
    return "";
}
class NumStr {
public:
    int   num;
    char *str;
};
static int compns(NumStr *a, NumStr *b){
    return strcmp(b->str,a->str); // the higher weight the prior URL
}

char *reversedom(char *dst, char *src){
    char *dp = dst;
    char *sp = src;
    char *yp, *up = 0;
    while( (yp = strrchr(sp,'.')) ){
        if(yp[1]){
            strcpy(dp,yp+1);
            dp += strlen(dp);
            *dp++ = '.';
        }
        *dp = 0;
        *yp = 0;
        if(up) *up = '.';
        up = yp;
    }
    strcpy(dp,sp);
    dp += strlen(dp);
    if(up) *up = '.';
    return dp;
}
void SX_urlForSort(char *srt, char *url){
     char *dp = srt, *sp = url, ch;
     while( (ch = *sp++) ){
         *dp++ = ch;
         if(ch == ':'){
             break;
         }
     }
     if( ch == ':'){
         if(sp[0]=='/' && sp[1]=='/'){
             *dp++ = *sp++;
             *dp++ = *sp++;
             char *up = strpbrk(sp,"/?");
             char uch = 0;
             if(up){
                 uch = *up;
                 *up = 0;
             }
             dp = reversedom(dp,sp);
             sp = up;
             if(up){
                 *up = uch;
             }
         }
         if(sp) // without url-path
         while( (ch = *sp++) ){
             *dp++ = ch;
         }
     }
     *dp = 0;
}
int *makesortUrl(Fstream &descf, int &len){
    if(!descf.good()){
        cerr << "makesortUrl: src file is bad" << endl;
        return 0;
    }
    int fsiz = descf.size();
    if( fsiz == -1 ){
        cerr << "makesortUrl: file size is bad" << endl;
        return 0;
    }
    int siz = fsiz/sizeof(FileDescRecord);
    NumStr *urls;
    FileDescRecord rec;

    urls = new NumStr[siz];

    memset(&rec,0,sizeof(rec));
    for(int ni = 0; ni < siz; ni++){
        urls[ni].num = ni;
        descf.read((char*)&rec,sizeof(rec));
        if(descf.gcount() != sizeof(rec)){
            return 0;
        }
        urls[ni].str = new char[strlen(rec.d_url)+1];
        SX_urlForSort(urls[ni].str,rec.d_url);
    }
    qsort(urls,siz,sizeof(NumStr),(sortfunc)compns);
    int *weighted = new int[siz];
    for(int i = 0; i < siz; i++){
        weighted[urls[i].num] = i;
    }
    len = siz;

    for(int j = 0; j < siz; j++){
        delete []urls[j].str;
    }
    delete []urls;
    return weighted;
}


//
// 040824: copied from any2fdif.c
//
char *wbstrcpy(char *dst, char *dx, const char *src, int len){
	char *dp,ch;
	const char *sp;
	int ic,in2B;

	dp = dst;
	in2B = 0;
	ic = 1; /* reserve for terminator */
	for( sp = src; (len == 0 || ic < len) && (ch = *sp); sp++ ){
		if( dx <= dp )
			break;
		*dp++ = ch;
		ic++;
		if( in2B == 0 ){
			if( ch & 0x80 )
				in2B = 1;
		}else{
			in2B = 0;
		}
	}
	if( in2B ){
		dp--;
	}
	*dp++ = 0;
	return dp;
}
void makeRecord(FileDescRecord &desc, int size, int date, int lastmod,
                int crc32, const char *url, const char *author,
                const char *title, const char *digest){

	memset(&desc,0,sizeof(desc));
	desc.d_size = size;
	desc.d_date = date;
	desc.d_lastmod = lastmod;
	desc.d_crc32 = crc32;

	int rsiz,ulen,alen,tlen,dlen;
	char *dp,*dx;

        /*
	rsiz = sizeof(FileDescRecord) - ((int)&desc.d_url - (int)&desc);
        */
	rsiz = sizeof(FileDescRecord) - (((PtrInt)&desc.d_url) - ((PtrInt)&desc));
	ulen = strlen(url) + 1; 
	alen = strlen(author) + 1; 
	tlen = strlen(title) + 1; 
	dlen = strlen(digest) + 1; 
	if( rsiz < ulen+alen+tlen )
		if( (rsiz*2)/3 < ulen ) ulen = (rsiz*2)/3;
	if( rsiz < ulen+alen+tlen )
		if( rsiz/3 < alen ) alen = rsiz/3;
	if( rsiz < ulen+alen+tlen )
		if( rsiz < tlen ) tlen = (rsiz*3)/4;
	if( rsiz < ulen+alen+tlen+dlen )
		dlen = rsiz - (ulen+alen+tlen);
	dp = desc.d_url;
	dx = desc.d_url + (sizeof(desc.d_url)-1);
	dp = wbstrcpy(dp,dx,url,ulen);
	dp = wbstrcpy(dp,dx,author,alen);
	dp = wbstrcpy(dp,dx,title,tlen);
	dp = wbstrcpy(dp,dx,digest,dlen);
}
void
writeRecord(Fstream *descf,
    int size, int date, int time, int crc32,
    const char *url, const char *author, const char *title, const char *digest){

	FileDescRecord desc[2];
	makeRecord(desc[0],size,date,time,crc32,url,author,title,digest);
	for(int i = 0; i < D_NINTS; i++)
		desc[0].d_ints[i] = htonl(desc[0].d_ints[i]);
	descf->write((char*)&desc[0],sizeof(desc[0]));
}

int
makesumfiles(Fstream &descf, Fstream &sumf){
    FileDescRecord rec;
    char line[0x10000];
    int i;
    for(i = 0;;i++){
        descf.read((char*)&rec,sizeof(rec));
        if(descf.gcount() != sizeof(rec)){
            break;
        }

        sprintf(line,"%08d %08X %08X %08X %8d %s\n",i,
            ntohl(rec.d_date), ntohl(rec.d_lastmod), ntohl(rec.d_crc32),
            ntohl(rec.d_size), rec.d_url);
        sumf.write(line,strlen(line));
    }
    return i;
}

#define BSIZE (0x1000/4)
#define NHEAD 4
NumFile::NumFile(const char *path){
	for(int i = 0; i < N_NUMCRCS; i++){
		fps[i] = 0;
		vals[i] = 0;
	}
	nelem = 0;
	bsize = 0;
	fpr = fopen(path,"r");
	if(fpr != NULL){
		Fread(fpr,(char*)&nelem,sizeof(int));
		Fread(fpr,(char*)&bsize,sizeof(int));
		nelem = htonl(nelem);
		bsize = htonl(bsize);
/*
fprintf(stderr,"### NE=%d BS=%d %s\n",nelem,bsize,path);
*/
	}
}
int
NumFile::get(int did, int cid){
	if(nelem <= 0)
		return -1;
	if(did < 0 || nelem <= did)
		return -1;
	if(cid < 0 || N_NUMCRCS <= cid)
		return -1;
/*
fprintf(stderr,"### [%d][%d] ne=%d bsize=%d\n",cid,did,nelem,bsize);
*/
	if(vals[cid] == 0){
		int fd = fileno(fpr);
		off_t off = cid*bsize;
		vals[cid] = (int*)mmap(NULL,bsize,PROT_READ,MAP_SHARED,fd,off);
	}
	if(vals[cid] && vals[cid] != (int*)-1){
/*
fprintf(stderr,"[%d][%d] = %d\n",cid,did,vals[cid][did]);
*/
		return ntohl(vals[cid][NHEAD+did]);
	}
	return -1;
}
NumFile::NumFile(const char *path, int size){
	for(int i = 0; i < N_NUMCRCS; i++)
		fps[i] = 0;
	FILE *ufp = fopen(path,"w+");
/*
fprintf(stderr,"### size=%d %X %s\n",size,ufp,path);
*/
	if( ufp != NULL ){
		int bsize = ((NHEAD+size+BSIZE-1)/BSIZE)*BSIZE * sizeof(int);
		for(int i = 0; i < N_NUMCRCS; i++){
			fps[i] = fopen(path,"r+");
			if(fps[i] == NULL){
				break;
			}
			fseek(fps[i],i*bsize,0);
/*
fprintf(stderr,"### %0o\n",ftell(fps[i]));
*/
			int xsize = htonl(size);
			int xbsize = htonl(bsize);
			Fwrite(fps[i],(char*)&xsize,sizeof(int));
			Fwrite(fps[i],(char*)&xbsize,sizeof(int));
			fseek(fps[i],i*bsize+NHEAD*sizeof(int),0);
/*
fprintf(stderr,"##> size=%d %X %d %s\n",size,fps[i],ftell(fps[i]),path);
*/
		}
		fclose(ufp);
	}
}
NumFile::~NumFile(){
	for(int i = 0; i < N_NUMCRCS; i++){
		if(fps[i])
			fclose(fps[i]);
	}
}
int
NumFile::write(int did, FileDescRecord &rec, int usort){
	int val[8];
	val[N_URLCRC]  = htonl(strCRC32(0,rec.d_url));
	val[N_CONTCRC] = rec.d_crc32;
	val[N_SIZE]    = rec.d_size;
	val[N_DATE]    = rec.d_date;
	val[N_LASTMOD] = rec.d_lastmod;
	val[N_USORT]   = htonl(usort);

	for(int i = 0; i < N_NUMCRCS; i++){
/*
fprintf(stderr,"[%2d] %8d %8o %X\n",i,ftell(fps[i]),ftell(fps[i]),val[i]);
*/
		Fwrite(fps[i],(char*)&val[i],sizeof(int));
	}
	return 0;
}

int findUrl(NumFile &nums, const char *url){
	int nsize = nums.size();
	int crc = strCRC32(0,url);
	int did;
	int found_did = -1;
	for(did = 0; did < nsize; did++){
		if( nums.get(did,N_URLCRC) == crc){
			found_did = did;
			break;
		}
	}
	return found_did;
}
