// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Description.cc,v 0.10 1998-01-08 01:45:36+09 harada Exp harada $
//

#include "SX.h"
#include "DescRecord.h"
#include "htmlutil.h"

DescRecord::DescRecord(size_t blocksz)
    : blksize(blocksz), size(0), date(0), lastmodified(0),
    crc32(0),
    url(NULL), author(NULL), title(NULL), digest(NULL)
{
}

DescRecord::DescRecord(const char *buf, size_t blocksz)
    : blksize(blocksz), size(0), date(0), lastmodified(0),
    crc32(0),
    url(NULL), author(NULL), title(NULL), digest(NULL)
{
    (void)init(buf);
}

DescRecord::~DescRecord(){
    if(url) delete []url;
    if(author) delete []author;
    if(title) delete []title;
    if(digest) delete []digest;
}

void
DescRecord::init(const char *buf){
/*
    size = ntohl(*(size_t *)buf);
    date = ntohl(*(time_t *)(buf + sizeof(size)));
    lastmodified = ntohl(*(time_t *)(buf + sizeof(size) + sizeof(date)));
    const char *cp = buf + sizeof(size) + sizeof(date) + sizeof(lastmodified);
*/
    FileDescRecord *desc = (FileDescRecord*)buf;
    size = ntohl(desc->d_size);
    date = ntohl(desc->d_date);
    lastmodified = ntohl(desc->d_lastmod);
    crc32 = ntohl(desc->d_crc32);
    const char *cp = desc->d_url;

    url = mystrdup(cp);
    while(*cp && cp < buf + blksize - 1) cp++;

    author = mystrdup(++cp);
    while(*cp && cp < buf + blksize - 1) cp++;

    title = mystrdup(++cp);
    while(*cp && cp < buf + blksize - 1) cp++;
    
    // digest part might not be terminated by NUL.
    digest = new char[blksize - (cp - buf) + 1];
    digest[0] = 0; /* fix-140621a digest part might not exist */
/*
    strncpy(digest, ++cp, blksize - (cp - buf));
*/
    cp++;
    strncpy(digest, cp, blksize - (cp - buf));
}

void
DescRecord::reset(){
    if(url) delete []url;
    if(author) delete []author;
    if(title) delete []title;
    if(digest) delete []digest;
    url = author = title = digest = NULL;
    size = 0;
    date = lastmodified = 0;
}

DescRecord &
DescRecord::operator =(const DescRecord &that){
    reset();
    size = that.size;
    date = that.date;
    lastmodified = that.lastmodified;
    if(that.url) url = mystrdup(that.url);
    if(that.author) author = mystrdup(that.author);
    if(that.title) title = mystrdup(that.title);
    if(that.digest) digest = mystrdup(that.digest);

    return *this;
}

ostream&
operator <<(ostream &stream, DescRecord &that){
    char datebuf[80];
    if(that.date != 0)
        stream << "Date: "
               << rfc822date(datebuf, sizeof(datebuf), that.date) << endl;
    if(that.lastmodified != 0)
        stream << "Last-Modified: "
               << rfc822date(datebuf, sizeof(datebuf), that.lastmodified)
               << endl;
    if(that.url && that.url[0])
        stream << "URL: " << that.url << endl;
    if(that.title && that.title[0])
        stream << "Title: \"" << that.title << '"' << endl;
    if(that.author && that.author[0])
        stream << "Author: \"" << that.author << '"' << endl;
    if(that.digest && that.digest[0])
        stream << "Digest: \"" << that.digest << '"' << endl;
    return stream;
}
