// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Evaluator.cc,v 0.9 1997-12-27 07:10:40+09 harada Exp harada $
//

#include "SX.h"
#include "Evaluator.h"

int
/*
Evaluator::and(Result *dst, int dstLen, const Result *src, int srcLen) const {
*/
Evaluator::And(Result *dst, int dstLen, const Result *src, int srcLen) const {
    if(dst == NULL || dstLen == 0 || src == NULL || srcLen == 0)
        return 0;
    int r = 0;
    int d = 0;
    int s = 0;
    while(d < dstLen && s < srcLen){
        assert(dst[d].ntimes > 0);
        assert(src[s].ntimes > 0);
        if(dst[d].docid == src[s].docid){
            dst[r] = dst[d];    // docid, docsize
            saddintvl(dst[r].list,sizeof(dst[r].list),src[s].list,sizeof(src[s].list));
            dst[r].ntimes = dst[d].ntimes < src[s].ntimes
                ? dst[d].ntimes : src[s].ntimes; // is this valid?
            dst[r].weight = 2 * (dst[d].weight + src[s].weight);
            r++; d++; s++;
        }else if(dst[d].docid < src[s].docid){
            d++;
        }else
            s++;
    }
    return r;
}


int
/*
Evaluator::not(Result *dst, int dstLen, const Result *src, int srcLen) const {
*/
Evaluator::Not(Result *dst, int dstLen, const Result *src, int srcLen) const {
    if(src == NULL || srcLen == 0 || dstLen == 0)
        return dstLen;
    
    Result *tmp = new Result[dstLen];
    memcpy(tmp, dst, sizeof(Result) * dstLen);
    int r = 0;
    int d = 0;
    int s = 0;
    while(d < dstLen && s < srcLen){
        if(dst[d].docid < src[s].docid) tmp[r++] = dst[d++];
        else if(dst[d].docid > src[s].docid) s++;
        else { d++; s++; }
    }
    while(d < dstLen)
        tmp[r++] = dst[d++];
    memcpy(dst, tmp, sizeof(Result) * r);
    delete []tmp;
    return r;
}


int
/*
Evaluator::or(Result *dst, const Result *src1, int srcLen1,
*/
Evaluator::Or(Result *dst, const Result *src1, int srcLen1,
           const Result *src2, int srcLen2) const {
    if(src2 == NULL || srcLen2 == 0){
        if(src1 == NULL || srcLen1 == 0) return 0;
        memcpy(dst, src1, sizeof(Result) * srcLen1);
        return srcLen1;
    }
    if(src1 == NULL || srcLen1 == 0){
        memcpy(dst, src2, sizeof(Result) * srcLen2);
        return srcLen2;
    }

    int r = 0;
    int s1 = 0;
    int s2 = 0;
    for(;;){
        if(s1 >= srcLen1){
            while(s2 < srcLen2)
                dst[r++] = src2[s2++];
            break;
        }
        if(s2 >= srcLen2){
            while(s1 < srcLen1)
                dst[r++] = src1[s1++];
            break;
        }
        if(src1[s1].docid < src2[s2].docid)
            dst[r++] = src1[s1++];
        else if(src1[s1].docid > src2[s2].docid)
            dst[r++] = src2[s2++];
        else{
            dst[r] = src1[s1];
/*
            dst[r].ntimes = src1[s1].ntimes + src2[s2].ntimes;
*/
            int ntimes = src1[s1].ntimes + src2[s2].ntimes;
            if(255 < ntimes)
                ntimes = 255;
            dst[r].ntimes = ntimes;
            dst[r].weight = 2 * (src1[s1].weight + src2[s2].weight);
            r++; s1++; s2++;
        }
    }
    return r;
}


static int
compare_weight(const Result *a, const Result *b){
    return (b->weight > a->weight) ? 1 : -1;
}

int
Evaluator::rank(Result *result, int resultLen) const {
    int i;
    float maxweight = 0;
    for(i = 0; i < resultLen; i++){
        if(result[i].weight > maxweight)
            maxweight = result[i].weight;
    }
    int sortLen = (resultLen > MAXSORTRESULTS) ? MAXSORTRESULTS : resultLen;
    qsort(result, sortLen, sizeof(Result),
          (int (*)(const void *, const void *))compare_weight);

    // re-adjust scores so that the maximum weight to be 1000
    double mag = (double)1000 / maxweight;
    for(i = 0; i < resultLen; i++)
        result[i].weight = result[i].weight * mag;
    return 0;
}

