// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Expression.cc,v 0.12 1998-02-09 18:41:22+09 harada Exp $
//

#include "SX.h"
#include "Query.h"
#include "Expression.h"
#include "kcnvlib.h"
int SXopAND = 1;

char *Expression::errmsgs[] = {
    "",
    "empty query found",
    "asterisk should locate the last",
    "(dummy)",
    "not implemented",
    "empty expression",
    "prefix operator has no operand",
    "unterminated phrase",
    "at least one positive query required", 
};

int
Expression::fixstring(char *dst, const char *src) const {
    const char *cp = src;
    char *dp = dst;
    int replaced = 0;
    while(*cp){
        if(Is_kanji_here(cp)){
            *dp++ = *cp++;
            *dp++ = *cp++;
        }else{
            if(isalnum(*(unsigned char *)cp) || isspace(*(unsigned char *)cp)
               || *cp == '"' || *cp == '*' || *cp == '-'
               || *cp == '^' || *cp == '$'
               || *cp == '_' || *cp == '+')
                *dp++ = *cp++;
            else{
                *dp++ = ' ';
                cp++;
                replaced++;
            }
        }
    }
    *dp = '\0';
    return replaced;
}

SimpleExpression::SimpleExpression(const char *expression)
    : queryNum(0)
{
    if(expression == NULL){
        parse_error = EMPTYEXPRESSION;
        return;
    }
    for(int i = 0; i < MAXQUERIESINEXPRESSION; i++){
        queries[i] = NULL;
        reqtypes[i] = BASIC;
    }
    parse_error = parse(expression);
}

SimpleExpression::~SimpleExpression(){
    for(int i = 0; i < queryNum; i++)
        delete queries[i];
}

void
SimpleExpression::printDebug(ostream &out){
    out << "SimpleExpression: ";
    out << " queryNum = " << queryNum << ' ';
    for(int i = 0; i < queryNum; i++){
        if(SXopAND){
            if(reqtypes[i] == BASIC) out << '+';
            else if(reqtypes[i] == EXCLUDE) out << '-';
        }else
        if(reqtypes[i] == REQUIRE) out << '+';
        else if(reqtypes[i] == EXCLUDE) out << '-';
        queries[i]->print(out);
        out << ' ';
    }
    out << ") ";
}

ostream &
SimpleExpression::print(ostream &out){
    for(int i = 0; i < queryNum; i++){
        if(SXopAND){
            if(reqtypes[i] == BASIC) out << '+';
            else if(reqtypes[i] == EXCLUDE) out << '-';
        }else
        if(reqtypes[i] == REQUIRE) out << '+';
        else if(reqtypes[i] == EXCLUDE) out << '-';
        queries[i]->print(out);
        if(i < queryNum - 1) out << ' ';
    }
    return out;
}

void
SimpleExpression::addquery(Query *query, int reqtype){
    if(queryNum == MAXQUERIESINEXPRESSION)
        return;                 // ignore 
    queries[queryNum] = query;
    reqtypes[queryNum] = reqtype;
    queryNum++;
}

int
SimpleExpression::parse(const char *expression){
    char buf[MAXQUERYLENGTH];
    int reqtype;
    int inphrase;

    // cout << " parse = " << expression << endl;
    const char *cp = expression;
    while(*cp){
        inphrase = 0;
        reqtype = BASIC;
        char *dp = buf;
        *dp = '\0';
        if(isspace(*(unsigned char *)cp)){
            do { cp++; } while(isspace(*(unsigned char *)cp));
            continue;
        }
        if(SXopAND){
           if(*cp == '+'){
               cp++;
           }else
           if(*cp == '-'){
               reqtype = EXCLUDE;
               cp++;
           }else
               reqtype = REQUIRE;
        }else
        if(*cp == '+'){
            reqtype = REQUIRE;
            cp++;
        }else if(*cp == '-'){
            reqtype = EXCLUDE;
            cp++;
        }
        if(*cp == '"'){
            inphrase = 1;
            cp++;
        }else if(*cp == '\0' || isspace(*(unsigned char *)cp))
            return PREFIXHASNOOPERAND;

/*
        int type = 0;
*/
        int type = TYPE_ANY;
        if(! inphrase){
            if((type = ElemId(cp,&cp,":")) < 0)
                type = TYPE_ANY;
/*
                type = 0;
*/
/*
            // TODO: rewrite with a suitable factory method
            if(0 == strncmp(cp, "title:", 6)){
                type = 2; cp += 6;
            }else if(0 == strncmp(cp, "keyword:", 8)){
                type = 4; cp += 8;
            }else if(0 == strncmp(cp, "author:", 7)){
                type = 3; cp += 7;
            }else if(0 == strncmp(cp, "url:", 4)){
                type = 1; cp += 4;
            }
#if 0
            else if(0 == strncmp(cp, "summary:", 8)){
                type = 5; cp += 8;
            }else if(0 == strncmp(cp, "special:", 8)){
                type = 6; cp += 8;
            }
#endif
*/
/*
            if(type != 0 && *cp == '"'){
*/
            if(type != TYPE_ANY && *cp == '"'){
                inphrase = 1;
                cp++;
            }else{
                while(*cp && *cp != '"' && !isspace(*(unsigned char *)cp)){
                    *dp++ = *cp++;
                }
                *dp = '\0';
                char tmpstr[MAXQUERYLENGTH];
                if(fixstring(tmpstr, buf)){
                    PhraseQuery *phraseq = new PhraseQuery(tmpstr, type);
                    if(phraseq->error())
                        return phraseq->error();
                    addquery(phraseq, reqtype);
                }else{
                    SimpleQuery *simpleq = new SimpleQuery(buf, type);
                    if(simpleq->error())
                        return simpleq->error();
                    addquery(simpleq, reqtype);
                }
            }
        }
        if(inphrase){
            while(*cp && *cp != '"')
                *dp++ = *cp++;
            if(*cp == '\0')
                return UNTERMINATEDPHRASE;
            else{
                assert(*cp == '"');
                cp++;
            }
            *dp = '\0';
            char tmpstr[MAXQUERYLENGTH];
            fixstring(tmpstr, buf);
            PhraseQuery *phraseq = new PhraseQuery(tmpstr, type);
            if(phraseq->error()){
                return phraseq->error();
            }
            addquery(phraseq, reqtype);
        }
    }

    if(queryNum == 0)
        return EMPTYEXPRESSION;

    for(int i = 0; i < queryNum; i++){
        if(reqtypes[i] != EXCLUDE) return SUCCESS;
    }
    return ONLYNOTOPUSED;
}


Result *
SimpleExpression::eval(Retriever &ret, const Evaluator &evaluator,
                       int &resultLen){
    int i;
    resultLen = 0;

    // (1) process all required queries, and multiply them.
    Result *required = NULL;
    int requiredLen = 0;
    for(i = 0; i < queryNum; i++){
        if(reqtypes[i] != REQUIRE) continue;
        int tmpLen = 0;
        Result *tmp = queries[i]->eval(ret, tmpLen);
        if(tmp == NULL){        // not found
            if(required) delete []required;
            return NULL;
        }
        if(required == NULL){
            required = tmp;
            requiredLen = tmpLen;
        }else{
/*
            requiredLen = evaluator.and(required, requiredLen, tmp, tmpLen);
*/
            requiredLen = evaluator.And(required, requiredLen, tmp, tmpLen);
            if(requiredLen == 0){   // not found
                delete []tmp;
                delete []required;
                return NULL;
            }
            delete []tmp;
        }
    }
    
    // (2) process all basic( neither required nor excluded ) queries,
    //     and summing them.
    Result *basic = NULL;
    int basicLen = 0;
    for(i = 0; i < queryNum; i++){
        if(reqtypes[i] != BASIC) continue;
        int tmpLen;
        Result *tmp = queries[i]->eval(ret, tmpLen);
        if(tmp != NULL){
            if(basic == NULL){
                basic = tmp;
                basicLen = tmpLen;
            }else{
/*
                Result *or = new Result[basicLen + tmpLen];
                int orLen = evaluator.or(or, basic, basicLen, tmp, tmpLen);
*/
                Result *Or = new Result[basicLen + tmpLen];
                int orLen = evaluator.Or(Or, basic, basicLen, tmp, tmpLen);
                delete []basic;
                delete []tmp;
/*
                basic = or;
*/
                basic = Or;
                basicLen = orLen;
            }
        }
    }

    // (3) multiply (1) and (2)
    if(required){
        if(basic){
/*
            requiredLen = evaluator.and(required, requiredLen, basic, basicLen);
*/
            requiredLen = evaluator.And(required, requiredLen, basic, basicLen);
            delete []basic;
        }
    }else{
        if(basic){
            required = basic;
            requiredLen = basicLen;
        }else{
            return NULL;
        }
    }
    if(requiredLen == 0){
        delete []required;
        return NULL;
    }

    // (4) Finally, process excluded queries and summing up.
    assert(required != NULL);
    assert(requiredLen > 0);
    for(i = 0; i < queryNum; i++){
        if(reqtypes[i] != EXCLUDE) continue;
        int tmpLen;
        Result *tmp = queries[i]->eval(ret, tmpLen);
        if(tmp){
/*
            requiredLen = evaluator.not(required, requiredLen, tmp, tmpLen);
*/
            requiredLen = evaluator.Not(required, requiredLen, tmp, tmpLen);
            delete []tmp;
            if(requiredLen == 0){ // not found
                delete []required;
                return NULL;
            }
        }
    }

    //evaluator.rank(required, requiredLen);
    resultLen = requiredLen;
    return required;
}

// ------------------------------------------------------------

ComplexExpression::ComplexExpression(const char *expression)
    : expNum(0)
{
    if(expression == NULL){
        parse_error = EMPTYEXPRESSION;
        return;
    }
    exps[0] = exps[1] = NULL;
    parse_error = parse(expression);
}

ComplexExpression::~ComplexExpression(){
    if(exps[1]) delete exps[1];
    if(exps[0]) delete exps[0];
}

Result *
ComplexExpression::eval(Retriever &ret, const Evaluator &evaluator,
                        int &resultLen){
    resultLen = 0;
    if(expNum == 0)
        return NULL;
    else if(expNum == 1){
        return exps[0]->eval(ret, evaluator, resultLen);
    }else{
        assert(expNum == 2);
        int resultLen1, resultLen2;
        Result *result1 = exps[0]->eval(ret, evaluator, resultLen1);
        if(NULL == result1 || resultLen1 == 0)
            return NULL;        // not found
        Result *result2 = exps[1]->eval(ret, evaluator, resultLen2);
        if(NULL == result2 || resultLen2 == 0){
            delete []result1;
            return NULL;        // not found
        }
/*
        resultLen = evaluator.and(result1, resultLen1, result2, resultLen2);
*/
        resultLen = evaluator.And(result1, resultLen1, result2, resultLen2);
        delete []result2;
        if(resultLen == 0){
            delete []result1;
            return NULL;        // not found
        }
        return result1;
    }
}

void
ComplexExpression::printDebug(ostream &out){
    out << "ComplexExpression: ";
    out << " parse_error = " << parse_error;
    out << " expNum = " << expNum << endl << " (";
    if(expNum){
        exps[0]->print(out);
        for(int i = 1; i < expNum; i++){
            out << " & ";
            exps[i]->print(out);
        }
    }
    out << ")." << endl;
}

ostream &
ComplexExpression::print(ostream &out){
    if(expNum == 0){
        out << ' ';
    }else if(expNum == 1){
        exps[0]->print(out);
    }else if(expNum == 2){
        exps[0]->print(out);
        out << " & ";
        exps[1]->print(out);
    }else{                      // not implemented yet
        for(int i = 0; i < expNum; i++)
            exps[i]->print(out);
    }
    return out;
}

int
ComplexExpression::parse(const char *expression){
    char buf[MAXQUERYLENGTH];
    const char *cp = expression;
    int inphrase = 0;
    char *dp = buf;

    *dp = '\0';
    if(isspace(*(unsigned char *)cp)){
        do{ cp++; }while(isspace(*(unsigned char *)cp));
    }
    if(*cp == '\0')
        return EMPTYEXPRESSION;
    while(*cp){
        if(inphrase){
            if(*cp == '"'){     // end of phrase
                *dp++ = *cp++;
                *dp++ = ' ';
                inphrase = 0;
                continue;
            }
        }else{
            if(*cp == '"'){     // beginning of phrase
                *dp++ = *cp++;
                inphrase = 1;
                continue;
            }
            if(*cp == '|' || *cp == '(' || *cp == ')'){
                return NOTIMPLEMENTED;
            }
            if(*cp == '&' || *cp == '>'){
                *dp = '\0';
                SimpleExpression *leftexp = new SimpleExpression(buf);
                if(leftexp->error())
                    return leftexp->error();
                push(leftexp);
                cp++;
                while(isspace(*(unsigned char *)cp)) cp++;
                if(*cp){
                    ComplexExpression *rightexp = new ComplexExpression(cp);
                    if(rightexp->error())
                        return rightexp->error();
                    push(rightexp);
                }
                return SUCCESS;
            }
        }
        *dp++ = *cp++;
    }
    *dp = '\0';
    if(inphrase){
        // cout << " UNTERMINATEDPHRASE! ";
        return UNTERMINATEDPHRASE;
    }
    SimpleExpression *exp = new SimpleExpression(buf);
    if(exp->error())
        return exp->error();
    push(exp);
    return SUCCESS;
}

#ifdef EXPRESSION_DEBUG
int main(){
    ComplexExpression cexp("+harada harold -masanori & -\"foo bar\" ľ");
    cexp.printDebug();

    return 0;
}
#endif
