// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Expression.h,v 0.9 1997-12-27 07:11:42+09 harada Exp $
//
#ifndef EXPRESSION_H_INCLUDED
#define EXPRESSION_H_INCLUDED

#include "SX.h"
#include "Query.h"
#include "Retriever.h"
#include "Result.h"
#include "Evaluator.h"

// abstract class
class Expression {
protected:
    int parse_error;
    int fixstring(char *dst, const char *src) const;
public:
    enum error_type {
        SUCCESS = 0,
        DUMMY = Query::END_OF_QUERY_ERROR,
        NOTIMPLEMENTED,         // ޤƤޤ
        EMPTYEXPRESSION,        // ޤ
        PREFIXHASNOOPERAND,     // +,-θ˼ޤ
        UNTERMINATEDPHRASE,     // ե졼ĤƤޤ
        ONLYNOTOPUSED,          // NOTΤߤμϽǤޤ
    };
    static char *errmsgs[];
    Expression() : parse_error(SUCCESS) { }
    virtual ~Expression() { }
    virtual Result *eval(Retriever &ret, const Evaluator &evaluator,
                         int &resultLen) = 0;
    operator bool() const { return parse_error == SUCCESS; }
    int error() const { return parse_error; }
    virtual ostream &print(ostream &out = cout) = 0;
    virtual void printDebug(ostream &out = cout) = 0;
};

class SimpleExpression : public Expression {
protected:
    enum requirement { BASIC, REQUIRE, EXCLUDE };
    int queryNum;
    Query *queries[MAXQUERIESINEXPRESSION];
    int reqtypes[MAXQUERIESINEXPRESSION];
    virtual int parse(const char *expression);
    void addquery(Query *query, int reqtype);
public:
    SimpleExpression(const char *expression);
    virtual ~SimpleExpression();
    virtual Result *eval(Retriever &ret, const Evaluator &evaluator,
                         int &resultLen);
    virtual ostream &print(ostream &out = cout);
    virtual void printDebug(ostream &out = cout);
};

class ComplexExpression : public Expression {
protected:
    int expNum;
    Expression *exps[2];
    void push(Expression *exp){ exps[expNum++] = exp; }
    virtual int parse(const char *expression);
public:
    ComplexExpression(const char *expression);
    virtual ~ComplexExpression();
    virtual Result *eval(Retriever &ret, const Evaluator &evaluator,
                         int &resultLen);
    virtual ostream &print(ostream &out = cout);
    virtual void printDebug(ostream &out = cout);
};

#endif
