// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: HashBag.h,v 0.9 1997-12-27 07:14:00+09 harada Exp $
//
#ifndef HASHBAG_H_INCLUDED
#define HASHBAG_H_INCLUDED

#include "SX.h"
#include "HashKey.h"

#define PRIMETABLESIZE (256)

extern unsigned int primeTbl[];
extern int HashLevel;

typedef void (*hashDumpFunc)(int ko, void *key, void *val);
typedef int (*hashSortFunc)(const void *, const void*);

template <class T>
class HashBag{
private:
    int slotNum;
    int usedNum;
    int cursor;
    int accessNum;
    int collisionNum;
    HashKey **keyslot;
    T *valueslot;
    unsigned int hash2(unsigned int h){
        return primeTbl[h % PRIMETABLESIZE];
    }
public:
    HashBag(int n);
    HashBag(const HashBag &that);
    ~HashBag();
    int get(const HashKey &key, T &value);
    int set(const HashKey &key, const T &value);
/*
    T & operator[](const HashKey &key);
*/
    T & operator[](HashKey &key);
    T *Valueslot(){ return valueslot; }
    operator int() const { return usedNum; }
    int size() const { return usedNum; }
    int capacity() const { return slotNum; }
    int accessed() const { return accessNum; }
    int collision() const { return collisionNum; }
    void rewind(){ cursor = 0; };
    int next(HashKey *&keyp, T *&valuep);
    int printDebug();
    int sortdump(hashDumpFunc func, int byvsize);
    void keystats(char *buf,int bsiz);
};

template <class T>
void HashBag<T>::keystats(char *buf,int bsiz){
	snprintf(buf,bsiz,
		"%4.1f%% (%7d/%7d)%5.1f%% coll.(%6d/%6d)",
		usedNum/(slotNum/100.0),
		usedNum,slotNum,
		collisionNum/(accessNum/100.0),
		collisionNum,accessNum);
}

// nʬΥǡ뤿Υϥåơ֥
// ưŪ˥礭ʤäϤʤ
template <class T>
HashBag<T>::HashBag(int n)
    : slotNum(16), usedNum(0), cursor(0)
{
    for(;;){
/*
        if(slotNum > 3 * n) break;
*/
        if(slotNum >= 2 * n) break;
        slotNum <<= 1;          // slotNum ɬ(äǿˤʤʤ)
    }
    typedef HashKey *HKP;
    keyslot = new HKP[slotNum];
    valueslot = new T[slotNum];
    memset(keyslot, 0, sizeof(HashKey *) * slotNum);
    accessNum = 0;
    collisionNum = 0;
if( 1024 < n )
LapReport(0,"new Hash(%d) %d slots, %d bytes, func#%d",
(int)n,(int)slotNum,(int)slotNum*(sizeof(HKP)+sizeof(T)),HashLevel);
}


template <class T>
HashBag<T>::HashBag(const HashBag &that)
    : slotNum(that.slotNum), usedNum(that.usedNum), cursor(that.cursor)
{
/*
    keyslot = new (HashKey *)[slotNum];
*/
    keyslot = new HashKey *[slotNum];
    valueslot = new T[slotNum];
    memset(keyslot, 0, sizeof(HashKey *) * slotNum);
    for(int i = 0; i < slotNum; i++){
        if(that.keyslot[i]){
            keyslot[i] = that.keyslot[i]->clone();
            valueslot[i] = that.valueslot[i];
        }
    }
    accessNum = 0;
    collisionNum = 0;
}


template <class T>
HashBag<T>::~HashBag(){
if( 128*1024 < usedNum )
LapReport(0,"... Hash::free keys...");
    for(int i = 0; i < slotNum; i++)
        if(keyslot[i]) delete (keyslot[i]);
    if(keyslot) delete []keyslot;
if( 128*1024 < usedNum )
LapReport(0,"... Hash::free values...");
    if(valueslot) delete []valueslot;
}


template <class T> int
HashBag<T>::printDebug(){
    cout << "---HashBag---" << endl;
    cout << "slotNum = " << slotNum << endl;
    cout << "usedNum = " << usedNum << endl;
    for(int i = 0; i < slotNum; i++){
        if(keyslot[i] != NULL){
            cout << setw(4) << i << ": \""
                 << keyslot[i]->toString() << "\" \t" << valueslot[i] << endl;
        }
    }
    return 0;
}


template <class T> int
HashBag<T>::set(const HashKey &key, const T &value){
    unsigned int h, h2;
    h = key.hashCode() % slotNum;
    h2 = hash2(h);
    accessNum++;
    while(NULL != keyslot[h]){
        if(key == *(keyslot[h])){
            valueslot[h] = value;
            return 0;
        }
        h = (h + h2) % slotNum;
        collisionNum++;
    }
    keyslot[h] = key.clone();
    valueslot[h] = value;
    usedNum++;
    return 0;
}


template <class T> T &
/*
HashBag<T>::operator[](const HashKey &key){
*/
HashBag<T>::operator[](HashKey &key){
    unsigned int h, h2;
    h = key.hashCode() % slotNum;
    h2 = hash2(h);
    accessNum++;
    for(;;){
        if(NULL == keyslot[h]){
            key.kid = usedNum;
            usedNum++;
            keyslot[h] = key.clone();
            return valueslot[h];
        }
/*
        if(key == *(keyslot[h])) return valueslot[h];
*/
        if(key == *(keyslot[h])){
            key.kid = (*keyslot[h]).kid;
            return valueslot[h];
        }
        h = (h + h2) % slotNum;
        collisionNum++;
    }
}

// Ĥä0ʳ֤
template <class T> int
HashBag<T>::get(const HashKey &key, T &value){
    unsigned int h, h2;
    h = key.hashCode() % slotNum;
    h2 = hash2(h);
    accessNum++;
    for(;;){
        if(NULL == keyslot[h]) return 0;
        if(key == *(keyslot[h])){
            value = valueslot[h];
            return 1;
        }
        h = (h + h2) % slotNum;
        collisionNum++;
    }
}

// ãƤ顤0 ֤
template <class T> int
HashBag<T>::next(HashKey *& keyp, T *&valuep){
    for( ; cursor < slotNum; cursor++){
        if(NULL != keyslot[cursor]){
            keyp = keyslot[cursor];
            valuep = &valueslot[cursor];
            cursor++;
            return 1;
        }
    }
    return 0;
}

static HashKey **kslot;
static int kcomp(int *a, int *b){
    return strcmp(kslot[*a]->toString(),kslot[*b]->toString());
}

static int *vslot;
static int vcomp(int *a, int *b){
    return vslot[*b] - vslot[*a];
}

template <class T> int
HashBag<T>::sortdump(hashDumpFunc dumpfunc, int byvsize){
    int *slots = new int[usedNum];
    int sx = 0;
    int i;

    if(byvsize){
       int *vsize = new int[usedNum];
       int *slotx = new int[usedNum];
       for(i = 0; i < slotNum; i++){
           if(keyslot[i]){
               slotx[sx] = i;
               vsize[sx] = valueslot[i].size();
/*
               slots[sx++] = sx;
*/
               slots[sx] = sx;
               sx++;
           }
       }
       vslot = vsize;
       qsort(slots,usedNum,sizeof(slots[0]),(hashSortFunc)vcomp);

       for(i = 0; i < usedNum; i++){
           int x = slotx[slots[i]];
           (*dumpfunc)(i, keyslot[x], &valueslot[x]);
       }
       delete []vsize;
       delete []slotx;
    }else{
       for(i = 0; i < slotNum; i++){
           if(keyslot[i]){
               slots[sx++] = i;
           }
       }
       kslot = keyslot;
       LapReport(0,"... Hash::sortdump(%d) qsort...",usedNum);
       qsort(slots,usedNum,sizeof(slots[0]),(hashSortFunc)kcomp);
       LapReport(0,"... Hash::sortdump qsort done, dump indices...");
       for(i = 0; i < usedNum; i++){
           int x = slots[i];
           (*dumpfunc)(i, keyslot[x], &valueslot[x]);
       }
       LapReport(0,"... Hash::sortdump dump indices done");
    }
    delete []slots;
    return usedNum;
}

#endif
