// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Indexer.cc,v 0.13 1998-06-22 00:48:34+09 harada Exp $
//

#include "SX.h"
#include "LexiconFile.h"
#include "Word.h"
#include "HeapBag.h"
#include "XMap.h"
#include "kcnvlib.h"
#include "settemplate.h"
#include "Indexer.h"
#include "ConText.h"
#include "DescFile.h"

/*
Indexer::Indexer(WordBreaker &wb, const Unifier &unif)
    : breaker(wb), unifier(unif)
*/
Indexer::Indexer(WordBreaker &wb, const Unifier &unif, ConText *ctx)
    : breaker(wb), unifier(unif),
      Ctx(ctx), SeedDocs(0), SeedWords(0)
{
    wb.Ctx = ctx;
    NewDocs = 0;
    idescf = 0;
}

Indexer::Indexer(ConText *ctx)
    : breaker(*(new WordBreaker(*(new Patricia)))),
      unifier(*(new Unifier)),
      Ctx(ctx), SeedDocs(0), SeedWords(0)
{
    NewDocs = 0;
    idescf = 0;
}

Indexer::~Indexer(){
}

void
Indexer::printDebug(ostream &out) const{
    out << "-- Indexer --" << endl;
    out << "(no useful info)" << endl;
}


void
Indexer::deentitize(char *dst, const char *src){
    const char *cp = src;
    while(*cp){
        if(*cp == '&'){
            if(0 == strncmp(cp, "&gt;", 4)){
                *dst++ = '>';
                cp += 4;
            }else if(0 == strncmp(cp, "&lt;", 4)){
                *dst++ = '<';
                cp += 4;
            }else if(0 == strncmp(cp, "&amp;", 5)){
                *dst++ = '&';
                cp += 5;
            }else if(0 == strncmp(cp, "&nbsp;", 6)){
                *dst++ = ' ';
                cp += 6;
            }else if(0 == strncmp(cp, "&quot;", 6)){
                *dst++ = '"';
                cp += 6;
            }else if(0 == strncmp(cp, "&apos;", 6)){
                *dst++ = '\'';
                cp += 6;
            }else if(0 == strncmp(cp, "&#39;", 5)){
                *dst++ = '\'';
                cp += 5;
            }else{
                *dst++ = *cp++;
            }
        }else{
            *dst++ = *cp++;
        }
    }
    *dst = '\0';
}

void
/*
Indexer::output_mapentry(ostream &xmapfile, ElemCtx &ectx, int size, int nwords){
*/
Indexer::output_mapentry(Fstream &xmapfile, ElemCtx &ectx, int size, int nwords){
    XMapEntry entry;
/*
    entry.size = ectx.c_words;
*/
/*
    entry.last = nwords; // the last word in the document
*/
    entry.last = SeedWords + nwords; // the last word in the document

    entry.last = htonl(entry.last);
    xmapfile.write((char*)&entry,sizeof(entry));

    if(idescf){
        FileDescRecord irec;
        if(Fsize == 0 ){
            idescf->read((char*)&irec,sizeof(irec));
            Fsize = ntohl(irec.d_size);
            Fdate = ntohl(irec.d_date);
            Mtime = ntohl(irec.d_lastmod);
        }else{
            idescf->seekg(sizeof(irec),ios::cur);
        }
    }
    writeRecord(descf,Fsize,Fdate,Mtime,Crc32,Url,Author,Title,Digest);
}

void
Indexer::addMETA(const char *key, HashBag<valueSet<pos_t> > &bag,
                 ElemCtx &ectx, pos_t offset){
    int savtype = ectx.c_type;
    Word term(strlen(key), key);
    ectx.c_words++;
    ectx.c_type = TYPE_META;
/*
    bag[term] += toPos(ectx,offset);
*/
    pos_t pos;
    bag[term] += pos = toPos(ectx,offset);
/*
{
urlid_t did;
size_t off;
int type;
frompos(pos,did,off,type);
fprintf(stderr,"[did=%d off=%d typ=%d]\n",did,off,type);
}
*/
    if(Ctx) Ctx->CtxPutpos(term,ectx,offset);
    ectx.c_type = savtype;
}

int IndexSerno;
extern int WordCapitalNum;
extern int WordCapitalLen;
int VScapa;
int VSfill;
int VSmax;
int VScpy;
int HashLevel;

/*
int hashFull(HashBag < valueSet<pos_t> > &words,istream &source){
*/
int hashFull(HashBag < valueSet<pos_t> > &words,istream &source,int ccount){

	if( 128*1024*1024 < words.Valueslot()->valuestotal()){
		return 1;
	}

	unsigned int capa = words.capacity();
	unsigned int fill = words.size();
	unsigned int nacc = words.accessed();
	unsigned int coll = words.collision();

	if( 1 <= HashLevel ){
		if( capa/4 < fill && nacc*2 < coll )
		{
			/* don't enlarge when collsion is heavy ... */
			return 1;
		}
		if( capa*7/8 < fill ) /* low collision ratio */
		{
			if( MinBUFFERINGWORDS < MaxBUFFERINGWORDS )
			if( MinBUFFERINGWORDS < ccount/32 )
/*
			if( MinBUFFERINGWORDS < source.tellg()/32 )
*/
			{
				MinBUFFERINGWORDS *= 2;
				LapReport(0,"### expanded HashTable --> %d",
					MinBUFFERINGWORDS);
			}
			return 1;
		}
		return 0;
	}

	if( capa/128 < fill && nacc*4 < coll ){
		/* early dicision for heavy collision */
	}else
	if( capa/4   < fill && nacc*2 < coll ){
		/* usual dicision */
	}else
	if( capa/2   < fill && nacc/4 < coll ){
		/* collision might become heavy... */
	}else{
		return 0;
	}

	HashLevel++;
	LapReport(0,"### too often collisions %.2f%%, HashLelvel ---> %d",
		(coll*100.0)/nacc,HashLevel);
	return 2;
}
static void bidxReport(HashBag<valueSet<pos_t> > &words,istream &source,off_t soff,double Stime,double Stime0){
	char stat[128];

	LapReport(0,"found %d words ... build sub-index#%d ...",
		words.size(),IndexSerno-1);

	LapReport(0,"scanned %d bytes (%dK/s), TOTAL processed %lldM (%.2fM/s)",
		(int)((off_t)source.tellg()-soff),
		(int)(((off_t)source.tellg()-soff)/(Time()-Stime))/1000,
		/*
		(off_t)source.tellg()/1000000,
		*/
		(unsigned CODEINT)source.tellg()/1000000,
		(int)(((off_t)source.tellg())/(Time()-Stime0))/1000000.0
		);

	words.keystats(stat,sizeof(stat));
	LapReport(0,"... Lex %s",stat);
	words.Valueslot()->stats(stat,sizeof(stat));
	LapReport(0,"... Idx %s",stat);
}

int
/*
Indexer::processText(istream &source, LexiconFile &lexfile,
*/
Indexer::processText(istream &source, off_t &ccount, LexiconFile &lexfile,
                     IndexFile &indexfile, Fstream &xmapfile, pos_t &offset){
    // Ⱦѥľ'-'¿ʤɡˤƤޤΤۤ
    // Ĺʤ뤳Ȥ⤢
    char linebuf[MAXSOURCELINELENGTH + 1024];
    char unifiedline[MAXSOURCELINELENGTH + 512];
    static int type = TYPE_TEXTW3;
    static int docid = -1;
    static ElemCtx ectx = {0, TYPE_TEXTW3, 0, 0};
    static pos_t start_offset = offset;
    static int nwords = 0;
    static double Stime0 = 0;
    double Stime = Time();
    off_t soff = source.tellg();

    if( Stime0 == 0 )
         Stime0 = Time();

/*
    HashBag<valueSet<pos_t> > words(MaxBUFFERINGWORDS);
*/
    HashBag<valueSet<pos_t> > words(MinBUFFERINGWORDS);
    linebuf[MAXSOURCELINELENGTH - 1] = '\0';

    for(int lineno = 0 ; ; lineno++){
	if( lineno % 1000000 == 0 ){
		static int last,lt;
		int ct = time(0);
		int cur = (int)source.tellg();

		if( 0 < lineno )
		LapReport(0,"scanned %4d docs, %8d lines, %10lld bytes (%dK/s)",
			docid,lineno,(unsigned CODEINT)source.tellg(),
			/*
			docid,lineno,(off_t)source.tellg(),
			*/
			((last-cur)/(lt-ct))/1000);
		last = source.tellg();
		lt = ct;
	}
/*
	if( words.accessed() && lineno % 100000 == 0 ){
		words.stats(stat,sizeof(stat));
		LapReport(0,"%s",stat);
        }
*/

        source.getline(linebuf, MAXSOURCELINELENGTH, '\n');
	ccount += source.gcount();

        if(source.fail()){
            if(source.eof()) break;
            // linebufǥޥХʸǤ줿ν
            // cerr << "fail! line = \"" << linebuf << "\"" << endl;
            source.clear();
            // source.clear(ios::failbit); Τۤ?
            char *cp = linebuf;
            while(*cp){
                if(Iskanji1(*cp) || Ishkana1(*cp)){
                    if(*(cp + 1) == '\0'){
                        source.putback(*cp);
                        ccount--;
                        *cp = '\0';
                        break;
                    }
                    if(Iskanji2(*(cp + 1))) cp +=  2;
                    else if(Ishkana2(*(cp + 1))) cp += 2;
                    else {
                        PRINTERRMSG("Abnormal Kanji sequence found. ignore.");
                        *cp++ = ' ';
                        *cp++ = ' ';
                    }
                } else
                    cp++;
            }
        }
        // cout << "lineno = " << lineno << " line=\"" << linebuf;
        // cout << "\" size: " << words.size() << endl;
        if(linebuf[0] == '\0') continue;
        if(linebuf[0] == '<'){
            ectx.c_delch = 0;
            if(0 == strncasecmp(linebuf, "<DOC HREF=\"", 11)){
                start_offset = offset;
                //cout << "start_offset = " << start_offset << endl;
                char *url = linebuf + 11;
                char *q = strchr(url, '"');
                if(q) *q = '\0';
                deentitize(unifiedline, url);

                Strncpy(Url,url,sizeof(Url));
                Docpath[0] = 0;
                Author[0] = 0;
                Title[0] = 0;
                Digest[0] = 0;
                Fsize = 0;
                Fdate = 0;
                Mtime = 0;
                Crc32 = 0;

                ectx.c_docid = docid + 1;
                ectx.c_words = 0;

                addMETA(" begin",words,ectx,offset);

                ectx.c_type = TYPE_URL;
                (void)breaker.breakText(words, unifiedline,
                                        ectx, offset, WordBreaker::FORINDEX);
                if( opt_v )
                cout << '[' << unifiedline << flush;
                docid++;
                type = TYPE_TEXTW3;

    if((lineno%100) == 0)
    if(0 < LapTime())
    LapReport(1,"scanned %6d docs (%4d/s) %8d words (%6d/s)",
    docid,(int)(docid/LapTime()),nwords,(int)(nwords/LapTime()));

            }else if(0 == strncasecmp(linebuf, "<TEXT WEIGHT=\"", 14)){
                int weight = atoi(linebuf + 14);
                if(weight <= 0 || 3 < weight) weight = 3;
                switch (weight){
                case 3: type = TYPE_TEXTW3; break;
                case 2: type = TYPE_TEXTW2; break;
                case 1: type = TYPE_TEXTW1; break;
                default: break;
                }
            }else if(linebuf[0]=='<' && 0<(type=ElemId(linebuf+1,NULL," >"))){
            }else
            if(0 == strncasecmp(linebuf, "<DOCPATH>",9)){
                Strncpy(Docpath,linebuf+9,sizeof(Docpath));
                char *dp;
                if((dp = strchr(Docpath,'<')))
                    *dp = 0;
            }else
            if(0 == strncasecmp(linebuf, "<METAX ",7)){
                char *attr = linebuf+7;
                if(strncasecmp(attr,"name=SIZE>",10) == 0)
                     sscanf(attr+10,"%x",&Fsize);
                else
                if(strncasecmp(attr,"name=DATE>",10) == 0)
                     sscanf(attr+10,"%x",&Fdate);
                else
                if(strncasecmp(attr,"name=LASTMODIFIED>",18) == 0)
                     sscanf(attr+18,"%x",&Mtime);
                else
                if(strncasecmp(attr,"name=CRC32>",11) == 0)
                     sscanf(attr+11,"%x",&Crc32);

            }else if(0 == strncmp(linebuf, "</", 2)){
                if(0 == strncasecmp(linebuf, "</DOC", 5)){
                    nwords += ectx.c_words;
                    output_mapentry(xmapfile, ectx, offset-start_offset,nwords);
                    if( opt_v )
                    cout << ']'<< endl;
                }
                type = TYPE_TEXTW3;
                ectx.c_delch = 0;
                offset++;
            }
            continue;
        }

        if(noUnifier()){
        deentitize(linebuf, linebuf);
        }else{
        unifier.unify(unifiedline, linebuf);
        deentitize(linebuf, unifiedline);
        }

        switch(type){
          case TYPE_AUTHOR: Strncpy(Author,linebuf,sizeof(Author)); break;
          case TYPE_TITLE: Strncpy(Title,linebuf,sizeof(Title)); break;
          case TYPE_DIGEST: Strncpy(Digest,linebuf,sizeof(Digest)); break;
        }

        // ʬ
        ectx.c_type = type;
        (void)breaker.breakText(words, linebuf,
                                ectx, offset, WordBreaker::FORINDEX);

/*
        if(words.size() > MaxBUFFERINGWORDS){
            LapReport(0,"found %d words ... build index",words.size(),0,0,0);
*/
/*
        if( hashFull(words,source) ){
*/
        if( hashFull(words,source,ccount) ){
	    bidxReport(words,source,soff,Stime,Stime0);
            buildindex(lexfile, indexfile, words);

	    words.Valueslot()->newstats();
            LapReport(0,"... built sub-index#%d",IndexSerno-1);
            return 1;           // ޤ³
        }
    }
    if(words.size() > 0){
      source.clear();
      bidxReport(words,source,soff,Stime,Stime0);

      LapReport(0,"scanned %6d docs (%4d/s) %8d words (%6d/s).",
          docid+1,(int)(docid/LapTime()),
          nwords,(int)(nwords/LapTime()));
/*
      LapReport(0,"found %d words ... build index",words.size(),0,0,0);
        buildindex(lexfile, indexfile, words);
      LapReport(0,"scanned %d bytes.",offset,0,0,0);
      LapReport(0,"registered %d words occurrences (%d registered).",
          nwords,ectx.c_words_total,0,0);
*/
      LapReport(0,"found %d words ... build index",words.size());

        buildindex(lexfile, indexfile, words);

     /*
      LapReport(0,"scanned %lld bytes",(off_t)offset);
      LapReport(0,"registered %d words occurrences (%lld registered).",
          nwords,ectx.c_words_total);
     */
      LapReport(0,"scanned %lld bytes",(unsigned CODEINT)offset);
      LapReport(0,"registered %d words occurrences (%lld registered).",
          nwords,(unsigned CODEINT)ectx.c_words_total);
    }
    offset++;
    NewDocs = docid + 1;
    return 0;                   // Ϥ
}


void
Indexer::makefilename(char *lexfilename, char *idxfilename,
                      const char *basename, int nth){
    const char *cp = basename;
    char *dp = lexfilename;
    while((*dp = *cp++)) dp++;
    *dp++ = '.';
    *dp++ = '0' + nth / 100;
    *dp++ = '0' + (nth % 100) / 10;
    *dp++ = '0' + nth % 10;
    *dp = '\0';
    strcpy(idxfilename, lexfilename);
    strcat(idxfilename, ".idx");
    strcat(lexfilename, ".lex");
}

int
Indexer::makeIndices(istream &source, const char *basename,
                     Fstream &xmapfile, pos_t &offset){
    char lexfilename[PATH_MAX];
    char idxfilename[PATH_MAX];
    int no;
    off_t soff;
    double Stime0 = Time();
    double Stime;
    off_t ccount = 0; /* source.tellg() for stream input */

    xmapfile.clear();

    if(SeedDocs == 0 )
    xmapfile.seekp(0);

    IndexSerno = 0;
    for(no = 0; no <= 999; no++){
	IndexSerno++;
/*
this makes lower locality between Word and string at sorting
	if( stringBank )
		stringBank->reset();
	else	stringBank = new Strings;
*/
/*
LapReport(0,"--- --- creating idx#%d stringBank=%X",no,(int)stringBank);
*/
LapReport(0,"--- --- creating idx#%d stringBank=%X",no,(ArgInt)(PtrInt)stringBank);
/*
soff = source.tellg();
*/
soff = ccount;
Stime = Time();

        makefilename(lexfilename, idxfilename, basename, no);
        LexiconFile lexicon(lexfilename,"w");
        if(1){
            IndexFile indexfile(idxfilename, "w");
/*
            if(0 == processText(source, lexicon,indexfile,xmapfile, offset)){
*/
            if(0 == processText(source,ccount,lexicon,indexfile,xmapfile, offset)){
                no++;
                break;
            }
LapReport(0,"--- +++ %d bytes (%.1fK/s), TOTAL %lld (%.1fK/s)",
(int)(ccount-soff), ((ccount-soff)/(Time()-Stime))/1000,
(unsigned CODEINT)ccount, (ccount/(Time()-Stime0))/1000);
/*
ccount, (ccount/(Time()-Stime0))/1000);
*/
/*
(int)((off_t)source.tellg()-soff),
(((off_t)source.tellg()-soff)/(Time()-Stime))/1000,
(off_t)source.tellg(),
((off_t)source.tellg()/(Time()-Stime0))/1000
);
*/

        }else{
            PRINTERRMSG("Internal error: Cannot open files");
            cerr << "  lexfilename = \"" << lexfilename << "\"\n";
            cerr << "  idxfilename = \"" << idxfilename << "\"\n";
            return 0;
        }
    }
    IndexSerno = 0;
    return no;                  // ե(Ȥ)ο
}

int
Indexer::mergeIndices(const char *newbase, const char *tmpbase, int tmpNum,
                      IMask *mask, const char *seedbase, Files &files){
    assert(tmpNum > 0);
    int tmpnum = tmpNum;
    if(seedbase != NULL){
        tmpNum++;
    }

    if(mask == NULL)
    if(tmpNum == 1){            // Ĥξ̾ѹ
        char newlexfilename[PATH_MAX];
        char newidxfilename[PATH_MAX];
        char tmplexfilename[PATH_MAX];
        char tmpidxfilename[PATH_MAX];

        makefilename(tmplexfilename, tmpidxfilename, tmpbase, 0);
        strcpy(newlexfilename, newbase); strcat(newlexfilename, ".lex");
        strcpy(newidxfilename, newbase); strcat(newidxfilename, ".idx");

        if(-1 == Rename(tmplexfilename, newlexfilename)) PRINTERR();
        if(-1 == Rename(tmpidxfilename, newidxfilename)) PRINTERR();
        
        files.add("LexiconFile",".lex");
        files.add("IndexFile",".idx");
        return 0;
    }

    // tmpNumȤΥƥݥեreadonlyǰƤ˳
    int i;
/*
    char idxfilenames[tmpNum][PATH_MAX];
    char lexfilenames[tmpNum][PATH_MAX];
    LexiconFile *lexicons[tmpNum];
    IndexFile *indices[tmpNum];
*/
    typedef char Fname[PATH_MAX];
    typedef LexiconFile *LFP;
    typedef IndexFile *IFP;
    Fname *idxfilenames = new Fname[tmpNum];
    Fname *lexfilenames = new Fname[tmpNum];
    LexiconFile **lexicons = new LFP[tmpNum];
    IndexFile **indices = new IFP[tmpNum];

    int ni = 0;
    if(seedbase != NULL){
        sprintf(lexfilenames[ni],"%s.lex",seedbase);
        sprintf(idxfilenames[ni],"%s.idx",seedbase);
        ni++;
    }
    for(i = 0; i < tmpnum; i++){
        makefilename(lexfilenames[ni], idxfilenames[ni], tmpbase, i);
        ni++;
    }
    for(i = 0; i < tmpNum; i++){
        lexicons[i] = new LexiconFile(lexfilenames[i],"r");
        indices[i] = new IndexFile(idxfilenames[i], "r");
    }
/*
    for(i = 0; i < tmpNum; i++){
        makefilename(lexfilenames[i], idxfilenames[i], tmpbase, i);
        lexicons[i] = new LexiconFile(lexfilenames[i],"r");
        indices[i] = new IndexFile(idxfilenames[i], "r");
    }
*/

    // ե
    char newlexpath[PATH_MAX];
    char newidxpath[PATH_MAX];
    strcpy(newlexpath, newbase); strcat(newlexpath, ".lex");
    strcpy(newidxpath, newbase); strcat(newidxpath, ".idx");

    LexiconFile newlexicon(newlexpath,"w");
    IndexFile newindex(newidxpath, "w");

    pos_t *shifttbl = NULL;
    if(seedbase != NULL){
        shifttbl = new pos_t[tmpNum];
        shifttbl[0] = 0;
        pos_t shift = topos(SeedDocs,0,0);
        for(i = 1; i < tmpNum; i++){
            shifttbl[i] = shift;
        }
    }

    // ޡ
/*
    mergeFiles(&newlexicon, &newindex, tmpNum, lexicons, indices, NULL);
*/
    mergeFiles(&newlexicon, &newindex, tmpNum, lexicons, indices, files, mask, shifttbl);

    if(shifttbl)
        delete shifttbl;

    for(i = 0; i < tmpNum; i++){
        delete indices[i];
        delete lexicons[i];
        unlink(lexfilenames[i]);
        unlink(idxfilenames[i]);
    }
    delete []lexfilenames;
    delete []idxfilenames;
    delete []lexicons;
    delete []indices;

    return 0;
}


static LexiconFile *lex;
static IndexFile *idx;
static ConText *outCtx;
static FILE *outKfp;

static void storeEntry(int ko, void *wp, void *vp){
    if(outKfp != NULL){
	/*
        int off = lex->tellp();
        Fwrite(outKfp,(char*)&off,4);
	*/
	Lexoff_t off = lex->tellp();
        Fwrite(outKfp,(char*)&off,sizeof(off));
    }
    if(outCtx){
        outCtx->addkey(((Word *)wp)->toString(),ko);
    }
    LexiconEntry entry(*(Word*)wp);
    entry.store(*idx, *(valueSet<pos_t>*)vp, ((valueSet<pos_t>*)vp)->size());
    lex->appendEntry(entry);

    if(outCtx){
        outCtx->addpos(((Word*)wp)->kid);
    }
}
int
Indexer::buildindex(LexiconFile &lexfile, IndexFile &indexfile,
                    HashBag<valueSet<pos_t> > &hashbuf){
    if(outKfp != NULL ){
        fseek(outKfp,0,0);
    }else{
        outKfp = tmpfile();
    }
    if(Ctx){
        Ctx->startMap(hashbuf.size());
    }

    lex = &lexfile;
    idx = &indexfile;
    outCtx = Ctx;
    int keyNum = hashbuf.sortdump(storeEntry,0);

    if(Ctx){
        Ctx->doneMap();
    }
    if(outKfp != NULL){
        fflush(outKfp);
        int kfpsize = ftell(outKfp);
        fseek(outKfp,0,0);
        lexfile.appendKeymap(outKfp,kfpsize);
        fclose(outKfp);
        outKfp = 0;
    }
    return keyNum;
}

typedef struct {
    int ix;
    int nx;      // serial number of the Lexcion in the file
    LexiconFile *lex;
    IndexFile *idx;
    Fstream *lexstreamp;
    LexiconEntry *entry;
    pos_t *list;
    int listLen;
/*
    pos_t shift; // non-zero when merging existing files
    IMask *mask; // bit-map of docid's to be purged
*/
    ShiftMask shiftmaskb;
    ShiftMask *shiftmask;
} LexEnt;
//
// What ShiftMask does:
//   local docid in each partial index is shifted to be global-docid
//   then the global-docid is removed(-1) or mapped to another docid
//
int fetch1(LexEnt *le){
    *le->lexstreamp >> *le->entry;
    if(le->entry->eof()){
        le->list = 0;
        le->listLen = 0;
        return -1;
    }else{
        le->nx++;
/*
        le->list = le->entry->fetch(*le->idx, le->listLen);
*/
/*
        le->list = le->entry->fetch(*le->idx, le->listLen, le->mask);
*/
        le->list = le->entry->fetch(*le->idx, le->listLen, le->shiftmask);
/*
        if(le->shift != 0){
            for(int j = 0; j < le->listLen; j++)
                le->list[j] += le->shift;
        }
*/
        return 0;
    }
}
static int compLE(LexEnt **a, LexEnt **b){
     LexiconEntry *la = (*a)->entry;
     LexiconEntry *lb = (*b)->entry;
     if(lb->eof()) return -1;
     if(la->eof()) return 1;
     if(int diff = strcmp(la->getkey(),lb->getkey()))
         return diff;
     return (*a)->ix - (*b)->ix;
}

class Stat {
    private:
        enum { MAXSAMPLE = 10 };
        int lens[MAXSAMPLE];
        int lenrefs[MAXSAMPLE];
        int refs[MAXSAMPLE];
        int reflens[MAXSAMPLE];
        int types[32];
    public:
        Stat();
        void sample(int klen, int refs);
        void pos(pos_t *list, int listLen);
        void print();
};
Stat::Stat(){
    int i;
    for(i = 0; i < MAXSAMPLE; i++) lens[i] = 0;
    for(i = 0; i < MAXSAMPLE; i++) lenrefs[i] = 0;
    for(i = 0; i < MAXSAMPLE; i++) refs[i] = 0;
    for(i = 0; i < MAXSAMPLE; i++) reflens[i] = 0;
    for(i = 0; i < 32; i++) types[i] = 0;
}
void
Stat::sample(int klen, int nref){
    int kl = klen - 1;
    int rs = (nref/100);
    if(MAXSAMPLE <= kl) kl = MAXSAMPLE-1;
    if(MAXSAMPLE <= rs) rs = MAXSAMPLE-1;
    lens[kl] += 1;
    lenrefs[kl] += klen*nref;
    refs[rs] += 1;
    reflens[rs] += klen*nref;
}
void
Stat::pos(pos_t *list, int listLen){
    for(int i = 0; i < listLen; i++){
        int type = pos2type(list[i]);
        types[type] += 1;
    }
}
void
Stat::print(){
    int i;
    int lenb = 0;
    for(i = 0; i < MAXSAMPLE; i++) lenb += lenrefs[i];
    fprintf(stderr,"wordLength   count  sumBytes ratio\n");
    for(i = 0; i < MAXSAMPLE; i++){
       if(lens[i])
          fprintf(stderr,"%10d %7d %9d %4.1f%%\n",i+1,lens[i],
              lenrefs[i],(lenrefs[i]*100.0)/lenb);
    }

    int refb = 0;
    for(i = 0; i < MAXSAMPLE; i++) refb += reflens[i];
    fprintf(stderr,"occurences   count  sumBytes ratio\n");
    for(i = 0; i < MAXSAMPLE; i++){
       if(refs[i])
          fprintf(stderr,"%10d %7d %9d %4.1f%%\n",(i+1)*100,refs[i],
              reflens[i],(reflens[i]*100.0)/lenb);
    }

    fprintf(stderr,"type   count\n");
    for(i = 0; i < 32; i++){
       if(types[i]){
           fprintf(stderr,"%4d %7d\n",i,types[i]);
       }
    }
}

int
Indexer::mergeFiles(LexiconFile *newlexicon, IndexFile *newindex, int n,
                    LexiconFile *lexicons[], IndexFile *indices[],
                    Files &files,
                    IMask *mask, pos_t *shifttbl)
/*
                    pos_t *shifttbl)
*/
{
    Stat *stat = 0;
    if(opt_S)
        stat = new Stat;
    int nwords = 0;
/*
    int nbytes = 0;
    int npos = 0;
*/
    off_t nbytes = 0;
    off_t npos = 0;

    newlexicon->prepareAppend();
    FILE *kfp = tmpfile();
    int i;

/*
    LexEnt ilex[n];
    LexEnt *sortedlex[n];
*/
    typedef LexEnt *LEP;
    LexEnt *ilex = new LexEnt[n];
    LexEnt **sortedlex = new LEP[n];

    for(i = 0; i < n; i++){
        ilex[i].ix = i;
        ilex[i].nx = 0;
        ilex[i].lex = lexicons[i];
        ilex[i].shiftmaskb.shift = 0;
        ilex[i].shiftmaskb.mask = 0;
        if(shifttbl){                // merging existing indices
          if(!Ctx->isnewidx(i)){     // this is not a new index
            if(Ctx != NULL){         // with ConText
              int lsize = lexicons[i]->nentry();
              Ctx->newMap(i,lsize);  // map from local lex# to merged lex#
            }
          }
        }
/*
fprintf(stderr,"#### DO SHIFT/MAP for new too: %d\n",Ctx->isnewidx(i));
*/
        ilex[i].shiftmaskb.mask = mask;

        ilex[i].idx = indices[i];
        ilex[i].lexstreamp = &(lexicons[i]->rewind());
/*
        ilex[i].shift = shifttbl ? shifttbl[i] : 0;
*/
        ilex[i].shiftmaskb.shift = shifttbl? pos2did(shifttbl[i]) : 0;
        if(ilex[i].shiftmaskb.shift || ilex[i].shiftmaskb.mask)
            ilex[i].shiftmask = &ilex[i].shiftmaskb;
        else
            ilex[i].shiftmask = 0;
/*
fprintf(stderr,"#### SHIFT=%d MASK=%X SHIFTMASK=%X\n",
ilex[i].shiftmaskb.shift,
(int)ilex[i].shiftmaskb.mask,(int)ilex[i].shiftmask);
*/

        ilex[i].entry = new LexiconEntry;
        sortedlex[i] = &ilex[i];
        fetch1(&ilex[i]);
    }
    qsort(sortedlex,n,sizeof(LexiconEntry*),(sortfunc)compLE);

    LapReport(0,"... merge %d indices ...",n,0,0,0);
    int actNum = n;
    typedef pos_t *POSP;
    pos_t **lists = new POSP[n];
    int *listLens = new int[n];

    while(0 < actNum){
        if(nwords)
        if((nwords%100) == 0)
/*
        LapReport(1,"merged %7d words %8d occurrences, %8d bytes",
            nwords,npos,nbytes,0);
*/
	LapReport(1,"merged %7d words %8lld occurrences, %8lld bytes",
            nwords,(unsigned CODEINT)npos,(unsigned CODEINT)nbytes);
/*
            nwords,npos,nbytes);
*/

        int minNum = 0;
        const char *key0 = 0;
        for(minNum = 0; minNum < actNum; minNum++){
            LexiconEntry *ep = sortedlex[minNum]->entry;
            if( ep->eof() )
                break;
            if(minNum == 0){
                key0 = sortedlex[0]->entry->getkey();
            }else{
                if(strcmp(key0,ep->getkey()) != 0)
                    break;
            }
        }
        if(minNum == 0)
            break;

        LexiconEntry newentry(key0);
        int totalLen = 0;

        if(kfp != NULL){
/*
            int lexoff = newlexicon->tellp();
            Fwrite(kfp,(char*)&lexoff,4);
*/
            Lexoff_t lexoff = newlexicon->tellp();
            Fwrite(kfp,(char*)&lexoff,sizeof(lexoff));
        }
        if(Ctx != NULL){
            for(i = 0; i < minNum; i++){
                Ctx->addKeyX(key0,sortedlex[i]->ix,sortedlex[i]->nx-1,nwords);
            }
        }
        if(minNum == 1){
            pos_t *list = sortedlex[0]->list;
            int listLen = sortedlex[0]->listLen;
          if(listLen == 0){
          }else{
            newentry.store(*newindex,list,listLen);
            newlexicon->appendEntry(newentry);
            totalLen = listLen;
            if(opt_S){
                stat->pos(list,listLen);
            }
          }
            delete []list;
        }else{
            for(i = 0; i < minNum; i++){
                lists[i] = sortedlex[i]->list;
                listLens[i] = sortedlex[i]->listLen;
                totalLen += sortedlex[i]->listLen;
            }
          if( 1024*1024*1024 < totalLen*sizeof(pos_t) ){
fprintf(stderr,"--- list is too large to merge ... %d [%s]\n",
totalLen,key0);
            totalLen = 1024*1024;
          }
          if(totalLen == 0 ){
          }else
          if(0x10000 < totalLen){
            pos_t *merged = new pos_t[totalLen];
            int mergedLen = mergeLists(merged,totalLen,lists,listLens,minNum);
            newentry.store(*newindex, merged, mergedLen);
            newlexicon->appendEntry(newentry);
            if(opt_S){
                stat->pos(merged,mergedLen);
            }
            delete []merged;
          }else{
            pos_t merged[0x10000];
            int mergedLen = mergeLists(merged,totalLen,lists,listLens,minNum);
            newentry.store(*newindex, merged, mergedLen);
            newlexicon->appendEntry(newentry);
            if(opt_S){
                stat->pos(merged,mergedLen);
            }
          }
            for(i = 0; i < minNum; i++)
                delete []lists[i];
        }
        nwords++;
        npos += totalLen;
        int klen = strlen(key0);
        nbytes += totalLen * klen;
        if(opt_S)
            stat->sample(klen,totalLen);
/*
        if(1){
            printf("%d %s\n",totalLen,key0);
        }
*/

        int closed = 0;
        for(i = 0; i < minNum; i++){
            if( fetch1(sortedlex[i]) != 0 )
                closed++;
        }
        {
          int i; // the entry to be inserted
          for(i = minNum-1; 0 <= i; i--){
            int j; // the entry larger than i
            for(j = i+1; j < actNum; j++){
                if(compLE(&sortedlex[i],&sortedlex[j]) <= 0){
                    break;
                }
            }
            if(i+1 < j){
              LexEnt *sav = sortedlex[i];
                int k;
              for(k = i; k < j-1; k++)
                sortedlex[k] = sortedlex[k+1];
              sortedlex[k] = sav;
            }
          }
        }
        actNum -= closed;
    }
    
/*
    LapReport(0,"merged %7d words %8d occurrences, %8d bytes.",
        nwords,npos,nbytes,0);
*/
    LapReport(0,"merged %7d words %8lld occurrences, %8lld bytes.",
        nwords,(unsigned CODEINT)npos,(unsigned CODEINT)nbytes);
/*
        nwords,npos,nbytes);
*/
    if(opt_S)
        stat->print();

    fflush(kfp);
    int kfpsize = ftell(kfp);
    fseek(kfp,0,0);
    newlexicon->appendKeymap(kfp,kfpsize);
    fclose(kfp);

    files.add("LexiconFile",".lex");
    files.add("IndexFile",".idx");

    delete []ilex;
    delete []sortedlex;
    delete []lists;
    delete []listLens;

    return nwords;
}

void createSortFiles(const char *base, Files &files){
    char descname[PATH_MAX];
    strcpy(descname,base);
    strcat(descname,".dsc");
    Fstream desc(descname,"r");

    if(desc.good()){
        LapReport(0,"creating sort files...",0,0,0,0);

        char path[PATH_MAX];
        strcpy(path,base); strcat(path,".sum");
        files.add("SumFile",".sum");
        Fstream sum(path,"w");

        int nd = 0;
        if(sum.good()){
            desc.clear();
            desc.seekg(0,ios::beg);
            nd = makesumfiles(desc,sum);
        }
        fprintf(stderr,"%d Documents.\n",nd);

        desc.clear();
        desc.seekg(0,ios::beg);
        int len;
        int *usort = makesortUrl(desc,len);

        char crcpath[PATH_MAX];
        sprintf(crcpath,"%s%s",base,NUM_EXT);
        NumFile numf(crcpath,nd);
        desc.clear();
        desc.seekg(0,ios::beg);
        for(int i = 0;;i++){
            FileDescRecord rec;
            desc.read((char*)&rec,sizeof(rec));
            if(desc.gcount() != sizeof(rec))
                break;
            numf.write(i,rec,usort[i]);
        }
        files.add("NumFile",NUM_EXT);
/*
numf.flush();
NumFile crcr(crcpath);
if(crcr.good())
for(int i = 0; i < nd; i++)
fprintf(stderr,"[%2d] %2d %8d %08X %08X %08X\n",i,crcr.get(i,N_USORT),
crcr.get(i,N_SIZE),
crcr.get(i,N_CONTCRC),
crcr.get(i,N_URLCRC),
crcr.get(i,N_LASTMOD)); 
*/
    }
}

void
Files::renames(){
    char tpath[PATH_MAX];
    char npath[PATH_MAX];
    char spath[PATH_MAX];

/*
    int total = 0;
*/
    off_t total = 0;
    for(int i = 0; i < num(); i++){
        sprintf(tpath,"%s#%s",base(),ext(i));
        sprintf(npath,"%s%s",base(),ext(i));
        sprintf(spath,"%s%s~",base(),ext(i));
        Rename(npath,spath);
        Rename(tpath,npath);
        total += FileSize(npath);
/*
        fprintf(stderr,"%-16s %8d %s\n",desc(i),FileSize(npath),npath);
*/
	/*
        fprintf(stderr,"%-12s %10lld %s\n",desc(i),(unsigned CODEINT)FileSize(npath),npath);
	*/
        LapReport(0,"%^%-12s %10lld %s",desc(i),(unsigned CODEINT)FileSize(npath),npath);
    }
/*
    fprintf(stderr,"%-16s %8d %d files\n","total",total,num());
*/
/*
    fprintf(stderr,"%-12s %10lld %d files\n","total",(unsigned CODEINT)total,num());
*/
    LapReport(0,"%^%-12s %10lld %d files","total",(unsigned CODEINT)total,num());
}

typedef struct {
    int num;
    int val;
    int size;
} NumVal;
static int cmpnv(NumVal *a, NumVal *b){
    int diff;
    if((diff = ((NumVal*)a)->val - ((NumVal*)b)->val) != 0)
         return diff;
    return ((NumVal*)b)->num - ((NumVal*)a)->num;
}

static int dbgdel = 0;
int AllowDuplicated = 0;

IMask *
Indexer::toberemoved(const char *base, int ndocs){
    char descname[PATH_MAX];
    strcpy(descname,base);
    strcat(descname,".dsc");
    Fstream desc(descname,"r");

/*
    NumVal crc32[ndocs];
*/
    NumVal *crc32 = new NumVal[ndocs];

    int ndel = 0;
    if(desc.good()){
        int i;
        for(i = 0;;i++){
            FileDescRecord frec;
            desc.read((char*)&frec,sizeof(frec));
            if(desc.gcount() != sizeof(frec))
                break;
            crc32[i].num = i;
            crc32[i].val = frec.d_crc32;
            crc32[i].size = frec.d_size;
if(dbgdel)
fprintf(stderr,"--- %2d/%2d/%2d %X\n",i,SeedDocs,ndocs,frec.d_crc32);
        }
    }
    qsort(crc32,ndocs,sizeof(NumVal),(int(*)(const void*,const void*))cmpnv);

    Mask bmask(ndocs);
if(!AllowDuplicated){
    int prev = crc32[0].val;
    for(int i = 1; i < ndocs; i++){
        int cur = crc32[i].val;

        if(cur != 0 && cur == prev && crc32[i-1].size == crc32[i].size){
fprintf(stderr,"ignore dup. doc. %5d %5d CRC=%08X size=[%6d %6d]\n",
crc32[i-1].num,crc32[i].num,cur,crc32[i-1].size,crc32[i].size);
            bmask.set(crc32[i].num);
            ndel++;
if(dbgdel)
fprintf(stderr,"## %d/%d del %d\n",crc32[i].num,SeedDocs,ndel);
        }else{
            prev = cur;
        }
    }
}
    if(ndel){
        fprintf(stderr,"%d duplication removed.\n",ndel);
    }
    delete []crc32;

    IMask *mask = 0;
    if(ndel){
        if(bmask.num_set()){
            mask = new IMask(ndocs);
            int docid = 0;
            for(int mi = 0; mi < ndocs; mi++){
               if(bmask.is_set(mi)){
                   mask->set(mi,-1);
               }else{
                   mask->set(mi,docid++);
               }
if(dbgdel)
fprintf(stderr,"## MAP[%2d/%2d] %2d\n",mi,SeedDocs,mask->get(mi));
            }
        }
    }

    char xmapname[PATH_MAX];
    strcpy(xmapname,base);
    strcat(xmapname,".map");
    Fstream xmap(xmapname,"r+");
    if(xmap.good()){
        int nwords = 0;
        int olddoc = 0;
        int newdoc = 0;
        int last = 0;

/*
        XMapEntry newmap[ndocs-ndel];
*/
        XMapEntry *newmap = new XMapEntry[ndocs-ndel];
        XMapEntry entry;
        for(int di = 0; di < ndocs; di++){
            xmap.read((char*)&entry, sizeof(XMapEntry));
            int size = entry.last - last;
            last = entry.last;
            if(!bmask.is_set(olddoc)){
                nwords += size;
                entry.last = nwords;
                newmap[newdoc++] = entry;
            }
            olddoc++;
        }
	xmap.clear();
        xmap.seekp(0);
        xmap.write((char*)newmap,newdoc*sizeof(XMapEntry));
        xmap.truncate();
        xmap.close();
        delete []newmap;
    }

    Fstream newdesc(descname,"r+");
    if(newdesc.good()){
        desc.clear();
        desc.seekg(0,ios::beg);
        for(int di = 0;;di++){
            FileDescRecord frec;
            desc.read((char*)&frec,sizeof(frec));
            if(desc.gcount() != sizeof(frec))
                break;
            if(!bmask.is_set(di)){
                newdesc.write((char*)&frec,sizeof(frec));
            }
        }
        newdesc.truncate();
        newdesc.close();
    }

    if(desc.good())
        desc.close();
    return mask;
}
