//
// $Id: Label.cc,v 0.9 1998-01-10 01:43:43+09 harada Exp harada $
//
#include "SX.h"
#include "Result.h"
#include "Label.h"

Label::Label(int fildsc, int infofildsc)
    : lbl_fd(fildsc), lbi_fd(infofildsc),
      mapaddr(NULL), infoaddr(NULL), labelInfo(NULL)
{
    struct stat lbl_st, lbi_st;
    errno = 0;
    fstat(lbl_fd, &lbl_st);
    lbl_filesize = lbl_st.st_size;
    recordNum = lbl_filesize / (sizeof(labelid_t) * MAXLABELS);
    mapaddr = mmap(NULL,        // start address (hint)
                   lbl_filesize, // size
                   PROT_READ,   // protection
                   MAP_SHARED,  // flags
                   lbl_fd,      // file descriptor
                   0 );         // file offset

    fstat(lbi_fd, &lbi_st);
    lbi_filesize = lbi_st.st_size;
    infoaddr = mmap(NULL,
                    lbi_filesize,
                    PROT_READ,
                    MAP_SHARED,
                    lbi_fd,
                    0 );
    if(errno) PRINTERRMSG("cannot initialize Label Object");

    labelInfo = new (struct label_info)[65536];
    memset(labelInfo, 0, sizeof(label_info) * 65536);
    const char *cp = infoaddr;
    for(int i = 0; i < 65536 && (size_t)(cp - infoaddr) < lbi_filesize; i++){
        labelInfo[i].id = i + 1;
        memcpy(&(labelInfo[i].usedNum), cp, sizeof(int));
        cp += sizeof(int);
        labelInfo[i].name = cp;
        while(*cp++)
            ;
    }
}

void
Label::resetInfo(){
    for(int i = 0; i < 65536; i++){
        labelInfo[i].count = 0;
        labelInfo[i].score = 0;
    }
}

Label::~Label(){
    if(labelInfo) delete []labelInfo;
    if(mapaddr && -1 == munmap(mapaddr, lbl_filesize)){
        PRINTERRMSG("*.lbl munmap(2) failed.");
    }
    if(-1 == close(lbl_fd))
        PRINTERRMSG("*.lbl close(2) failed.");
    if(mapaddr && -1 == munmap(infoaddr, lbi_filesize)){
        PRINTERRMSG("*.lbi munmap(2) failed.");
    }
    if(-1 == close(lbi_fd))
        PRINTERRMSG("*.lbi close(2) failed.");
}


int
Label::countLabels(struct label_info infotbl[],
                   const Result *results, int resultsNum){
    int foundlabel = 0;
    for(int r = 0; r < resultsNum; r++){
        const labelid_t * const labels = getLabels(results[r].docid);
        for(int i = 0;
            labels[i] && i < (int)(MAXLABELS / sizeof(labels[0])); i++){
            if(1 == ++infotbl[labels[i] - 1].count){
                foundlabel++;
            }
            infotbl[labels[i] - 1].score += results[r].weight;
        }
    }
    return foundlabel;
}

int
Label::countLabels2(struct label_info infotbl[],
                    const Result *results, int resultsNum){
    int foundlabel = 0;
    for(int r = 0; r < resultsNum; r++){
        const labelid_t * const labels = getLabels(results[r].docid);
        int labelNum = 0;
        int i;
        for(i = 0; labels[i] && i < (int)(MAXLABELS / sizeof(labels[0])); i++){
            labelNum++;
        }
        for(i = 0; i < labelNum; i++){
            if(1 == ++infotbl[labels[i] - 1].count){
                foundlabel++;
            }
            infotbl[labels[i] - 1].score += results[r].weight / labelNum;
        }
    }
    return foundlabel;
}

static int
compare_count(const Label::label_info *a, const Label::label_info *b){
    return b->count - a->count;
}

// ˡ1: ¿ޤޤƤ٥ͥ褹
int
Label::recommend1(const Result *results, int resultsNum){
    resetInfo();
    int labelNum = countLabels(labelInfo, results, resultsNum);
    qsort(labelInfo, 65536, sizeof(label_info),
          (int (*)(const void *, const void *))compare_count);
    for(int i = 0; i < labelNum && i < MAXDISPLAYLABELS; i++){
        cout << "label ID." << labelInfo[i].id << " appears "
             << labelInfo[i].count << " times. ("
             << labelInfo[i].name << ' '
             << labelInfo[i].usedNum << ")\n";
    }

    return 0;
}

static int
compare_score(const Label::label_info *a, const Label::label_info *b){
    if(b->score == a->score) return 0;
    return (b->score > a->score) ? 1 : -1;
}

// ˡ2: ιפ¤ؤ
int
Label::recommend2(const Result *results, int resultsNum){
    resetInfo();
    int labelNum = countLabels(labelInfo, results, resultsNum);
    qsort(labelInfo, 65536, sizeof(label_info),
          (int (*)(const void *, const void *))compare_score);
    for(int i = 0; i < labelNum && i < MAXDISPLAYLABELS; i++){
        cout << labelInfo[i].name << '('
             << labelInfo[i].count << " / "
             << labelInfo[i].score / resultsNum << "pt)\n";
    }

    return 0;
}


// ˡ3: ˡ1٥λѲ
int
Label::recommend3(const Result *results, int resultsNum){
    resetInfo();
    int labelNum = countLabels(labelInfo, results, resultsNum);
    int i;
    for(i = 0; i < 65536; i++)
        labelInfo[i].score
            = (double)labelInfo[i].count / (log(labelInfo[i].usedNum + 1) + 1);
    qsort(labelInfo, 65536, sizeof(label_info),
          (int (*)(const void *, const void *))compare_score);
    for(i = 0; i < labelNum && i < MAXDISPLAYLABELS; i++){
        cout << "label ID." << labelInfo[i].id << " "
             << labelInfo[i].count << " times. "
             << labelInfo[i].score << " pt. ("
             << labelInfo[i].name << ' '
             << labelInfo[i].usedNum << ")\n";
    }

    return 0;
}


// ˡ4: ˡ2٥λѲ
int
Label::recommend4(const Result *results, int resultsNum){
    resetInfo();
    int labelNum = countLabels(labelInfo, results, resultsNum);
    int i;
    for(i = 0; i < 65536; i++){
            labelInfo[i].score /= labelInfo[i].usedNum + 1000;
    }
    qsort(labelInfo, 65536, sizeof(label_info),
          (int (*)(const void *, const void *))compare_score);
    for(i = 0; i < labelNum && i < MAXDISPLAYLABELS; i++){
        cout << "label ID." << labelInfo[i].id << " "
             << labelInfo[i].count << " times. "
             << labelInfo[i].score << " pt. ("
             << labelInfo[i].name << ' '
             << labelInfo[i].usedNum << ")\n";
    }

    return 0;
}

// ˡ5: ˡ21URLΥ٥ο
int
Label::recommend5(const Result *results, int resultsNum){
    resetInfo();
    int labelNum = countLabels2(labelInfo, results, resultsNum);
    qsort(labelInfo, 65536, sizeof(label_info),
          (int (*)(const void *, const void *))compare_score);
    for(int i = 0; i < labelNum && i < MAXDISPLAYLABELS; i++){
        cout << labelInfo[i].name << '('
             << labelInfo[i].count << " / "
             << labelInfo[i].score / resultsNum << "pt)\n";
    }

    return 0;
}

// ˡ6: ˡ5򤢤Υ٥λѲ
int
Label::recommend6(const Result *results, int resultsNum){
    resetInfo();
    int labelNum = countLabels2(labelInfo, results, resultsNum);
    int i;
    for(i = 0; i < 65536; i++){
        labelInfo[i].score /= labelInfo[i].usedNum + 1;
    }
    qsort(labelInfo, 65536, sizeof(label_info),
          (int (*)(const void *, const void *))compare_score);
    for(i = 0; i < labelNum && i < MAXDISPLAYLABELS; i++){
        cout << "label ID." << labelInfo[i].id << " "
             << labelInfo[i].count << " times. "
             << labelInfo[i].score << " pt. ("
             << labelInfo[i].name << ' '
             << labelInfo[i].usedNum << ")\n";
    }

    return 0;
}



#ifdef LABEL_DEBUG
int main(){
    int fd = open("TEST.lbl", O_RDONLY, 0666);
    if(fd == -1) PRINTERRMSG("cannot open label file");
    Label label(fd);
    const labelid_t *labels = label.getLabels(0);
    cout << "found: " << flush;
    while(*labels > 0){
        cout << *labels << ',';
        labels++;
    }
    cout << endl;

    return 0;
}
#endif
