// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: LexiconEntry.cc,v 0.99 2004-08-10 y.sato Exp $
//   Data Representation in the Lexicon File of FreyaSX
//   [0]          keyLength(K=0..63) [|0x80] 
//   [1..K]       key (without terminator)
//   [K+1..K+1+N] listLen (compressed)
//   [K+1+N+1..]
//    with [0]0x80 // LEXLOCALINDEX
//                pos_t_1 pos_t_2 ... (compressed)
//    otherwise
//                wordid (compressed) ... points to pos_t lists in Index File
//
//
#include "SX.h"
#include "settemplate.h"
#include "LexiconEntry.h"
#include "IndexFile.h"



void
LexiconEntry::printDebug(ostream &out) const{
}
 LexiconEntry::LexiconEntry(){
   key[0] = 0;
   valueNum = 0;
   listSize = 0;
   _eof = 0;
   wid = ILL_WID;
   lid = 0;
 }
 LexiconEntry::LexiconEntry(const char *keystr){
   strcpy(key, keystr);
   valueNum = 0;
   listSize = 0;
   _eof = 0;
   wid = ILL_WID;
   lid = 0;
 }
 LexiconEntry::~LexiconEntry(){
 }
 LexiconEntry &
 LexiconEntry::store(IndexFile &indexfile, const pos_t *list, int listLen){
   listSize = listLen;
   if(listLen <= MaxLEXLOCALINDEX){
     valueNum = listLen;
     memcpy(values,list,sizeof(pos_t)*listLen);
   }else{
     valueNum = 0;
     indexfile.store(list, listLen, wid);
   }
   return *this;
 }
/*
 pos_t *LexiconEntry::fetch(IndexFile &indexfile, int &listLen) const{
*/
/*
 pos_t *LexiconEntry::fetch(IndexFile &indexfile, int &listLen, IMask *mask) const{
*/
 pos_t *LexiconEntry::fetch(IndexFile &indexfile, int &listLen, ShiftMask *mask) const{
   pos_t *list;
   if(0 < valueNum){
      listLen = valueNum;
      list = new pos_t[listLen];
      memcpy(list,values,sizeof(list[0])*listLen);
      if(mask != NULL){
          listLen = maskList(list,listLen,mask);
      }
      return list;
   }
   if(wid == ILL_WID){
      PRINTERRMSG("[SX]ILLEGAL ENTRY FETCHED");
      listLen = 0;
/*
      return 0;
*/
      list = new pos_t[1];
      return list;
   }
   listLen = 0;
/*
   list = indexfile.fetch(wid, listLen);
*/
   list = indexfile.fetch(wid, listLen, mask);
   return list;
 }
/*
 ostream &operator<<(ostream &stream, const LexiconEntry &entry){
*/
 Fstream &operator<<(Fstream &stream, const LexiconEntry &entry){
/*
int off = stream.tellp();
*/
   unsigned char tag0,klen;
   klen = strlen(entry.key);
   if(MaxKEYLENGTH <= klen){
       PRINTERRMSG("[SX]too long key\n");
       klen = MaxKEYLENGTH;
   }
   tag0 = klen;
   if(0 < entry.valueNum)
       tag0 |= 0x80;
/*
   stream << tag0;
   stream << entry.key;
*/
   stream.write((char*)&tag0,1);
   stream.write((char*)entry.key,strlen(entry.key));

   pos_t val = entry.listSize;
   stream.Fputintvl(&val);
/*
   fputintvl(stream.file(),&val);
*/

//fprintf(stdout,"## %X %d/%4d PUT[%s]\n",off,blen,entry.listSize,entry.key);

   if(0 < entry.valueNum){
       for(int i = 0; i < entry.listSize; i++){
           val = EnPACK(entry.values[i]);
           stream.Fputintvl(&val);
/*
           fputintvl(stream.file(),&val);
*/
       }
   }else{
       pos_t wwid = (pos_t)entry.wid;
       stream.Fputintvl(&wwid);
/*
       fputintvl(stream.file(),&wwid);
*/
   }
   return stream;
 }
/*
 istream &operator>>(istream &stream, LexiconEntry &entry){
*/
 Fstream &operator>>(Fstream &stream, LexiconEntry &entry){
/*
int off = stream.tellg();
*/
   entry.valueNum = 0;
   entry.wid = ILL_WID;
   entry._eof = 0;

   unsigned char tag0 = 0xFF;
   unsigned char klen = 0xFF;

   // stream >> tag0; // bad, but why?
   stream.read((char*)&tag0,1);
   if(stream.eof()){
       LexiconEntry nullentry;
       entry = nullentry;
       entry._eof = 1;
       return stream;
   }
   klen = tag0 & 0x3F;

   int ki = 0;
   for(ki = 0; ki < klen; ki++){
       stream.read(&entry.key[ki],1);
       if(entry.key[ki] == 0){
           PRINTERRMSG("[SX]got zero value in key-string");
           exit(0);
       }
       if(stream.eof()){
           PRINTERRMSG("[SX]got EOF during reading lexicon");
       }
   }
   entry.key[klen] = 0;
   unsigned CODEINT lsiz;
   // int blen =
   fgetintvl(stream.file(),&lsiz);
   entry.listSize = lsiz;

   if(tag0 == 0 && lsiz == 0){
/*
fprintf(stderr,"#### SOFT EOF ####\n");
*/
       LexiconEntry nullentry;
       entry = nullentry;
       entry._eof = 1;
       return stream;
   }
/*
fprintf(stderr,"##%X %d/%4d GET[%s]\n",off,blen,(int)lsiz,entry.key);
*/

   entry.valueNum = 0;
   entry.wid = ILL_WID;
   if( tag0 & 0x80 ){
       unsigned char vnum = lsiz;
       if(MAXLEXLOCALINDEX < vnum){
          PRINTERRMSG("[SX]got too large LEXLOCALINDEX");
       }
       int i;
       for(i = 0; i < vnum; i++){
           pos_t val;
           int len = fgetintvl(stream.file(),&val);
           if(len <= 0){
               PRINTERRMSG("[SX]EOF getting LLI");
               break;
           }
           entry.values[i] = DePACK(val);
       }
       if(i == vnum){
           entry.valueNum = vnum;
       }
   }else{
       pos_t wwid;
       int len = fgetintvl(stream.file(),&wwid);
       if(len <= 0){
           PRINTERRMSG("[SX]EOF getting WID");
       }else{
           entry.wid = wwid;
       }
       /* the real values will be loaded at fetch() */
   }
   if( entry.valueNum != 0 || entry.wid != ILL_WID )
   /* is valid entry */
   if( stream.eof() ){
       // Not REAL EOF, for MacOS X
       stream.clear();
   }
   return stream;
 }
