int lexb;
// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: LexiconFile.cc,v 0.10 1998-02-17 22:23:46+09 harada Exp $
//

// TODO: offsetTbl  Coder ưפ˰̤Ǥ

#include "SX.h"
#include "jlib.h"
#include "LexiconFile.h"

char LexiconFile::MAGIC[] = "_FREYALEX_996_\n"; // 16byte(including '\0')

LexiconFile::LexiconFile(const char *path, const char *mode)
    : file(*(new Fstream(path,mode))),
      readonly(strcmp(mode,"r")==0)
{
    if(!file.good()){
        cerr << "Cannot open LexiconFile: " << path << endl;
        exit(-1);
    }
    int newfile = strcmp(mode,"w") == 0;
    name = new char[strlen(path)+1];
    strcpy(name,path);
    kindex = 0;

/*
    offsetTbl = new size_t[65536];
*/
    offsetTbl = new Lexoff_t[65536];
    err = 0;
    if(newfile){
/*
        assert(sizeof(size_t) == 4);
*/
        assert(sizeof(Lexoff_t) == 4);
        memset((char*)&Head, 0, HEADERSIZE);
        memcpy(Head.magic,MAGIC, sizeof(MAGIC));
        memset(offsetTbl, 0, 65536 * sizeof(offsetTbl[0]));
        Head.fileSize = HEADERSIZE + 65536 * sizeof(offsetTbl[0]);

        writeHeader();
    }else{
        readHeader();
/*
        kindex = (int*)mmap(NULL,Head.nentry*4,PROT_READ,MAP_SHARED,file.fd(),Head.keymapoff);
*/
/*
        for(int i = 0; i < Head.nentry; i++){
          LexiconEntry entry;
          getlexstr(i,entry);
          fprintf(stderr,"[%d] %s\n",i,entry.getkey());
        }
*/
    }
}
void
LexiconFile::dump(){
    if(!file.good()){
        return;
    }
    for(int i = 0; i < Head.nentry; i++){
        LexiconEntry entry;
        getlexstr(i,entry);
/*
        printf("%d %s\n",entry.getrefcount(),entry.getkey());
*/
        printf("%d %d %s\n",i,entry.getrefcount(),entry.getkey());
    }
}
int LexiconFile::lexid(const char *key){
	if(!file.good()){
		return -1;
	}
	int li = Head.nentry / 2;
	int inc = Head.nentry / 4;
	LexiconEntry entry;
	for(;;){
		getlexstr(li,entry);
/*
		fprintf(stderr,"%4d %4d %s\n",li,inc,entry.getkey());
*/
		if( strcmp(key,entry.getkey()) == 0 )
			return li;
		if( inc == 0 )
			break;
		if( 0 < strcmp(key,entry.getkey()) )
			li += inc;
		else	li -= inc;
		inc /= 2;
	}
	getlexstr(li,entry);
	if( 0 < strcmp(key,entry.getkey()) )
		inc = 1;
	else	inc = -1;
	int lx;
	for( lx = 0; lx < 100; lx++ ){
		li += inc;
		getlexstr(li,entry);
/*
		fprintf(stderr,"%4d %4d %s\n",li,inc,entry.getkey());
*/
		if( strcmp(key,entry.getkey()) == 0 )
			return li;
		if( 0 < inc ){
			if( strcmp(key,entry.getkey()) < 0 )
				break;
		}else{
			if( strcmp(key,entry.getkey()) > 0 )
				break;
		}
	}
	return -1;
}


LexiconFile::~LexiconFile(){
    if(! readonly){
        writeHeader();
    }
    file.close();
    delete []offsetTbl;
}

void
LexiconFile::printDebug(ostream &out) const{
    out << "-- LexiconFile --" << endl;
    for(int i = 0; i < 256; i++){
        for(int j = 0; j < 256; j++){
            int k = i * 256 + j;
            if(offsetTbl[k]){
                out << "0x" << setfill('0') << setw(4) << hex << k << ' ';
                out << "0x" << offsetTbl[k] << " \""
                    << (unsigned char)i << (unsigned char)j << '"' << endl;
            }
            out << dec;
        }
    }
}

void
/*
htonHead(LexHead &head, size_t *offsetTbl){
*/
htonHead(LexHead &head, Lexoff_t *offsetTbl){
    head.nentry = htonl(head.nentry);
    head.fileSize = htonl(head.fileSize);
    head.endofLexicons= htonl(head.endofLexicons);
    head.keymapoff= htonl(head.keymapoff);
    for(int i = 0; i < 65536; i++){
        offsetTbl[i] = ntohl(offsetTbl[i]);
    }
}
void
/*
ntohHead(LexHead &head, size_t *offsetTbl){
*/
ntohHead(LexHead &head, Lexoff_t *offsetTbl){
    head.nentry = ntohl(head.nentry);
    head.fileSize = ntohl(head.fileSize);
    head.endofLexicons= ntohl(head.endofLexicons);
    head.keymapoff= ntohl(head.keymapoff);
    for(int i = 0; i < 65536; i++){
        if(offsetTbl[i])
            offsetTbl[i] = ntohl(offsetTbl[i]);
    }
}


int
LexiconFile::readHeader(){
    file.clear();
    file.seekg(0);
    file.clear();

/*
    file.read(headerbuf, HEADERSIZE);
*/
    file.read((char*)&Head, HEADERSIZE);
    if(file.gcount() != HEADERSIZE || file.fail()){
        err = "Bad Header";
        PRINTERRMSG("Invalid lexicon file.");
        return 1;
    }
/*
    if(0 != strcmp(headerbuf, MAGIC)){
*/
    if(0 != strcmp(Head.magic, MAGIC)){
        err = "Bad Magic";
        PRINTERRMSG("Invalid lexicon file! Please check the file.");
        return 2;
    }

    file.read((char*)offsetTbl, 65536 * sizeof(offsetTbl[0]));
    ntohHead(Head,offsetTbl);

    file.clear();
    file.seekg(0, ios::end);
    return file.fail();
}

int
LexiconFile::getlexstr(int lid, LexiconEntry &entry){
/*
    int off;
*/
    Lexoff_t off;

  if(kindex){
    off = kindex[lid];
  }else{
    file.clear();
/*
    file.seekg(Head.keymapoff+lid*sizeof(int),ios::beg);
*/
    file.seekg(Head.keymapoff+lid*sizeof(Lexoff_t),ios::beg);
    file.read((char*)&off,sizeof(off));
  }
    off = ntohl(off);

    file.clear();
    file.seekg(off,ios::beg);
    file >> entry;
    return 0;
}

int
LexiconFile::writeHeader(){
    file.flush();
    file.clear();
    file.seekp(0);
    file.clear();

    LexHead XHead = Head;
/*
    size_t *XoffsetTbl = new size_t[66636];
*/
    Lexoff_t *XoffsetTbl = new Lexoff_t[66636];
    memcpy((char*)XoffsetTbl,(char*)offsetTbl,65536*sizeof(offsetTbl[0]));
    htonHead(XHead,XoffsetTbl);
    file.write((char*)&XHead, HEADERSIZE);
    file.write((char*)XoffsetTbl, 65536 * sizeof(offsetTbl[0]));
    file.flush();
    return file.fail();
}

/*
istream &
*/
Fstream &
LexiconFile::rewind(){
    off_t off;
    file.clear();
/*
    file.seekg(HEADERSIZE + sizeof(offsetTbl[0]) * 65536);
*/
    off = HEADERSIZE + sizeof(offsetTbl[0]) * 65536;
    file.seekg(off);
    return file;
}

void
LexiconFile::prepareAppend(){
    memset(offsetTbl, 0, 65536 * sizeof(offsetTbl[0]));
    file.clear();
    file.seekp(HEADERSIZE + sizeof(offsetTbl[0]) * 65536);
    file.clear();
    Head.fileSize = file.tellp();
}

int
LexiconFile::indexhash(const char *key){
    unsigned char hi, lo;
    hi = *(unsigned char *)key++;
    lo = *(unsigned char *)key;
    return hi * 256 + lo;
}

int
LexiconFile::appendEntry(const LexiconEntry &entry){
    int idx = indexhash(entry.getkey());
    if(0 == offsetTbl[idx])
        offsetTbl[idx] = file.tellp();
    file << entry;
    Head.fileSize = file.tellp();
    Head.nentry += 1;
    return 0;
}

void
LexiconFile::appendKeymap(FILE *kfp, int size){
    char buf[64*1024];

    Head.endofLexicons = file.tellp();
    // append EOL (End Of Lexicons)
    for(int i = 0; i < 4; i++)
        buf[i] = 0;
    file.write(buf,4);

    int rcc;
    Head.keymapoff = file.tellp();
    int kid;
    while(0 < (rcc = Fread(kfp,(char*)&kid,sizeof(kid)))){
       kid = htonl(kid);
       file.write((char*)&kid,rcc);
    }
/*
    while(0 < (rcc = Fread(kfp,buf,sizeof(buf)))){
       file.write(buf,rcc);
    }
*/
    file.flush();
/*
fprintf(stderr,"##Lexicons: %d,%d %s\n",Head.endofLexicons,file.tellp(),name);
*/
}


// exact match
int
LexiconFile::search(const char *key, LexiconEntry &result){
/*
    size_t from, to;
*/
    Lexoff_t from, to;
    if(0 == findRegion(key, from, to))
        return 0;
    //cout << "LexiconFile::search: from = " << from << " to = " << to << endl;
    file.clear();
    file.seekg(from);
/*
    while((size_t)file.tellg() < to){
*/
    while((off_t)file.tellg() < (off_t)to){
        LexiconEntry entry;
        file >> entry;
        //cout << "  key = \"" << key << "\" entry = \""
        //     << entry.getkey() << "\"" << endl;
        int cmp = strcmp(entry.getkey(), key);
        if(cmp >= 0){
            if(0 == cmp){
                result = entry;
                return 1;       // found
            }else break;        // not found
        }
    }
    return 0;                   // not found
}

// ׸
//   : MAXPREFIXMATCHĤߤĤʤ
//         ɤȾ1ʸΤȤexact matchƱ
LexiconEntry **
LexiconFile::prefixsearch(const char *key, int &n){
    typedef LexiconEntry *LEP;
    LexiconEntry **entries = new LEP[MAXPREFIXMATCH + 1];
/*
    size_t from, to;
*/
    Lexoff_t from, to;
    n = 0;
    if(0 == findRegion(key, from, to))
        return NULL;

    int keyLen = strlen(key);
    file.clear();
    file.seekg(from);
    LexiconEntry *entryp;
/*
    while((size_t)file.tellg() < to){
*/
    while((Lexoff_t)file.tellg() < to){
        entryp = new LexiconEntry;
        file >> *entryp;
        int cmp = strncmp(entryp->getkey(), key, keyLen);
        if(cmp >= 0){
            if(cmp == 0){
                entries[n++] = entryp;
                if(n == MAXPREFIXMATCH) break;
            }else{
                delete entryp;
                break;
            }
        }else
            delete entryp;
    }
    entries[n] = NULL;
    return entries;
}

// find longest prefix (mainly for Japanese word breaking)
// : ѿʸ(1byte)prefixϸĤʤ
int
LexiconFile::findPrefix(const char *text, LexiconEntry &result){
/*
    size_t from, to;
*/
    Lexoff_t from, to;
    if(0 == findRegion(text, from, to))
        return 0;               // not found
    //cout << "LexiconFile::findPrefix: from = " << from
    //     << " to = " << to << endl;
    file.clear();
    file.seekg(from);
    int matchLen = 0;
/*
    while((size_t)file.tellg() < to){
*/
    while((Lexoff_t)file.tellg() < to){
        LexiconEntry entry;
        file >> entry;
        //cout << "  text = \"" << text << "\" entry = \""
        //     << entry.getkey() << "\"" << endl;
        int cmp = jisprefix(entry.getkey(), text);
        if(cmp >= 0){
            if(cmp == 0){
                matchLen = strlen(entry.getkey());
                result = entry;
            }else
                break;
        }
    }
    return matchLen;
}

int
/*
LexiconFile::findRegion(const char *key, size_t &from, size_t &to){
*/
LexiconFile::findRegion(const char *key, Lexoff_t &from, Lexoff_t &to){
    int idx = indexhash(key);
    from = offsetTbl[idx];
    if(0 == from)
        return 0;
/*
    to = filesize;
*/
    to = Head.endofLexicons;
    for(int i = idx + 1; i < 65536; i++){
        if(0 != offsetTbl[i]){
            to = offsetTbl[i];
            return 2;
        }
    }
    return 1;
}
