// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Query.cc,v 0.9 1997-12-27 07:21:10+09 harada Exp $
//
#include "SX.h"
#include "Query.h"

void
Query::lowerword(char *word){
    char *p = word;
    if(MinCAPITALCOL){
        int capcol = 0;
        for(int i = 0; word[i]; i++){
            if(isupper(word[i])){
                capcol = i;
            }
        }
        if(capcol < MinCAPITALCOL-1){
            unsigned char ch;
            for(char *p = word; (ch = *p); p++){
                if(isupper(ch))
                    *p = tolower(*p);
            }
        }
    }else
    while(*p){
        if(isupper(*(unsigned char *)p))
            *p = *p - 'A' + 'a';
        p++;
    }
}

void
Query::printtype(ostream &out){
/*
    char *types[] = {"url:", "title:", "author:",
                     "keyword:", "", ""};
    if(type > 0)
        out << types[type - 1];
*/
    const char *name;
    if((name = ElemQName(type)) != NULL)
        out << name << ":";
}


// ------------------------------------------------------------

SimpleQuery::SimpleQuery(const char *str, int t)
    : Query(t), word(NULL), isprefixsearch(0)
{
    char wordbuf[MAXKEYLENGTH + 1];
    const char *cp = str;
    while(isspace(*(unsigned char *)cp)) cp++;

    char *dp = wordbuf;
    int haslower = 0;
    while(*cp && !isspace(*(unsigned char *)cp)
          && dp - wordbuf < MAXKEYLENGTH){
        if(islower(*(unsigned char *)cp)) haslower = 1;
        *dp++ = *cp++;
    }
    *dp = '\0';
    if(haslower) lowerword(wordbuf);

    int wordLen = dp - wordbuf;
    if(wordLen <= 0){
        err = EMPTYQUERY;
        return;
    }
    if(wordLen > 1 && wordbuf[wordLen - 1] == '*'){
        isprefixsearch = 1;
        wordbuf[--wordLen] = '\0';
    }
    for(int i = 0; i < wordLen; i++){
        if(wordbuf[i] == '*'){
/*
fprintf(stderr,"#### %d[%s]\n",i,wordbuf);
            if(strcmp(wordbuf,"**")==0){
            }else{
*/
            err = ASTERISKLOCATION;
            return;
/*
            }
*/
        }
    }
    word = new char[wordLen + 1];
    strcpy(word, wordbuf);
}

SimpleQuery::~SimpleQuery(){
    if(word) delete []word;
}

ostream &
SimpleQuery::print(ostream &out){
    if(word != NULL){
        printtype(out);
        out << word;
        if(isprefixsearch)
            out << '*';
    }
    return out;
}

void
SimpleQuery::printDebug(ostream &out){
    out << "SimpleQuery: word = ";
    if(NULL == word) out << "(null)";
    else{
        out << '"';
        print(out);
        out << '"';
    }
    out << endl;
}

Result *
SimpleQuery::eval(Retriever &ret, int &resultLen){
    resultLen = 0;
    if(NULL == word) return NULL;
/*
    return ret.retrieve(word, resultLen, type, isprefixsearch);
*/
    Result *result;
    if( (word[0] & 0x80) && (word[1] & 0x80) && (word[2] == 0) ){
	/* a single Japanese charcter */
        result = ret.retrieve(word, resultLen, type, 1);
        return result;
    }
    result = ret.retrieve(word, resultLen, type, isprefixsearch);
    if( resultLen == 0 && (word[0] & 0x80) ){
	/* automatially retry prefix-search for Japanese word */
	result = ret.retrieve(word, resultLen, type, 1);
    }
    return result;
}

// ------------------------------------------------------------

PhraseQuery::PhraseQuery(const char *str, int t)
    : Query(t), wordNum(0)
{
    if(str == NULL) return;
    tokenize(str);
}

PhraseQuery::~PhraseQuery(){
    for(int i = 0; i < wordNum; i++)
        delete []words[i];
}

Result *
PhraseQuery::eval(Retriever &ret, int &resultLen){
    resultLen = 0;
    if(0 == wordNum){
        return NULL;
    }else if(1 == wordNum){
        return ret.retrieve(words[0], resultLen, type);
    }
    return ret.retrievePhrase(words, offsets, wordNum, resultLen, type);
}

int
PhraseQuery::tokenize(const char *str){
    const char *cp = str;
    while(isspace(*(unsigned char *)cp)) cp++;
/*
    int offset = 0;
*/
    while(*cp){
        char wordbuf[MAXKEYLENGTH + 1];
        char *dp = wordbuf;
        int haslower = 0;
        while(*cp && !isspace(*(unsigned char *)cp)){
            if(islower(*(unsigned char *)cp)) haslower = 1;
            *dp++ = *cp++;
        }
        *dp = '\0';
        if(haslower){           // ٤ƾʸ
/*
            char *p = wordbuf;
            while(*p){
                if(isupper(*(unsigned char *)p))
                    *p = *p - 'A' + 'a';
                p++;
            }
*/
            lowerword(wordbuf);
        }
        int wordLen = dp - wordbuf;
        words[wordNum] = new char[wordLen + 1];
        strcpy(words[wordNum], wordbuf);
/*
        offsets[wordNum] = offset;
        offset += wordLen + 1;
*/
        offsets[wordNum] = topos(0,0,wordNum);
        wordNum++;
        
        while(isspace(*(unsigned char *)cp)) cp++;
    }
    return wordNum;
}

ostream &
PhraseQuery::print(ostream &out){
    if(wordNum == 0){
        out << "(empty phrase)";
    }else{
        printtype(out);
        out << '"' << words[0];
        for(int i = 1; i < wordNum; i++)
            out << ' ' << words[i];
        out << '"';
    }
    return out;
}

void
PhraseQuery::printDebug(ostream &out){
    out << "PhraseQuery: wordNum = " << wordNum << " \"";
    for(int i = 0; i < wordNum; i++)
        out << words[i] << '(' << offsets[i] << ") ";
    out << '"' << endl;
}

#ifdef QUERY_DEBUG
int main(){
    PhraseQuery ph("harada masanori");
    ph.printDebug();

    return 0;
}
#endif
