// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Retriever.cc,v 0.9 1997-12-27 07:23:31+09 harada Exp $
//
#include "SX.h"
#include "kcnvlib.h"
#include "settemplate.h"
#include "LexiconEntry.h"
#include "Retriever.h"

Retriever::Retriever(LexiconFile &lex, IndexFile &idxfile, const XMap &map)
    : lexiconfile(lex), indexfile(idxfile), xmap(map)
{
    wordoff = 0;
    Idxid = 0;
    attop = 0;
}

Retriever::~Retriever(){
}

pos_t *
Retriever::searchJapaneseWordX(const char *key, int &resultLen,
                              int isprefixsearch){
    pos_t *pfxlist = 0;
    int pfxlistLen = 0;
    if(isprefixsearch){
        // SX: search words including the key as a prefix
        //  -- ex. "keyword" for "key*"
        // which is not got in searchJapaneseWord()
        pfxlist = prefixsearchWord(key, pfxlistLen);
    }

    int listLen = 0;
    pos_t *list = searchJapaneseWord(key, listLen, isprefixsearch);
    if(pfxlistLen == 0 && listLen == 0){
         resultLen = 0;
         return NULL;
    }
    if(pfxlistLen == 0){
         resultLen = listLen;
         return list;
    }
    if(listLen == 0){
         resultLen = pfxlistLen;
         return pfxlist;
    }

    // merge
    pos_t *mlist = new pos_t[pfxlistLen+listLen];
    int mergedLen = mergeList(mlist,pfxlist,pfxlistLen,list,listLen);
    delete []list;
    delete []pfxlist;
    resultLen = mergedLen;
    return mlist;
}

pos_t *
Retriever::searchJapaneseWord(const char *key, int &resultLen,
                              int isprefixsearch){
    pos_t *list = NULL;
    int listLen = 0;
    int p = 0;
    resultLen = 0;
    int swordoff = wordoff;
    while(key[p] != '\0'){
        LexiconEntry entry;
        // cout << "key = " << key + p << endl;
        int matchLen = lexiconfile.findPrefix(key + p, entry);

#ifdef DEBUG
fprintf(stderr,"###(%d) J p=[%2d] keyLen=%d matchLen=%d [%s]%d\n",
wordoff,p,strlen(key),matchLen,key+p,strlen(entry.getkey()));
#endif

        // cout << "matchLen = " << matchLen << endl;
        if(! isprefixsearch){
            if(0 == matchLen){
                if(list) delete []list;
                resultLen = 0;
                return NULL; // not found
            }
        }else{
            if(0 == matchLen || *(key + p + matchLen) == '\0'){
                int tmpLen = 0;
                pos_t *tmplist = prefixsearchWord(key + p, tmpLen);
                if(tmplist){
                    if(list){
                        int mergedLen = (tmpLen < listLen) ? tmpLen : listLen;
                        pos_t *mergedlist = new pos_t[mergedLen];
                        int diff = topos(0,0,wordoff-swordoff);
                        mergedLen = multiplyList(mergedlist, list, listLen,
                                                 tmplist, tmpLen, diff);
/*
                                                 tmplist, tmpLen, p);
*/
                        delete []tmplist;
                        delete []list;
                        if(mergedLen > 0){
                            resultLen = mergedLen;
                            return mergedlist;
                        }else{  
                            delete []mergedlist;
                            resultLen = 0;
                            return NULL;    // not found
                        }
                    }else{
                        assert(p == 0);
                        resultLen = tmpLen;
                        return tmplist;
                    }
                }
            }
        }
        assert(matchLen % 2 == 0);
        if(matchLen == 0){
        // SX: fetching empty string will cause error
        // and continueing search for prefix* will find strange result *fix*
            resultLen = 0;
            break;
        }
        if(list == NULL){
            list = entry.fetch(indexfile, listLen);
            if(list == NULL) PRINTERRMSG("Internal error.");
            resultLen = listLen;
            assert(list != NULL);
            assert(listLen > 0);
        }else{
            int tmplistLen = 0;
            int mergedLen = 0;
            pos_t *tmplist = entry.fetch(indexfile, tmplistLen);
            pos_t *mergedlist;

            if(tmplist == NULL) PRINTERRMSG("Internal error.");
            assert(tmplist != NULL);
            assert(tmplistLen > 0);
            mergedLen = (listLen < tmplistLen) ? listLen : tmplistLen;
            mergedlist = new pos_t[mergedLen];

            int diff = topos(0,0,wordoff-swordoff);;
            mergedLen = multiplyList(mergedlist, list, listLen,
                                     tmplist, tmplistLen, diff);
/*
                                     tmplist, tmplistLen, p);
*/

            delete []tmplist;
            delete []list;
            if(mergedLen > 0){
                list = mergedlist;
                resultLen = listLen = mergedLen;
            }else{
                delete []mergedlist;
                resultLen = 0;
                return NULL;    // not found
            }
        }
        wordoff++;
        if(key[p + matchLen] == '\0')
            break;
        if(matchLen <= 4){
            p += 2;
        }else{
            p += matchLen;
        }
    }
    return list;
}

pos_t *
Retriever::searchWord(const char *key, int &resultLen){
    LexiconEntry entry;
    resultLen = 0;
    if(0 == lexiconfile.search(key, entry))
        return NULL;

#ifdef DEBUG
fprintf(stderr,"###(%d) W [%s]\n",wordoff,key);
#endif

    wordoff++;
    return entry.fetch(indexfile, resultLen);
}


/*
set offset to next word for each pos_t into itself
ex. at (pos_t&0xFF00000000000000) and masking it on comparation

template <class T>
int multiplyListGT(T *z, const T *x, int xn, const T *y, int yn, int diff){
    int zn;
    int i, j;
    zn = i = j = 0;
    while(i < xn && j < yn){
        if(x[i]+diff <= y[j]) z[zn++] = x[i];
        if(x[i] < y[j]) i++;
        else if(x[i] > y[j]) j++;
        else{ i++; j++; }
    }
    return zn;
}
*/


// ե졼
pos_t *
Retriever::searchPhrase(char **keys, int offsets[],
                        int nkeys, int &resultLen, int isprefixsearch){
    pos_t *list = NULL;
/*
    int afterwildcard = 0;
*/
    int pwordoff = 0;
    wordoff = 0;
    for(int k = 0; k < nkeys; k++){
        if(0 < k){
            if(strcmp(keys[k],"*") == 0){
                wordoff++;
                continue;
            }
/*
            if(strcmp(keys[k],"**") == 0){
                afterwildcard = 1;
                continue;
            }
*/
        }

        pos_t *tmplist;
        int tmplistLen;

        pwordoff = wordoff;
        if(k == nkeys - 1)
            tmplist = search(keys[k], tmplistLen, isprefixsearch);
        else
            tmplist = search(keys[k], tmplistLen, 0);
        if(tmplist == NULL){    // not found
            if(list) delete []list;
            return NULL;
        }
        if(list == NULL){
            list = tmplist;
            resultLen = tmplistLen;
        }else{
            int mergedLen = (resultLen < tmplistLen) ? resultLen : tmplistLen;
            pos_t *mergedlist = new pos_t[mergedLen];
/*
            if(afterwildcard){
fprintf(stderr,"#### AFTER WILD CARD [%d][%s] listlen=%d\n",k,keys[k],tmplistLen);
            mergedLen = multiplyListGT(mergedlist, list, resultLen,
                                     tmplist, tmplistLen offsets[k]);
            afterwildcard = 0;
            }else
*/

            int diff = topos(0,0,pwordoff);
            mergedLen = multiplyList(mergedlist, list, resultLen,
                                     tmplist, tmplistLen, diff);
/*
                                     tmplist, tmplistLen, offsets[k]);
*/

#ifdef DEBUG
fprintf(stderr,"###(%d)-%d phrase[%d] off=%d {CAND=%d HIT=%d}->%d\n",
pwordoff,wordoff,k,offsets[k],resultLen,tmplistLen,mergedLen);
/*
for(int i=0;i<resultLen;i++)
fprintf(stderr,"L[%2d] %2d %d -> %d\n",i,pos2did(list[i]),
pos2off(list[i]),pos2off(list[i])+pwordoff);
for(int i=0;i<tmplistLen;i++)
fprintf(stderr,"N[%2d] %2d %d\n",i,pos2did(tmplist[i]),pos2off(tmplist[i]));
*/
#endif

            delete []list;
            delete []tmplist;
            if(mergedLen > 0){
                list = mergedlist;
                resultLen = mergedLen;
            }else{ // not found
                delete []mergedlist;
                return NULL;
            }
        }
    }
    return list;
}

pos_t *
Retriever::prefixsearchWord(const char *key, int &resultLen){
    int i, n;
    LexiconEntry **entries;

    resultLen = 0;
    entries = lexiconfile.prefixsearch(key, n);
    if(NULL == entries) return NULL; // not found

    //cout << "prefixsearch: key=" << key << endl;
    //for(i = 0; i < n; i++){
    //    cout << "  " << entries[i]->getkey() << endl;
    //}

    pos_t *list;
    if(n == 1){
        list = entries[0]->fetch(indexfile, resultLen);
    }else{
/*
        pos_t *lists[n];
        int listLens[n];
*/
        typedef pos_t *POSP;
        pos_t **lists = new POSP[n];
        int *listLens = new int[n];
        int totalLen = 0;
        for(i = 0; i < n; i++){
            lists[i] = entries[i]->fetch(indexfile, listLens[i]);
            assert(lists[i] != NULL);
            assert(listLens[i] > 0);
            totalLen += listLens[i];
        }
        list = new pos_t[totalLen];

/*
LapReport(0,"SORT START %d/%d",totalLen,n,0,0);
*/
        int mergedLen =
        mergeLists(list,totalLen,lists,listLens,n);

        for(i = 0; i < n; i++)
            delete []lists[i];
/*
LapReport(0,"SORT DONE",0,0,0,0);
*/

/*
        resultLen = totalLen;
*/
        delete []lists;
        delete []listLens;
        resultLen = mergedLen;
    }

    for(i = 0; i < n; i++)
        delete entries[i];
    delete []entries;
    if(resultLen == 0){
        if(list) delete []list;
        return NULL;
    }
    return list;
}

// ޤޤʤθ
//      search()searchPhrase()ߤ˸ƤӤ礦
//      NPפΤ褦ܸե졼ե졼˸뤳Ȥ뤿
pos_t *
Retriever::search(const char *text, int &resultLen, int isprefixsearch){
    char keybufs[MAXWORDSINPHRASE][MAXKEYLENGTH];
    char *keys[MAXWORDSINPHRASE];
    int offsets[MAXWORDSINPHRASE];
    const char *cp = text;
    pos_t *list = NULL;
    int k = 0;
    while(*cp){
        offsets[k] = cp - text;
        char *dp = keys[k] = keybufs[k];
        if(Is_kanji_here(cp)){
            do { *dp++ = *cp++; *dp++ = *cp++; } while(Is_kanji_here(cp));
        }else{
            do { *dp++ = *cp++; } while(*cp && !Iskanji1(*cp));
        }
        *dp = '\0';
        k++;
    }
    if(k == 1){
        if(Is_kanji_here(keys[0])){
/*
            list = searchJapaneseWord(keys[0], resultLen, isprefixsearch);
*/
            list = searchJapaneseWordX(keys[0], resultLen, isprefixsearch);
        }else{
            if(isprefixsearch)
            {
                list = prefixsearchWord(keys[0], resultLen);
                wordoff++;
            }
            else
                list = searchWord(keys[0], resultLen);
        }
    }else if(k > 1){
        list = searchPhrase(keys, offsets, k, resultLen, isprefixsearch);
    }
    if(resultLen == 0){
        if(list) delete []list;
        return NULL;
    }
    return list;
}


int isnot_attop(ConText *Ctx,int id,const XMap *xmap,int inoffset){
	char pw[32];
	char cw[32];
	int len;

	off_t base = xmap->getdocbase(id);
	Ctx->getword(pw,sizeof(pw),base,inoffset-2);
	if( pw[0] & 0x80 ){
		len = strlen(pw);
	}else	len = 0;
	Ctx->getword(pw+len,sizeof(pw)-len,base,inoffset-1);
	Ctx->getword(cw,sizeof(cw),base,inoffset-0);

	if( (cw[0] & 0x80) && (pw[0] & 0x80) && 2 <= strlen(pw) ){
		const char *pp;

		for( pp = pw; pp[2]; pp++ );
		if( pp[0] == cw[0] && pp[1] == cw[1] )
			pp -= 2;

		if( pw <= pp ){
			if( Is_katakana_here(cw) && Is_katakana_here(pp) ){
				return 1;
			}
			if( Is_hiragana_here(cw) && Is_hiragana_here(pp) ){
				return 1;
			}
		}
	}
	return 0;
}

Result *
Retriever::translateList(const pos_t *list, int listLen,
                         int &resultLen, int type){
    int i;
    assert(listLen > 0);
    Result *result = new Result[listLen];
/*
    int docsize[listLen];
*/
    int *docsize = new int[listLen];
    int r = 0;
    urlid_t lastid = (urlid_t)-1;

    int siz = sizeof(result[0].list);
    int rem = siz;
    for(i = 0; i < listLen; i++){
        urlid_t id;
        size_t inoffset;
        int t;
/*
        if(0 != xmap.translate(list[i], id, inoffset, t)){
*/
        if(0 != frompos(list[i], id, inoffset, t)){
            PRINTERRMSG("translation fail!");
            // cerr << "-> list[" << i << "]=" << list[i] << endl;
/*
        }else if(type == 0 || type == t){
*/
        }else if(type == TYPE_ANY || type == t){
            if( attop && Ctx ){
		if( isnot_attop(Ctx,id,&xmap,inoffset) )
			continue;
            }

            if(id == lastid) r--;
            else rem = siz;
            result[r].docid = id;
/*
            result[r].docsize = xmap.getdocsize(id);
*/
            docsize[r] = xmap.getdocsize(id);
            if(result[r].ntimes < 255)
            result[r].ntimes += 1;
/*
            result[r].weight += xmap.score(t);
*/
            result[r].weight += ElemScore(id,t);

            if(0 < rem)
            if(t == type || t == TYPE_TEXTW3 || t == TYPE_LINK){
                pos_t val[1];
                val[0] = inoffset;
                int len = sputintvl(result[r].list+(siz-rem),rem,val,1);
                if( 0 < len )
                    rem -= len;
                else
                    rem = 0;
            }
            r++;
            lastid = id;
        }
    }
    if(r == 0){                 // not found
        delete []result;
        delete []docsize;
        return NULL;
    }
    resultLen = r;
    // calculate the average size
    double sum = 0;
    for(i = 0; i < resultLen; i++)
        sum += docsize[i];
/*
        sum += result[i].docsize;
*/
    double avg = (size_t)(sum / resultLen);
    // calculate `tf * idf'
    double idf = log((double)(2 * xmap.getdocNum() / resultLen));
    for(i = 0; i < resultLen; i++)
/*
        if(result[i].docsize > avg / 2)
            result[i].weight *= (avg * idf) / result[i].docsize;
*/
        if(docsize[i] > avg / 2)
            result[i].weight *= (avg * idf) / docsize[i];
        else
            result[i].weight *= idf;
    delete []docsize;
    return result;
}


Result *
Retriever::retrievePhrase(char **keys, int *offsets, int nkeys,
                          int &resultLen, int type){
    int listLen;
    pos_t *list = searchPhrase(keys, offsets, nkeys, listLen);
    if(list == NULL)
        return NULL;
    else{
        Result *result = translateList(list, listLen, resultLen, type);
        delete []list;
        return result;
    }
}

void h_addkey(int xid,int kid,const char *key);
Result *
Retriever::retrieve(const char *text, int &resultLen,
                    int type, int isprefixsearch){
    int listLen;
    if( *text == '^' ){
        attop = 1;
        text++;
    }

    pos_t *list = search(text, listLen, isprefixsearch);
    if(list == NULL)
        return NULL;
    else{
        h_addkey(Idxid,lexiconfile.lexid(text),text);
        Result *result = translateList(list, listLen, resultLen, type);
        delete []list;
        return result;
    }
}
