// Copyright (c) 2004 Yutaka Sato. All rights reserved.
// SX.cc -- extensions for FreyaSX
//

#include <stdarg.h>
#include "SX.h"

#define SXBASE	"freyasx"
#define SXDICT	"etc/icot.dic"
#define pubPfx	"pub-"
#define SXCACHE	"/tmp/freyasx"

int opt_v;
int opt_S;

char *Strncpy(char *s1, const char *s2, int n){
	char *sp;

	sp = s1;
	while( 1 < n-- )
		if( (*sp++ = *s2++) == 0 )
			break;
	*sp = 0;
	return s1;
}
static void xbcopy(char *src, char *dst, int ssize, char *dstend){
	int dsize = dstend - dst;
	if( dsize < ssize ){
		fprintf(stderr,"String too large: %d/%d %s\n",ssize,dsize,src);
		ssize = dsize;
	}
	memmove(dst,src,ssize);
}
void strins(char *str, char *dst, char *dstend){
	xbcopy(dst,dst+strlen(str),strlen(dst)+1,dstend);
	memcpy(dst,str,strlen(str));
}
/*
int FileSize(const char *path){
*/
off_t FileSize(const char *path){
	struct stat st;
	if(stat(path,&st) == 0)
		return st.st_size;
	return -1;
}
int fileMtime(char *path){
	struct stat st;
	int time;
	if(stat(path,&st) == 0)
		time = st.st_mtime;
	else	time = -1;
	return time;
}
int pathIsdir(char *dir){
	struct stat st;
	if(stat(dir,&st) == 0)
		return (st.st_mode & S_IFDIR) != 0;
	return 0;
}
int Rename(const char *from, const char *to){
	int rcode;
	if( FileSize(from) < 0 ){
		return -1;
	}
	if( 0 <= FileSize(to) ){
		if( unlink(to) != 0 ){
			fprintf(stderr,"Can't remove: %s\n",to);
		}
	}
	rcode = rename(from,to);
	if( rcode != 0 ){
		fprintf(stderr,"Can't rename: %s %s\n",from,to);
		perror("rename");
	}
	return rcode;
}

static char *sDate = "%y%m%d";
const char *shortDate(char *buf, size_t bufsiz, time_t &t){
	(void)strftime(buf, bufsiz, sDate, localtime(&t));
	return buf;
}
double Time(){
	struct timeval tv;
	gettimeofday(&tv,NULL);
	return tv.tv_sec + (double)tv.tv_usec/1000000;
}
int fileReadable(const char *path){
	Fstream fs(path,"r");
	if( fs.good() ){
		return 1;
	}else	return 0;
}
static char *fsxhome;
char *FSXHOME(){
	char *home,path[1024];

	if( fsxhome )
		return fsxhome;

	home = getenv("FSXHOME");
	if( home == NULL ){
		if( (home = getenv("HOME")) != NULL ){
			sprintf(path,"%s/%s",home,SXBASE);
			home = strdup(path);
		}
	}
	if( home == NULL )
		home = "/usr/local/freyasx";
	fsxhome = home;
	return fsxhome;
}
char *SX_defaultDict(char *path, int psiz){
	Strncpy(path,SXDICT,psiz);
	if( fileReadable(path) ){
		return path;
	}
	/*
	char *home = getenv("HOME");
	*/
	char *home = FSXHOME();
	if( home != NULL ){
		Strncpy(path,home,psiz-64);
		sprintf(path+strlen(path),"/%s",SXDICT);
		if( fileReadable(path) ){
			return path;
		}
	}
	char *dictpath = DEFAULT_MORPHDIC_PATH;
	if( fileReadable(dictpath) ){
		return dictpath;
	}
	if( fileReadable("./icot.dic") ){
		return "./icot.dic";
	}
	return dictpath;
}

//
// find index base ...
// { ./index/ ./bank/index/ ../bank/index/ $HOME/freyasx/bank/index/ }
// should be castomizable with FSXPATH=...
//
int isLstFile(char *path){
	char lpath[PATH_MAX],xpath[PATH_MAX],idx[PATH_MAX];
	sprintf(lpath,"%s.lst",path);
fprintf(stderr,"########### [%s]\n",lpath);
	if( 0 < FileSize(lpath) ){
		FILE *fp;
		if( (fp = fopen(lpath,"r")) != NULL ){
			xpath[0] = 0;
			fgets(xpath,sizeof(xpath),fp);
			idx[0] = 0;
			sscanf(xpath,"%s",idx);
fprintf(stderr,"########### [%s][%s]\n",lpath,idx);
			fclose(fp);
		}
	}
	return 0;
}
int idxPathIsdir(char *idxpath){
	if( pathIsdir(idxpath) )
		return 1;
	char *dp;
	if( (dp = strrchr(idxpath,'.')) ){
		if( strchr(dp,'/') == 0 ){
			*dp = '/';
			if( pathIsdir(idxpath) )
				return 1;
			*dp = '.';
		}
	}
	return 0;
}
int SX_Base(char *idx, char *base, int size, int create){
	char home[PATH_MAX],*env;
	char tmp[PATH_MAX];
	char *name;

	if( idx != base )
		Strncpy(base,idx,size);

	if( idxPathIsdir(base) )
		goto FOUND;

	if( strchr(base,'/') == 0 ){
		Strncpy(home,idx,sizeof(home));
		strins("bank/",home,home+sizeof(home));
		if( idxPathIsdir(home) ){
			Strncpy(base,home,size);
			goto FOUND;
		}
		strins("../",home,home+sizeof(home));
		if( idxPathIsdir(home) ){
			Strncpy(base,home,size);
			goto FOUND;
		}

		/*
		if( (env = getenv("HOME")) != 0 ){
		*/
		if( (env = FSXHOME()) != 0 ){
			Strncpy(home,env,sizeof(home)-32-strlen(base));
			strcat(home,"/bank");
			if( idxPathIsdir(home) ){
				sprintf(home+strlen(home),"/%s",base);
				if( idxPathIsdir(home) ){
					Strncpy(base,home,size);
					goto FOUND;
				}
				if( create ){
					Strncpy(base,home,size);
				}
			}
		}
	}

	if( create )
	if( idxPathIsdir(base) == 0 )
		mkdir(base,0750);

FOUND:
	if( idxPathIsdir(base) ){
		Strncpy(tmp,base,sizeof(tmp));
		if( (name = strrchr(tmp,'/')) != 0 )
			name = name + 1;
		else	name = tmp;
		Strncpy(base,tmp,size-1-strlen(name));
		sprintf(base+strlen(base),"/%s",name);
	}
	return 0;
}

int SX_createBase(char *idx, char *base, int size){
	return SX_Base(idx,base,size,1);
}
int SX_searchBase(char *idx, char *base, int size){
	return SX_Base(idx,base,size,0);
}

// search "head.html" with "pfx=ja"
//   -> "pub-head-ja.html"
//   -> "head-ja.html"
// or without "pfx"
//   -> "pub-head.html"
//   -> "head.html"
//
int pubFileOK(const char *path, char *xpath, int xsize){
	if( fileReadable(path) ){
		return 0;
	}
	if( strncmp(path,pubPfx,strlen(pubPfx)) == 0 ){
		return -1;
	}
	if( xpath != path )
		Strncpy(xpath,path,xsize);
	strins(pubPfx,xpath,xpath+xsize);
	if( fileReadable(xpath) ){
		return 1;
	}
	return -1;
}

static int c_ok;
static char c_pfx[32],c_path[1024],c_xpath[1024];
int SX_pfxPath(char *pfx, const char *path, char *xpath, int xsize){
	int ok;

	if(strcmp(pfx,c_pfx)==0)
	if(strcmp(path,c_path)==0){
		Strncpy(xpath,c_xpath,xsize);
		return c_ok;
	}

	if( *pfx != 0 ){
		Strncpy(xpath,path,xsize);
		if(char *dp = strrchr(xpath,'.')){
			strins(pfx,dp,xpath+xsize);
			strins("-",dp,xpath+xsize);
			if( 0 <= (ok = pubFileOK(xpath,xpath,xsize)) ){
				c_ok = 1;
				goto OK;
/*
				return 1;
*/
			}
		}
	}
	if( 0 <= (ok = pubFileOK(path,xpath,xsize)) ){
		c_ok = ok;
		goto OK;
/*
		return ok;
*/
	}
	return -2;

OK:
	Strncpy(c_pfx,pfx,sizeof(c_pfx));
	Strncpy(c_path,pfx,sizeof(c_path));
	Strncpy(c_xpath,pfx,sizeof(c_xpath));
	return c_ok;
}

int SX_openCache(int wr, const char *Key, int Mtime, Fstream &cache){
	char path[PATH_MAX];
	int crc = 0;
	crc = strCRC32(crc,Key);
	sprintf(path,"%s/%02x",SXCACHE,0xFF&(crc>>24));
	mkdir(path,0775);
	sprintf(path+strlen(path),"/%08x",crc);

	int mtime = fileMtime(path);
	if( mtime != -1 && mtime < Mtime ){
		unlink(path);
		return -1;
	}

	if( wr )
		cache.open(path,"w");
	else{
		cache.open(path,"r");
	}
	// fprintf(stderr,">> %d %d %s %s\n",cache.good(),wr,path,Key);
	if(!cache.good())
		return -1;
	else{
		return 0;
	}
}

/*
char *scanBracket(char *str, char *subst, int ssize){
*/
char *scanBracket(char *str, char *ts, int tl, char *fs=NULL, int fl=0){
/*
	char *sp = subst;
*/
	char *sp = ts;
	int rem = tl;
	int lev = 0;
	int esc = 0;
	char *ep;
	char buf[1024];
	if( fs == NULL )
		fs = buf; 
	*fs = 0;

	for( ep = str; *ep; ep++ ){ 
		if( esc ){
			esc = 0;
			*sp++ = *ep;
		}else
/*
		if( *ep == '\\' && (ep[1] == '{' || ep[1] == '}') ){
*/
		if( *ep == '\\' && (ep[1]=='{' || ep[1]=='}' || ep[1]==':') ){
			esc = 1;
		}else{
			if( *ep == ':' ){
				if( lev == 0 ){
					*sp = 0;
					sp = fs;
					rem = fl;
					continue;
				}
			}
			if( *ep == '{' )
				lev++;
			else
			if( *ep == '}' ){
				lev--;
				if( lev < 0 )
					break;
			}
			*sp++ = *ep;
		}
	}
	*sp++ = 0;
	if( *ep == '}' ){
		ep = ep+1;
	}
	return ep;
}

#define SUB_DFLT	1 // ${name:-Word} -> $name if defined or use Word
#define SUB_UNDEF	2 // ${name:+Word} -> Word if $name is undefined
#define SUB_DEC		3 // ${num-1} -> num-1
#define SUB_IFEQ	4 // ${name==Pat?Str} -> Str if $name equals to Pat
#define SUB_IFNEQ	5 // ${name!=Pat?Str} -> Str if $name not equals to Pat
#define SUB_IFINC	6 // ${name=>Pat?Str} -> Str if $name includes Pat
#define SUB_INCLUDE	7 // ${incl*=file:format} -> include file with format

// apply format for each line in file
//
static int inclines(char *str, int size, FILE *afp, char *format){
	char line[256],a1[256],a2[256],a3[256];
	char *dp = str;
	int rem = size;
	while( 0 < rem && fgets(line,sizeof(line),afp) != NULL){
		if( *line == '#' || *line == '\r' || *line == '\n' )
			continue;
		a1[0] = a2[0] = a3[0] = 0;
		sscanf(line,"%[^\t\r\n] %[^\t\r\n] %[^\t\r\n]",a1,a2,a3);
		for(char *fp = format; *fp; fp++){
			if(*fp == '\\'){
				if( fp[1] == 0 )
					break;
				switch(fp[1]){
					case 's': *dp++ = ' '; break;
					case 't': *dp++ = '\t'; break;
					case 'n': *dp++ = '\n'; break;
					default: *dp++ = fp[1]; break;
				}
				*dp = 0;
				rem--;
				fp++;
				continue;
			}
			if(*fp == '$'){
				switch(fp[1]){
				case '1': Strncpy(dp,a1,rem); fp++; break;
				case '2': Strncpy(dp,a2,rem); fp++; break;
				case '3': Strncpy(dp,a3,rem); fp++; break;
				}
			}else{
				*dp++ = *fp;
				*dp = 0;
			}
			rem -= strlen(dp);
			dp += strlen(dp);
		}
	}
	*dp = 0;
	return 0;
}
int condSubst(char *str, int size, char *(*getrv)(const char *)){
	char *ip = strstr(str,"${");
	if( ip == NULL )
		return 0;

	char *np = ip + 2;
	char name[32];
	sscanf(np,"%32[a-z]",name);
	if( *name == 0 )
		return 0;

	char *bp = np+strlen(name);
	int type = 0;
	char pat[256];
	if( strncmp(bp,"*=",2) == 0 ){
		bp += 2;
		char file[64],format[1024],*ep;
		ep = scanBracket(bp,file,sizeof(file),format,sizeof(format));
		FILE *afp = fopen(file,"r");
		if(afp != NULL){
			inclines(str,size,afp,format);
		}
		return 1;
	}else
	if( strncmp(bp,"==",2) == 0
	 || strncmp(bp,"!=",2) == 0
	 || strncmp(bp,"=>",2) == 0
	){
		if(strncmp(bp,"==",2) == 0) type = SUB_IFEQ; else
		if(strncmp(bp,"!=",2) == 0) type = SUB_IFNEQ; else
		if(strncmp(bp,"=>",2) == 0) type = SUB_IFINC;
		bp += 2;
		char *pp = pat;
		for(; *bp; bp++){
			if(*bp == '?'){
				bp++;
				break;
			}
			*pp++ = *bp;
		}
		*pp = 0;
	}else
	if( strncmp(bp,"-1",2) == 0 ){
		type = SUB_DEC;
		bp += 2;
	}else
	if( strncmp(bp,":-",2) == 0 ){
		type = SUB_DFLT;
		bp += 2;
	}else
	if( strncmp(bp,":+",2) == 0 ){
		type = SUB_UNDEF;
		bp += 2;
	}else{
		return 0;
	}

	char subst[LINE_MAX];
	char selse[LINE_MAX];
/*
	char *ep = scanBracket(bp,subst,sizeof(subst));
*/
	char *ep = scanBracket(bp,subst,sizeof(subst),selse,sizeof(selse));

	char *sval = subst; 
	char *val = getrv(name);
	switch( type ){
		case SUB_IFEQ:
			if(strcmp(val,pat) == 0)
				sval = subst;
			else	sval = selse;
			break;
		case SUB_IFNEQ:
			if(strcmp(val,pat) != 0)
				sval = subst;
			else	sval = selse;
			break;
		case SUB_IFINC:
			if(strstr(val,pat) != 0)
				sval = subst;
			else	sval = selse;
			break;
		case SUB_DFLT:
			if( *val != 0 )
				sval = val;
			else	sval = subst;
			break;
		case SUB_UNDEF:
			if( *val != 0 )
				sval = subst;
			else	sval = selse;
			break;
		case SUB_DEC:
			sprintf(subst,"%d",atoi(val)-1);
			sval = subst;
			break;
	}
	memmove(ip,ep,strlen(ep)+1);
	strins(sval,ip,str+size);
	return 1;
}
int condSubsts(char *str, int size, char *(*getrv)(const char *)){
	int i;
	/* it must be large for inclines() but should be small to detect loop */
	for( i = 0; i < 32; i++ ){
		if( condSubst(str,size,getrv) == 0 )
			break;
	}
	return i;
}

int SX_tempName(char *path){
	strcat(path,"#");
	return 0;
}
int SX_realName(char *path){
	int len = strlen(path);
	if( path[len-1] == '#' ){
		char rpath[PATH_MAX];
		strcpy(rpath,path);
		rpath[len-1] = 0;
		if( Rename(path,rpath) == 0 ){
			strcpy(path,rpath);
			return 0;
		}
	}
	return -1;
}

int SX_arrayMap(char *path, int **addrp){
	int fd;
	struct stat st;
	int *addr;

	fd = open(path,0);
	if( 0 <= fd ){
		fstat(fd,&st);
		int size = st.st_size;
		addr = (int*)mmap(NULL,size,PROT_READ,MAP_SHARED,fd,0);
		if( addr != (int*)-1 ){
			*addrp = addr;
			return size;
		}else{
			close(fd);
		}
	}
	return -1;
}
int SX_unmap(void *addr,int len){
	munmap(addr,len);
	return 0;
}
/*
int SX_newMMap(int fd,int off,int size,void **addrp){
	int *addr;
	int i;
	FILE *fp;

	fp = fdopen(dup(fd),"w");
	fseek(fp,off,0);
	for( i = 0; i < size; i++ )
		putc(0,fp);
	fclose(fp);
	addr = (int*)mmap(NULL,size,PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
	if( addr != (int*)-1 ){
		*addrp = addr;
		return size;
	}else{
		close(fd);
		*addrp = 0;
		return -1;
	}
}
*/

#define TYPE_NUM 32
typedef struct {
        /*
	char *name;
	char *qname;
        */
	const char *name;
	const char *qname;
	int score;
} ElemTypes;

ElemTypes elemTypes[TYPE_NUM] = {
    {"META",	"meta",		 1,     }, /* 0x00 pseudo element */
    {"URL",	"url",		16,     }, /* 0x01 URL of the document */
    {"TEXTW3",	NULL,		 8,     }, /* 0x02 body of text */
    {"TEXTW2",	NULL,		 4,     }, /* 0x03 Heads (H1,H2,...) */
    {"TEXTW1",	NULL,		 6,     }, /* 0x04 Description */
    {"DIGEST",	"digest",	20,     }, /* 0x05 generated from TEXTWx */
    {"TITLE",	"title",	 9,     }, /* 0x06 extracted from TEXTWx */
    {"AUTHOR",	"author",	 9,     }, /* 0x07 extracted from TEXTW3 */
    {"LINK",	"link",		 8,     }, /* 0x08 HREF */
    {"KEYWORD",	"keyword",	12,     }, /* 0x09 optional, use 4bits */
};

const
char *ElemQName(int type){
	if( 0 <= type && type < TYPE_NUM )
		return elemTypes[type].qname;
	return NULL;
}
int ElemScore(int docid, int type){
	if(0 <= type && type < TYPE_NUM)
		return elemTypes[type].score;
	return -1;
}
int ElemId(const char *name, const char **nextp, const char *brk){
    for(int i = 0; i < TYPE_NUM; i++){
        const char *tp = elemTypes[i].name;
        if(tp == 0)
            break;
        const char *np = name;
        char tc,nc;
        for(; (tc = *tp) != 0; tp++){
            nc = *np;
            if(islower(nc))
               nc = toupper(nc);
            if(tc != nc)
                break;
            np++;
        }
        if(*tp == 0){
            if( *brk == 0 ){
                if(nextp != NULL)
                    *nextp = np+1;
                return i;
            }
            for(const char *bp = brk; *bp; bp++)
                if(*bp == *np){
                     if(nextp != NULL )
                         *nextp = np+1;
                     return i;
                }
        }
    }
    return -1;
}

typedef struct {
	int	*ival;
  const char	*name;
  const char	*abbr;
	int	 min;
	int	 max;
	int	 mul;
} Limit;

int MaxBUFFERINGWORDS = MAXBUFFERINGWORDS;
int MinBUFFERINGWORDS = MINBUFFERINGWORDS;
int MaxLEXLOCALINDEX  = 4;
int MaxKEYLENGTH      = MAXKEYLENGTH;
int MinKEYLENGTH      = MINKEYLENGTH;
int MinCAPITALCOL     = MINCAPITALCOL; // min. column of capitcal char. in a word to be Lexicon
int IoBUFFERSIZE      = 0;

Limit limits[] = {
/*
{&MaxBUFFERINGWORDS,	"MAXBUFFERINGWORDS",	"MBW",	4,	1024,	1024},
*/
{&MaxBUFFERINGWORDS,	"MAXBUFFERINGWORDS",	"MBW",	4,   32*1024,	1024},
{&MinBUFFERINGWORDS,	"MINBUFFERINGWORDS",	"NBW",	4,	1024,	1024},
{&MaxLEXLOCALINDEX,	"MAXLEXLOCALINDEX",	"MLI",	0,	  15,	   1},
{&MaxKEYLENGTH,		"MAXKEYLENGTH",		"MKL",	8,	  64,	   1},
{&MinKEYLENGTH,		"MINKEYLENGTH",		"NKL",	1,	  16,	   1},
{&MinCAPITALCOL,        "MINCAPITALCOL",        "MCC",  1,        64,      1},
{&IoBUFFERSIZE,         "IOBUFFERSIZE",         "IOB",  0,        64,   1024},
{0}
};
void SX_scanCONF(const char *conf){
	const char *dp,*cval;
	int i,iv,clen;
	Limit *lp;

	if( (dp = strchr(conf,'=')) ){
	    clen = dp - conf;
	    cval = &conf[clen + 1];
	    for( i = 0; limits[i].ival; i++ ){
		lp = &limits[i];
		if( strncasecmp(conf,lp->name,clen) == 0
		 || strncasecmp(conf,lp->abbr,clen) == 0
		){
			iv = atoi(cval);
			if( iv < lp->min )
				iv = lp->min;
			else
			if( lp->max < iv )
				iv = lp->max;
			if( 1 < lp->mul )
				*lp->ival = iv * lp->mul;
			else	*lp->ival = iv;

			fprintf(stderr,"*** %s = %d",lp->name,iv);
                        if( 1 < lp->mul )
				fprintf(stderr," * %d",lp->mul);
			fprintf(stderr,"\n");
		}
	    }
	}
}

extern int IndexSerno;
static double LapStart;
static double LapNext;
static double LapIntvl = 5;
double LapTime(){
	return LapNext - LapStart;
}
/*
void LapReport(int update, const char *fmt, int v1, int v2, int v3, int v4){
*/
void LapReport(int update, const char *fmt, ...){
	double now = Time();
	char *v[8];
	int dolap = 1;
	int ai;
	va_list ap;
	va_start(ap,fmt);
	for( ai = 0; ai < 8; ai++ )
		v[ai] = va_arg(ap,char*);

	if( LapNext == 0 ){
		LapStart = now;
		LapNext = now + LapIntvl;
	}
	if( !update || LapNext <= now ){
		if( strncmp(fmt,"%^",2) == 0 ){
			fmt += 2;
			dolap = 0;
		}
#ifdef _MSC_VER
		if( strstr(fmt,"ll") ){
			char xfmt[128];
			int fc;
			char *xfp;
			const char *xfx;
			const char *fp;
			xfp = xfmt;
			xfx =  &xfmt[sizeof(xfmt)-3];
			for( fp = fmt; fc = *fp; fp++ ){
				if( xfx <= xfp )
					break;
				*xfp++ = fc;
				if( fc != '%' ){
					continue;
				}
				while( isdigit(*++fp) )
					*xfp++ = *fp;
				if( fp[0] == 'l' && fp[1] == 'l' ){
					strcpy(xfp,"I64");
					xfp += 3;
					fp++;
				}else{
					*xfp++ = *fp;
				}
			}
			*xfp = 0;
			fmt = xfmt;
		}
#endif

		if( dolap ){
		if( IndexSerno )
			fprintf(stderr,"#%02d",IndexSerno-1);
		else	fprintf(stderr,"   ");
		if( update )
			fprintf(stderr,"*** ");
		else	fprintf(stderr,"    ");
		fprintf(stderr,"%7.3f ",now-LapStart);
		}

		/*
		fprintf(stderr,fmt,v1,v2,v3,v4);
		*/
		fprintf(stderr,fmt,v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7]);
		fprintf(stderr,"\n");
		if( update )
			LapNext += LapIntvl;
	}
}

#include "typedef.h"
#include "settemplate.h"
typedef struct {
	int   lsx;
	int   len;
	int   off;
	pos_t val;
} ListElem;
static int cmplist(ListElem **a, ListElem **b){
	pos_t av = (*a)->val, bv = (*b)->val;
	if( bv < av ) return 1;
	if( av < bv ) return -1;
	return 0;
}

void checkList(const char *where, pos_t *list, int len){
	int rev = 0;
	pos_t val;
	pos_t prev = 0;
	for(int i = 0; i < len; i++ ){
		val = list[i];
		if( val < prev ){
			rev++;
			break;
		}
		prev = val;
	}
	if( rev ){
		/*
		fprintf(stderr,"%s error %X len=%d\n",where,(int)list,len);
		 */
		fprintf(stderr,"%s error %X len=%d\n",where,(ArgInt)(PtrInt)list,len);
		prev = 0;
		for(int i = 0; i < len; i++ ){
			val = list[i];
			fprintf(stderr,"%s[%d] %08X %08X\n",
				(val<prev?"?":" "),i,(int)(val>>32),(int)val);
			prev = val;
		}
	}
}
typedef int (*sortfunc)(const void*,const void*);

int mergeLists(pos_t *dst, int dsiz, pos_t **src, int *lens, int lsiz){
	if(lsiz == 0)
		return 0;

	if(lsiz == 1){
		memcpy(dst,src[0],sizeof(pos_t)*lens[0]);
		return lens[0];
	}
	if(lsiz == 2){
		int len = mergeList(dst,src[0],lens[0],src[1],lens[1]);
		return len;
	}

	ListElem listsb[256], *lpb[256];
	ListElem *lists, **lp;
	if( lsiz+1 <= 256 ){
		lists = listsb;
		lp = lpb;
	}else{
		typedef ListElem *LEP;
		lists = new ListElem[lsiz+1];
		lp = new LEP[lsiz+1];
	}

	int i,j,act,minx,off;
	pos_t val,prev;
	int rev = 0, dup = 0;

	for(i = 0; i < lsiz; i++){
		lists[i].lsx = i;
		lists[i].len = lens[i];
		lists[i].off = 0;
		lists[i].val = src[i][0];
		lp[i] = &lists[i];
	}
	lists[lsiz].val = ~(pos_t)0;
	qsort(lp,lsiz,sizeof(ListElem*),(sortfunc)cmplist);

	int dx = 0;
	act = lsiz;
	prev = 0;
	for(i = 0; i < dsiz; i++){
		if( act == 1 ){
			minx = lp[0]->lsx;
			off = lists[minx].off;
			int len = lp[0]->len;
			for(j = off; j < len; j++){
				if( dsiz <= dx ){
					// error: overflow
					break;
				}
				dst[dx++] = src[minx][off++];
			}
			break;
		}
		val = lp[0]->val;
		if( val == prev ){
			dup++;
		}else{
			if( val < prev ){
				// error: mal-order
				rev++;
			}
			dst[dx++] = val;
		}
		prev = val;

		minx = lp[0]->lsx;
		off = (lists[minx].off += 1);
		if( lens[minx] <= off ){
			lists[minx].val = ~(pos_t)0;
			for(j = 1; j < act; j++)
				lp[j-1] = lp[j];
			if(--act <= 0){
				break;
			}
		}else{
			val = lists[minx].val = src[minx][off];
			if(lp[1]->val < val){
				for(j = 2; j < act; j++){
					if( val <= lp[j]->val ){
						break;
					}
				}
				ListElem *slp = lp[0];
				int k;
				for(k = 1; k < j; k++)
					lp[k-1] = lp[k];
				lp[k-1] = slp;
			}
		}
	}

/*
	if( dup || rev ){
*/
	if( rev ){
		fprintf(stderr,"*** mergeLists %d/%d / %d  dup=%d rev=%d\n",
			dx,dsiz,lsiz,dup,rev);
		for( int i = 0; i < lsiz; i++ ){
			fprintf(stderr,"####%d len=%d\n",i,lens[i]);
			checkList("mergeLists",src[i],lens[i]);
		}
	}
	if( lists != listsb ){
		delete []lists;
		delete []lp;
	}
	return dx;
}

//
// put and get integer in variable length
//

// valiable length integer list
// one octet per 7bits
// most significant bit first
// each byte to be continued is ORed with 0x80
// 0x80 represents end of list
#define EOL 0x80
typedef unsigned char UChar;
int sputintvl(UChar *out, int osiz, unsigned CODEINT ival[], int ilen){
	int tlen = 0;
	int ii;
	for( ii = 0; ii < ilen; ii++ ){
		UChar buf[16];
		UChar *op = &buf[sizeof(buf)];
		unsigned CODEINT val = ival[ii];
		int len;
		for(len = 0; len < osiz; len++){
			*--op = (val & 0x7F);
			if((val & ~0x7FL) == 0)
				break;
			val = (val >> 7);
		}
		for(int i = 0; i < len; i++)
			op[i] |= 0x80;
		len += 1;
		if(len <= (osiz-tlen)){
			memcpy(&out[tlen],(char*)op,len);
			tlen += len;
		}else{
			break;
		}
	}
	if(tlen < osiz)
		out[tlen] = EOL; /* END OF LIST */
	return tlen;
}
/*
int sgetintvl(UChar *in, int len, unsigned CODEINT oval[], int osiz){
*/
int sgetintvl(const UChar *in, int len, unsigned CODEINT oval[], int osiz){
	int si = 0;
	int oi;
	for( oi = 0; oi < osiz; oi++ ){
		unsigned CODEINT val = 0;
		UChar ch;
		int il;
		for(il= 0; il < len; il++){
			if(len <= si){
				goto EXIT;
			}
			ch = in[si++];
			if(ch == EOL){
				goto EXIT;
			}
			il++;
			val = (val << 7) | (ch & 0x7F);
			if((ch & 0x80) == 0){
				oval[oi] = val;
				break;
			}
		}
	}
EXIT:
	return oi;
}
static int compCODEINT(unsigned CODEINT *a,unsigned CODEINT *b){
	return *a - *b;
}
int saddintvl(UChar *dst,int dsiz,const UChar *src,int ssiz){
	unsigned CODEINT dv[32];
	int dx;

	if( dst[dsiz-1] == EOL )
		return 0;
	dx = sgetintvl(dst,dsiz,dv,32);
	dx += sgetintvl(src,ssiz,dv+dx,32);
	qsort(dv,dx,sizeof(dv[0]),(sortfunc)compCODEINT);
	sputintvl(dst,dsiz,dv,dx);
	return dx;
}

int delSpaces(char *str){
    int ndel = 0;
    char *dp = str;
    int top = 1;
    char ch,pch = 0;
    for(char *sp = str; (ch = *sp) != 0; sp++){
        if(top && ch == ' '){
            ndel++;
        }else{
            top = (ch == ' ')?1:0;
            *dp++ = pch = ch;
        }
    }
    if(pch == ' ')
        dp--;
    *dp = 0;
    return ndel;
}

int maskList(pos_t *list, int listLen, ShiftMask *mask){
	int ox = 0;
	int ix;
	for(ix = 0; ix < listLen; ix++){
		pos_t pos1 = list[ix];
		int docid = pos2did(pos1);
/*
printf("## docid=%2d shift=%2d = %2d newid=%2d typ=%X off=%d\n",
docid,mask->shift,docid+mask->shift,
mask->mask?mask->mask->get(docid+mask->shift):docid,
pos2type(pos1),pos2off(pos1));
*/
		int newid = docid;
		if(mask->shift)
			newid += mask->shift;
		if(mask->mask)
			newid = mask->mask->get(newid);
           
		if(newid == -1){
			continue;
		}
		if(docid != newid){
			pos1 = setposdid(pos1,newid);
			list[ox++] = pos1;
		}else
		if(ox != ix){
			list[ox++] = pos1;
		}else{
			ox++;
		}
	}
	if(ox != ix){
		listLen -= (ix - ox);
	}
	return listLen;
}

FILE *openCount(const char *base, const char *name, const char *mode){
	FILE *fp;
	char path[1024];
	sprintf(path,"%s.%s.vot",base,name);
	fp = fopen(path,mode);
	return fp;
}
CounterElem *
SX_incCount(const char *base, const char *name, int id, int inc, int uid){
	static CounterElem ce;
	FILE *fp;

	if((fp = openCount(base,name,"r+")) == 0){
		fp = openCount(base,name,"w+");
        }
	if(fp == NULL)
		return NULL;
	fseek(fp,id*sizeof(CounterElem),0);
	memset(&ce,0,sizeof(CounterElem));
	if(sizeof(CounterElem) == fread((char*)&ce,1,sizeof(CounterElem),fp)){
		if(uid != -1)
		if(uid == ce.c_lastmoder){
			if( time(NULL)-ce.c_lastmod < 10 ){
				// IGNORE repetitive votes.
				return &ce;
			}
		}
	}
        ce.c_acount += 1;
	ce.c_count += inc;
	ce.c_nmod++;
	ce.c_lastmod = time(NULL);
	ce.c_lastmoder = uid;
	fseek(fp,id*sizeof(CounterElem),0);
	fwrite((char*)&ce,1,sizeof(CounterElem),fp);
	return &ce;
}
CounterElem *
SX_loadCounts(const char *base, const char *name, int &len){
	FILE *fp;
	if((fp = openCount(base,name,"r")) == 0){
		len = 0;
		return NULL;
	}
	fseek(fp,0,2);
	int size = ftell(fp);
	CounterElem *cnt;
	cnt = (CounterElem*)mmap(NULL,size,PROT_READ,MAP_SHARED,fileno(fp),0);
	if( cnt != (CounterElem*)-1 ){
		len = size/sizeof(CounterElem);
		return cnt;
	}else{
		fclose(fp);
		len = -1;
		return NULL;
	}
}

FILE *fopenComm(const char *url, int create){
	FILE *fp;
	fp = fopen("/tmp/freyasx-comment","r+");
	if(fp == NULL && create)
		fp = fopen("/tmp/freyasx-comment","w+");
	return fp;
}
int putComment(const char *url, const char *from, const char *comment){
	FILE *fp = fopenComm(url,1);
	if(fp != NULL){
		if(from){
			/*
			char *dp;
			*/
			const char *dp;
			if((dp = strstr(from,".-."))){
				from = dp + 3;
			}
		}else{
			from = "-";
		}
		char date[64];
		strcpy(date,"-");
		time_t now = time(NULL);
		strftime(date,sizeof(date),"%Y/%m/%d-%H:%M:%S",localtime(&now));
		fprintf(fp,"%s[%s] %s\n",date,from,comment);
		fclose(fp);
	}
	return 0;
}
int getComment(const char *url, const char *from, char *comment, int size){
	FILE *fp = fopenComm(url,0);
	comment[0] = 0;
	if(fp != NULL){
		fgets(comment,size,fp);
		fclose(fp);
	}
	return 0;
}

#ifdef _MSC_VER
int optind;
char *optarg;
int getopt(int ac, char **av, const char *spec){
	char *opt,opc,*opa;
	optarg = 0;
	if( optind == 0 ){
		if( optind < ac )
			optind++;
		else	return EOF;
	}
	if( ac <= optind )
		return EOF;
	opt = av[optind];
	if( *opt != '-' )
		return EOF;
	opc = opt[1];
	optind++;
/*
fprintf(stderr,"#### ac=%d optind=%d [%s] sp[%s]\n",ac,optind,opt,spec);
*/
	if( (opa = (char*)strchr(spec,opc)) == 0 )
		return EOF;
	if( opa[1] != ':' ){
		return opc;
	}
	if( opa[1] == ':' ){
		if( ac <= optind ){
			return EOF;
		}
		optarg = av[optind++];
	}
	return opc;
}
#endif

char *kv_getval(KeyVals *kv, const char *key){
	for( int i = 0; i < kv->len; i++ ){
		if( strcmp(kv->keyval[i].key,key) == 0 )
			return kv->keyval[i].val;
	}
	return "";
}

KeyVals *xkv;
char *xgetrv(const char *key){
	char *val = kv_getval(xkv,key);
	return val;
}
void add_indexlist(int add, char *file,char *index,char *description){
	char lstfile[PATH_MAX];
	char savfile[PATH_MAX];

	strcpy(lstfile,file);
	char *mode;
	if( 0 < FileSize(lstfile) )
		mode = "r";
	else	mode = "w+";

	sprintf(savfile,"%s~",lstfile);
	Fstream oldl(lstfile,mode);
	if( oldl.good() ){
		oldl.close();
		Rename(lstfile,savfile);
	}
/*
	Fstream list(lstfile,mode);
*/
	Fstream list(savfile,mode);
	if( list.good() ){
		char line[256],idx[256],*dp;

		SX_tempName(lstfile);
		Fstream newl(lstfile,"w");
		for(;;){
			*line = 0;
			if( list.gets(line,sizeof(line)) == NULL )
				break;
			strcpy(idx,line);
			if( (dp = strpbrk(idx,"\t\r\n")) )
				*dp = 0;
			if( strcmp(index,idx) == 0 )
				continue;
			newl.puts(line);
		}
		if( add ){
			sprintf(line,"%s\t%s\n",index,description);
			newl.puts(line);
		}
		list.close();
		newl.close();
		SX_realName(lstfile);
	}
}

#define CRC32POLY 0x04C11DB7
inline int strCRC32X(int crc, const char *str){
	int bi,ovf;
	const char *s;
	signed char oct;

	for( s = str; (oct = *s) != 0; s++ ){
		for( bi = 0; bi < 8; bi++ ){
			ovf = (crc < 0) ^ (oct < 0);
			oct <<= 1;
			crc <<= 1;
			if( ovf ) crc ^= CRC32POLY;
		}
	}
	return crc;
}
int strCRC32(int crc, const char *str){
/*
	static int CRC32cache[256];
	int ci;
	int c0;
	if( CRC32cache[1] == 0 ){
		char buf[2];
		buf[1] = 0;
		for( ci = 0; ci < 256; ci++ ){
			buf[0] = ci;
			CRC32cache[ci] = strCRC32X(0,buf);
		}
	}
	if( (c0 = (0xFF & *str)) ){
		crc = CRC32cache[c0];
		str++;
	}
*/
	return strCRC32X(crc,str);
}

#define CRC8PORY 0x8C
int strCRC8(int crc,const char *str){
	int oi,bi,ovf;
	const char *s;
	char oct;

	s = str;
	for( oi = 0; (oct = *s++); oi++ ){
		for( bi = 0; bi < 8; bi++ ){
			ovf = (crc ^ oct ) & 1;
			crc >>= 1;     
			oct >>= 1;
			if( ovf ) crc ^= CRC8PORY;
		}
	}
	return crc;
}

#define CRC8C 0x8C;
int strCRC8(int crc, const char *str, int len){
	int oi,bi,ovf;
	char oct;
	const char *s;

	s = str;
	for( oi = 0; oi < len; oi++ ){
		oct = *s++;
		for( bi = 0; bi < 8; bi++ ){
			ovf = (crc ^ oct ) & 1;
			crc >>= 1;     
			oct >>= 1;
			if( ovf ) crc ^= CRC8C;
		}
	}
	return crc;
}

Strings *stringBank;

Strings::Strings(){
	int i;
	for( i = 0; i < 256; i++ ){
		sx = 0;
		strid = 0;
		strbank_buf[i] = 0;
		strbank_siz[i] = 0;
		strbank_rem[i] = 0;
	}
}
void Strings::reset(){
	int i;

fprintf(stderr,"------- Strings reset: sx=%d strid=%d freed=%d\n",
	sx,strid,strfreed);

	for( i = 0; i < sx; i++ ){
/*
fprintf(stderr,"Strings clear[%d/%d]\n",i,sx);
*/
		free(strbank_buf[i]);
		strbank_buf[i] = 0;
	}
	sx = 0;
	strfreed = 0;
	strid = 0;
}
Strings::~Strings(){
	int i;
	for( i = 0; i < sx; i++ ){
/*
fprintf(stderr,"Strings free[%d/%d]\n",i,sx);
*/
		free(strbank_buf[i]);
		strbank_buf[i] = 0;
	}
	sx = 0;
}
char *Strings::Newstring(const char *str,int len){
	char *newstr;
	if( strbank_buf[sx] != 0 ){
		if( strbank_rem[sx] < len+1 ){
			sx++;
		}
	}
	if( strbank_buf[sx] == 0 ){
		strbank_siz[sx] = 1024*1024;
		strbank_rem[sx] = 1024*1024;
		strbank_buf[sx] = (char*)malloc(1024*1024);
	}
	newstr = strbank_buf[sx]+strbank_siz[sx]-strbank_rem[sx];
	memcpy(newstr,str,len);
	newstr[len] = 0;
	strbank_rem[sx] -= len+1;
/*
if( strid % 10000 == 0 )
fprintf(stderr,"---- string %d %d %X\n",sx,strid,(int)newstr);
*/
	return newstr;
}


static int xids[8];
static char *keys[8];
static int kids[8];
static int klen[8];
static int keyx;
void h_addkey(int xid,int kid,const char *key){
	int ki;
	if( 8 <= keyx )
		return;

	for( ki = 0; ki < keyx; ki++ ){
		if( (0 < kid && xids[ki] == xid && kids[ki] == kid)
		 || strcmp(keys[ki],key) == 0 ){
			return;
		}
	}
//fprintf(stderr,"---- ADDKEY %d %d %s\n",xid,kid,key);
	xids[keyx] = xid;
	keys[keyx] = strdup(key);
	klen[keyx] = strlen(key);
	kids[keyx] = kid;
	keyx++;
}
int h_matchkey(int xid,int kid,const char *key){
	int ki;
	for( ki = 0; ki < keyx; ki++ ){
		if( (0 < kid && xids[ki] == xid && kids[ki] == kid)
		 || strcmp(keys[ki],key) == 0
/*
		 || strncmp(keys[ki],key,klen[ki]) == 0
*/
		){
//fprintf(stderr,"---- MATCHKEY %d %d %s\n",xid,kid,key);
			return 1;
		}
	}
	return 0;
}
int h_hilight(char *text,int size){
	char buff[1024];
	const char *sp;
	char ch;
	char *dp;
	char *dx;
	int ri;
	const char *key;
	int len;

	if( *text == 0 )
		return 0;

/*
printf("--[%d][%s]\n",keyx,text);
for(int i=0;i<keyx;i++)
printf("==[%d][%s]\n",i,keys[i]);
*/

	Strncpy(buff,text,sizeof(buff));
	dp = text;
	dx = &text[size-1];

	for( sp = buff; (ch = *sp++); ){
		if( dx <= dp ){
			break;
		}
		for( ri = 0; ri < keyx; ri++ ){
			key = keys[ri];
			len = klen[ri];
			if( *key == ch && strncmp(sp,key+1,len-1) == 0 ){
				if( dx <= dp+64 ){
					break;
				}
			/* should use snprintf() */
			sprintf(dp,"<FONT color=#303080>%s</FONT>",key);
				dp += strlen(dp);
				sp += len-1;
				goto Next;
			}
		}
		*dp++ = ch;
		if( ch & 0x80 ){
			*dp++ = *sp++;
		}
	Next:;
	}
	*dp = 0;
/*
printf("++[%d][%s]\n",keyx,text);
*/
	return 0;
}
