// Copyright (c) 2004 Yutaka Sato. All rights reserved.

#ifndef SX_H_INCLUDED
#define SX_H_INCLUDED

extern "C" {
#include "SXc.h"
typedef int (*replaceFunc)(char*,const char*,const char*);
void SX_setConv(char *pfx);
void SX_deConv(replaceFunc,const char *k,char *v);
void SX_enConv(replaceFunc,char *linebuf);
}
int strCRC8(int crc, const char *str, int len);
int strCRC8(int crc,const char *str);
int strCRC32(int crc, const char *str);

#include "config.h"
/*
#include <iostream.h>
#include <iomanip.h>
*/
#include <iostream>
using namespace std; /* mod-140617e */
#include <iomanip>
#include <math.h>
void setAsyncIO();
int Rename(const char *, const char *);
int fileReadable(const char *path);

#define INLINE CPPINLINE
#include "Coder.h"

int sputintvl(unsigned char *out, int siz, pos_t ival[], int ilen);
int sgetintvl(const unsigned char *in, int len, pos_t oval[], int osiz);
int saddintvl(unsigned char *dst,int dsiz,const unsigned char *src,int ssiz);

inline int frompos(pos_t pos, urlid_t &urlid, size_t &offset, int &type){
	return _frompos(pos, &urlid, &offset, &type);
}
inline pos_t toPos(ElemCtx &ectx, pos_t offset){
	ectx.c_words_total++;
	return topos(ectx.c_docid, ectx.c_type, ectx.c_words);
}

typedef struct {
	int	c_acount;  // access counter
	int	c_lastmoder:8, // CRC8 of the owner of the last change
		c_nmod:24;    // the number of updates (votes)
	int	c_lastmod; // the date of the last change
	int	c_count;   // the value of counter
} CounterElem;
/*
#define c_acount   c_ints[0]
#define c_lastmodc c_ints[1]
#define c_lastmodt c_ints[2]
#define c_count    c_ints[3]
*/

CounterElem *SX_incCount(const char *base, const char *name, int id, int inc, int uid);
CounterElem *SX_loadCounts(const char *base, const char *name, int &len);


class Fstream {
private:
	char *fpath;
	FILE *fp;
	int rcc;
	int wcc;
        int dounlink;
	off_t offg;
	off_t offp;
	int ntellg;
	int ntellp;
public:
	void open(const char *path, const char *mode){
		fpath = new char[strlen(path)+1];
		strcpy(fpath,path);
		fp = Fopen(path,mode);
	}
/*
	Fstream(){ fpath = 0; fp = 0; rcc = -1; rcc = -1; dounlink = 0; }
*/
	Fstream(){
		fpath = 0; fp = 0; rcc = -1; rcc = -1; dounlink = 0;
		ntellg = ntellp = 0;
		offp = -1;
	}
	Fstream(const char *path, const char *mode){
		open(path,mode);
		rcc = -1;
		wcc = -1;
                dounlink = 0;
		ntellg = ntellp = 0; offp = -1;
		if( fp != NULL ) offp = Ftell(fp);
	}
	~Fstream(){
/*
fprintf(stderr,"### Fstream close %s %X %X\n",fpath?fpath:"",(int)&fp,(int)fp);
*/
		if(fp){
			fclose(fp);
			if( dounlink ){
/*
fprintf(stderr,"### Fstream unlink %s\n",fpath?fpath:"");
*/
				unlink((const char*)fpath);
			}
			delete []fpath;
		}
	};
	bool is_open(){ return fp != 0; }
/*
	int size(){ return Fsize(fp); };
*/
	off_t size(){ return Fsize(fp); };
        int remove(int on){ dounlink = on; return 0; }
        FILE *file(){ return fp; }
/*
	int write(char *buf, int siz){ return wcc = Fwrite(fp,buf,siz); }
*/
	int write(char *buf, int siz){
		wcc = Fwrite(fp,buf,siz);
		if( 0 < wcc ) offp += wcc;
		return wcc;
	}
	int read(char *buf, int siz){ return rcc = Fread(fp,buf,siz); }
        char *gets(char *buf, int siz){ return fgets(buf,siz,fp); }
/*
        int puts(char *buf){ return fputs(buf,fp); }
*/
        int puts(char *buf){
		int rcode;
		rcode = fputs(buf,fp);
		if( rcode == 0 ) offp += strlen(buf);
		return rcode;
	}
	int gcount(){ return rcc; }
	int pcount(){ return wcc; }
	int eof(){ return Feof(fp); }
	int fail(){ return Feof(fp); }
        void putback(int ch){
		ungetc(ch,fp);
	}
	void seekg(off_t off){
		Fseek(fp,off,0);
	} 
	void seekp(off_t off){
		Fseek(fp,off,0);
		offp = Ftell(fp);
	} 
	void seekg(off_t off, int type){
		if(type == ios::beg) Fseek(fp,off,0);
		if(type == ios::cur) Fseek(fp,off,1);
		if(type == ios::end) Fseek(fp,off,2);
	} 
	void rewind(){ fflush(fp); Fseek(fp,0,0); offp = 0; }
	off_t tellp(){

if( ntellp++ % 16*1024 == 0 ){ /* check the consistency ... */
	off_t off;
	off = Ftell(fp);
	if( off != offp ){
		fprintf(stderr,"--- ERROR tellp %d: Ftell()=%llX offp=%llX\n",
			ntellp,(unsigned CODEINT)off,(unsigned CODEINT)offp);
		sleep(1);
	}
}
		return offp;
	}
	off_t tellg(){
		ntellg++;
/*
if( ntellg % 1024 == 0 )
fprintf(stderr,"--- tellg() %d\n",ntellg);
*/
		return Ftell(fp);
	}
	off_t tellG(){
		if( feof(fp) )
			clearerr(fp);
		return Ftell(fp);
	}
        void flush(){ fflush(fp); }
	void clear(){ clearerr(fp); } 
        void close(){ if(fp) fclose(fp); fp = 0; }
        void truncate(){ if(fp) Ftruncate(fp); }
	int good(){ return fp != 0; }
        int fd(){ if(fp != NULL) return fileno(fp); else return -1; }
	int Fputintvl(pos_t *val){
		int len;
		len = fputintvl(fp,val);
		offp += len;
		return len;
	}
	int Fputintvls(const pos_t *list, int listLen){
		int len;
		len = fputintvls(fp,list,listLen);
		offp += len;
		return len;
	}
};

class Mask {
private:
	char *mask;
	unsigned int size;
	int numset;
public:
	Mask(unsigned int n){
		size = n;
		int bsize = (n+7)/8;
		mask = new char[bsize];
		memset(mask,0,bsize);
		numset = 0;
	}
	~Mask(){ delete []mask; }
	void set(unsigned int n){
		if(n < size){ mask[n/8] |= (1 << (n%8)); numset++; }
	}
	int is_set(unsigned int n){
		if(n < size)
			return mask[n/8] & (1 << (n%8));
		else	return -1;
	}
	int num_set(){ return numset; }
};

class IMask {
private:
	int *mask;
	unsigned int size;
public:
	IMask(unsigned int n){
		size = n;
		mask = new int[size];
		memset(mask,0,size*sizeof(int));
	}
	~IMask(){ delete []mask; }
	void set(unsigned int n, int m){
		if(n < size){
			mask[n] = m;
		}
	}
	int get(unsigned int n){
		if(n < size)
			return mask[n];
		else	return 0;
	}
        int getsize(){ return size; }
};

typedef struct {
    int shift;
    IMask *mask;
} ShiftMask; 
int maskList(pos_t *list, int listLen, ShiftMask *mask);

class Files {
private:
    char *_base;
    char **descs;
    char **exts;
    char **paths;
    int n;
public:
    Files(int size, const char *b):n(0){
        _base = strdup(b);
        descs = new char *[size];
        exts = new char *[size];
        paths = new char *[size];
    }
    void add(char *desc,char *ext){
        descs[n] = desc;
        exts[n] = ext;
        n++;
    }
    char *base(){ return _base; }
    int num(){
        return n;
    }
    char *desc(int i){ return descs[i]; }
    char *ext(int i){ return exts[i]; }
    void renames();
};

#define DESCBLOCKSIZE 256
#define D_NINTS 4
typedef struct {
    int  d_ints[D_NINTS];
    char d_url[DESCBLOCKSIZE-D_NINTS*sizeof(int)];
} FileDescRecord;
#define d_size    d_ints[0]
#define d_date    d_ints[1]
#define d_lastmod d_ints[2]
#define d_crc32   d_ints[3] // CRC of content
void writeRecord(Fstream *descf, int size, int date, int lastmod,
                int crc32, const char *url, const char *author,
                const char *title, const char *digest);

#define NUM_EXT	".num"
#define N_URLCRC	0
#define N_CONTCRC	1
#define N_SIZE		2
#define N_DATE		3
#define N_LASTMOD	4
#define N_USORT		5
#define N_NUMCRCS	6
class NumFile {
private:
	FILE *fps[8];
	FILE *fpr;
	int *vals[8];
	int nelem;
	int bsize;
public:
	NumFile(const char *path, int size);
	NumFile(const char *path);
	~NumFile();
	int write(int did, FileDescRecord &rec, int sort);
	void flush(){ for(int i = 0; i < N_NUMCRCS; i++) fflush(fps[i]); }
	int good(){ return fpr != NULL; }
	int get(int did, int cid);
	int size(){ return nelem; }
};

// true if s1 is longer thatn s2
inline int strlonger(const char *s1, const char *s2){
	for(;;){
		if( *s1++ == 0 ) return 0;
		if( *s2++ == 0 ) return 1;
	}
	return 0;
}

typedef struct {
	char *key;
	char *val;
} KeyVal;
typedef struct {
	int len;
	KeyVal *keyval;
} KeyVals;

class Strings{
private:
	int   sx;
	char *strbank_buf[256];
	int   strbank_siz[256];
	int   strbank_rem[256];
	int   strid;
	int   strfreed;
	char *Newstring(const char *str,int len);
public:
	Strings();
	~Strings();
	void reset();
	char *newstring(const char *str,int len){
		strid++;
		return Newstring(str,len);
	}
	char *newstring(const char *str){
		return newstring(str,strlen(str));
	}
	int freestring(const char *str){
		strfreed++;
		return 0;
	}
};
extern Strings *stringBank;

#endif // SX_H_INCLUDED
