// Copyright (c) 2004 Yutaka Sato. All rights reserved.
// SXc.c -- extensions for FreyaSX
//
#include "SXc.h"
#define INLINE
#include "Coder.h"

#ifdef _MSC_VER
#define PosOf(p) *(__int64*)(&p)
#else
#define PosOf(p) *(off_t*)(&p)
#endif

off_t Fseek(FILE *fp,off_t off,int whence){
	fpos_t pos;
	PosOf(pos) = off;
	if( fsetpos(fp,&pos) == 0 )
		return 0;
	else	return -1;
}
off_t Ftell(FILE *fp){
	fpos_t pos;
	if( fgetpos(fp,&pos) != 0 ){
		return -1;
	}else{
		return PosOf(pos);
	}
}
/*
int Fsize(FILE *fp){
	int off,size;
	off = ftell(fp);
	fseek(fp,0,2);
	size = ftell(fp) - off;
	fseek(fp,off,0);
	return size;
}
*/
off_t Fsize(FILE *fp){
	off_t off,size;
	off = Ftell(fp);
	fseek(fp,0,2);
	size = Ftell(fp) - off;
	Fseek(fp,off,0);
	return size;
}
FILE *Fopen(const char *path, const char *mode){
	FILE *fp;
	setBinaryIO();
	fp = fopen(path,mode);
/*
	if( fp != NULL && 0 < IoBUFFERSIZE ){
		void setbuffer(FILE *iop, char *abuf, size_t asize);
		char *buf = malloc(IoBUFFERSIZE);
		setbuffer(fp,buf,IoBUFFERSIZE);
	}
*/
	return fp;
}
int Feof(FILE *fp){
	return feof(fp);
}
int Fread(FILE *fp, char *buf, int siz){
	return fread(buf,1,siz,fp);
}
int Fwrite(FILE *fp, char *buf, int siz){
	return fwrite(buf,1,siz,fp);
}
int Fgetline(FILE *fp, char *buf, int siz, char del){
	buf[0] = 0;
	if(del == '\n'){
		char *dp;
		fgets(buf,siz,fp);
		if(( dp = strchr(buf,'\n') ))
			*dp = 0;
	}else{
	}
	return strlen(buf);
}
int Ftruncate(FILE *fp){
	int rcode;
	fflush(fp);
	rcode = ftruncate(fileno(fp),(off_t)ftell(fp));
	return rcode;
}

CINLINE
int fputintvl1(FILE *out, unsigned CODEINT *ival){
	unsigned CODEINT val = *ival;
	char buf[16];
	char *ox = &buf[sizeof(buf)];
	char *op = ox;
	int len;
	*--op = (val & 0x7F);
	val >>= 7;
	while(val){
		*--op = 0x80|(val & 0x7F);
		val >>= 7;
	}
	len = ox - op;
	fwrite(op,1,len,out);
	return len;
}
CINLINE
int fgetintvl1(FILE *in, unsigned CODEINT *valp){
	unsigned CODEINT val = 0;
	int ch;
	int len;
	for(len = 0; len < 10;){
		ch = getc(in);
		if(ch == EOF)
			break;
		len++;
		val = (val << 7) | (ch & 0x7F);
		if((ch & 0x80) == 0)
			break;
	}
	*valp = val;
	return len;
}
int fputintvl(FILE *out, unsigned CODEINT *ival){
	return fputintvl1(out,ival);
}
int fgetintvl(FILE *in, unsigned CODEINT *valp){
	return fgetintvl1(in, valp);
}

int fputintvls(FILE *file, const pos_t *list, int listLen){
	int i;
	pos_t prev = 0;
	int plen = 0;

	pos_t Clen = listLen;
	fputintvl1(file,&Clen);
	for(i = 0; i < listLen; i++){
		pos_t elem = EnPACK(list[i]);
		pos_t diff = elem - prev;
		prev = elem;
		plen += fputintvl1(file,&diff);
	}
	return plen;
}
pos_t *fgetintvls(FILE *file, int *listLen){
	int i,lLen;
	pos_t len;
	pos_t prev = 0;
	pos_t *list;

	fgetintvl1(file, &len);
	lLen = len;
	*listLen = len;
	list = (pos_t *)malloc(lLen*sizeof(pos_t));

	for(i = 0; i < lLen; i++){
		pos_t diff;
		fgetintvl1(file,&diff);
		prev = diff + prev;
		list[i] = DePACK(prev);
	}
	return list;
}

// conversion table file "ja.cnv" is used if "pfx=ja"
// each line of the file consists of InRep TAB* ExRep
// HTML output is converted from InRep to ExRep
// URL input is converted from ExRep to InRep
// the most naive implementation of text converter.
//
#define CONV_MAX 64
typedef struct {
	char *in; // Internal representation
	char *ex; // External representation
} Conv;
static Conv conv[CONV_MAX];
static int nconv = 0;
int SX_setConv(char *cnv){
	char path[PATH_MAX];
	char cnv1[LINE_MAX],*dp;
	FILE *fp;

	Strncpy(path,cnv,sizeof(path)-4);
	strcat(path,".cnv");
	fp = fopen(path,"r");
	if( fp == NULL )
		return -1;

	while( nconv < CONV_MAX ){
		if( fgets(cnv1,sizeof(cnv1),fp) == NULL )
			break;
		if( (dp = strpbrk(cnv1,"\r\n")) )
			*dp = 0;
		if(cnv1[0] == '#')
			continue;
		if((dp = strchr(cnv1,'\t')) != 0){
			*dp++ = 0;
			while(*dp == '\t')
				dp++;
			if(*dp != 0){
				conv[nconv].in = strdup(cnv1);
				conv[nconv].ex = strdup(dp);
				nconv++;
			}
		}
	}
	fclose(fp);
	return 0;
}
typedef int (*replaceFunc)(char*,const char*,const char*);
void SX_deConv(replaceFunc replace, const char *key, char *ex){
	int i;
	for(i = 0; i < nconv; i++)
		(*replace)(ex, conv[i].ex, conv[i].in);
}
void SX_enConv(replaceFunc replace, char *in){
	int i;
	for(i = 0; i < nconv; i++)
		(*replace)(in, conv[i].in, conv[i].ex);
}

#ifdef _MSC_VER
/* extracted from DeleGate/maker/strcasecmp.c */
strncasecmp(a,b,n)
	const char *a,*b;
	size_t n;
{	char ac,bc;
	int i;

	for( i = 1; i < n; i++ ){
		ac = *a++;
		bc = *b++;

		if(ac == 0){
			if(bc == 0)
				return 0;
			else	return -1;
		}else
		if(bc == 0)
			return 1;
		else
		if(ac != bc){
			if(islower(ac)) ac = toupper(ac);
			if(islower(bc)) bc = toupper(bc);
			if( ac != bc )
				return ac - bc;
		}
	}
	if(islower(*a)) ac = toupper(*a); else ac = *a;
	if(islower(*b)) bc = toupper(*b); else bc = *b;
	return ac - bc;
}
strcasecmp(a,b)
	const char *a,*b;
{
	return strncasecmp(a,b,0xFFFFFFF);
}

/* extracted from DeleGate/rary/windows.c */
#include <sys/timeb.h>
gettimeofday(tp, tzp)
	struct timeval *tp;
	struct timezone *tzp;
{	struct timeb timeb;

	ftime(&timeb);
	tp->tv_sec = timeb.time;
	tp->tv_usec = timeb.millitm * 1000;
	return 0;
}

/* extracted from DeleGate/maker/__locking.c */
#include <sys/locking.h>
static Locking(fd,type,bytes)
{ 	int off,rcode;

	rcode = -1;
	if( 0 <= (off = tell(fd)) )
	if( 0 <= lseek(fd,0,0) ){
		rcode = locking(fd,type,bytes);
		lseek(fd,off,0);
	}
	return rcode;
}
EXlockB(fd){	return	Locking(fd,_LK_LOCK,1); }
UNlock(fd){	return	Locking(fd,_LK_UNLCK,1); }


int ftruncate(int fd, off_t len){
	return -1;
}

void *mmap(void *ad, size_t le, int pr, int fl, int fd, off_t of){
	void *addr;
	HANDLE fh,fmh;
	fh = (HANDLE)_get_osfhandle(fd);
	fmh = CreateFileMapping(fh,NULL,PAGE_READONLY,0,le,NULL);
	addr = MapViewOfFile(fmh,FILE_MAP_READ,0,of,le);
	return addr;
}
int munmap(void *ad, size_t le){
	int ok = UnmapViewOfFile(ad);
	return ok ? 0 : -1;
}

/* fix-140708b for findex.cc from non-Windows-Console terminal */
int isaTty(int fd){
	HANDLE ph;
	int nready;
	char *env = getenv("SESSIONNAME");
	ph = (HANDLE)_get_osfhandle(fd);

	if( env != NULL && strcmp(env,"Console") == 0 /* fix-140502h */
	 || getenv("LANG") /* mod-140511c seems not set under a service */
	 || getenv("SSH_CONNECTION")
	 || getenv("CYGWIN")
	 || getenv("YYSH")
	 || getenv("YYEXEC")
	)
	/* Unix like pty is seen as PIPE, not FILE_TYPE_CHAR  */
	if( GetFileType(ph) == FILE_TYPE_PIPE )
	{
		if( PeekNamedPipe(ph,NULL,0,NULL,&nready,NULL) == TRUE ){
			if( nready == 0 ){
				if( fd == fileno(stdin)
				 || fd == fileno(stdout)
				 || fd == fileno(stderr)
				 ){
					return 1;
				}
			}
		}else{
		}
	}

	return _isatty(fd);
}
#endif

