// Copyright (c) 2004 Yutaka Sato. All rights reserved.

#include "SX.h"
#include <sstream>
#include <string>
using namespace std;

#include "typedef.h"
#include "config.h"

typedef int (*replaceFunc)(char*,const char*,const char*);
int pubFileOK(const char *path, char *xpath, int xsize);
extern "C" {
char *Strncpy(char *s1, const char *s2, int n);
void SX_enConv(replaceFunc,char *linebuf);
double Time();
}

char *escape_url(const char *url);
const char *html_entitize(char *buf, const char *src);
int is_valid_kanji(const char *kp);

int replace_string(char *buf, const char *keyword, const char *str);
int SX_pfxPath(char *pfx, const char *filepath, char *xpath, int xsize);
void SX_display(char *line, int size);
const char *html_entitize(char *buf, const char *src);

extern int param_from;
extern int param_n;
static char elapsedTimes[16];
extern char resultFroms[16];
extern char resultTos[16];

int param_zoom;
char param_index[NAME_MAX+1];
char param_style[16];
char param_pfx[16];

double Start;
stringstream shead;
stringstream sout;
stringstream dbgout;
char ContentType[64];

// $include{filename}
// shuold support $include(n1=v1,n2=v2,...){filename}
//
int SX_include(char *name, int nsiz, char *linebuf){
	char *inc = 0;
	for(char *p = linebuf; *p; p++){
		if(*p == '\\'){
			if( p[1] )
				p++;
			continue;
		}
		if(*p == '$'){
			if( strncmp(p,"$include{",9) == 0 ){
				inc = p;
				break;
			}
		}
	}
	if( inc == NULL )
		return 0;

	char *top = inc+9;
	char *end = strstr(top,"}");
	if( end == NULL )
		return 0;

	int len = end - top;
	char *sp;
	for(sp = linebuf; sp < inc && *sp != 0; sp++)
		sout << *sp;

	if( nsiz <= len )
		len = nsiz - 1;
	Strncpy(name,top,len+1);
	char xpath[256];
	if( pubFileOK(name,xpath,sizeof(xpath)) < 0 ){
		sout << "<!-- ## file not found ## -->";
	}
	char *dp = linebuf;
	for(sp = end+1; (*dp++ = *sp++) != 0;);
	return 1;
}

void sout_line(char *line){
	sout << line << endl;
}

void putMoved(char *url){
	shead << "Status: 302 Moved\r\n";
	shead << "Location: " << url << "\r\n";
}

void putNotAuthorized(){
	shead << "Status: 401 Not Authorized\r\n";
	shead << "WWW-Authenticate: Digest realm=editlist\r\n";
}

void putError(const char *msg1, const char *msg2){
        sout << "<HTML><HEAD><TITLE>" << msg1
             << "</TITLE></HEAD>\n<BODY><H1>" << msg2
             << "</H1></BODY>\n</HTML>\n";
}

void scan_META(char *line, int size){
	char *dp;
	if( (dp = strstr(line,"<META ")) == NULL )
		return;
	dp += 6;
	if( strncasecmp(dp,"HTTP-EQUIV=",11) != 0 )
		return;
	dp += 11;
	if( *dp == '"' ) dp++;
	if( strncasecmp(dp,"Content-Type",12) != 0 )
		return;
	dp += 12;
	if( *dp == '"' ) dp++;
	if( *dp == ' ' ) dp++;
	if( strncasecmp(dp,"CONTENT=",8) != 0 )
		return;
	dp += 8;
	if( *dp == '"' ) dp++;
	Strncpy(ContentType,dp,sizeof(ContentType));
	if( (dp = strpbrk(ContentType,"\">")) )
		*dp = 0;
	*line = 0;
// fprintf(stderr,"Content-Type: %s\n",ContentType);
}

#ifndef NOSTRINGSTREAM
#if defined(__GNUC__) && __GNUC__ < 3
#define NOSTRINGSTREAM 1
#else
#define NOSTRINGSTREAM 0
#endif
#endif

static int flushed;
void sout_flush(){
    if(flushed)
        return;
    flushed = 1;

    char params[256];
    params[0] = 0;

    sout << '\0';

    int ic;
/*
    if(sout.tellp() <= 0){ // stringstream seems be implemented incompletely
*/
    if(NOSTRINGSTREAM || sout.tellp() <= 0){
        int ch;
        for(ic = 0;;ic++){
            ch = sout.str().c_str()[ic];
            if(ch == 0 || ch == EOF)
                break;
            if(ch & 0x80){
                char buf[3];
                buf[0] = ch;
                buf[1] = sout.str().c_str()[ic+1];
                buf[2] = 0;
                if(is_valid_kanji(buf)){
                    strcpy(params,"; charset=EUC-JP");
                }
                break;
            }
        }
    }else{
        sout.seekg(0);
        int ch;
        for(ic = 0; (ch = sout.get()) != EOF; ic++){
            if(sout.eof())
                break;
            if(ch & 0x80){
                char buf[3];
                buf[0] = ch;
                buf[1] = sout.get();
                buf[2] = 0;
                if(is_valid_kanji(buf)){
                    strcpy(params,"; charset=EUC-JP");
                }
                break;
            }
        }
/*
        sout.clear();
        sout.seekg(0);
        while((ch = sout.get()) != EOF)
            putc(ch,stdout);
*/
    }
    fputs(shead.str().c_str(),stdout);
/*
    fprintf(stdout,"Content-Type: text/html%s\r\n",params);
*/
    // ContentType might be "text/shtml" to be interpreted by DeleGate
    fprintf(stdout,"Content-Type: %s%s\r\n",ContentType,params);
    fprintf(stdout,"\r\n");
    if(params[0]){
        // the content type in META should be text/html
        fprintf(stdout,"<META HTTP-EQUIV=Content-Type CONTENT=\"%s%s\">\n",
            "text/html",params);
/*
            ContentType,params);
*/
    }
    fputs(sout.str().c_str(),stdout);
    fflush(stdout);
    printf("<!-- (%f seconds) -->\n",Time()-Start);
    fputs(dbgout.str().c_str(),stdout);
}

void setrv(const char *name, const char *val);
char *getrv(const char *name);
char *getrvx(const char *name);

void makeParams(const char *query, int from){
	stringstream params;
	while(*query == ' ')
		query++;

	if(query[0]){
		char qquery[LINE_MAX],*dp;
		char *equery = escape_url(query);
		html_entitize(qquery, equery);
		delete []equery;
		for(dp = qquery; *dp; dp++)
			if(*dp == ' ')
				*dp = '+';
		if(qquery[0])
			params << "key="   << qquery << "&";
	}

	char *rv;
	if(from){
		if( (rv = getrvx("docid")) )
			params << "docid=" << rv << "&";
		else
		if( 1 < param_from )
			params << "from="  << param_from-1 << "&";
	}
	if( (rv = getrvx("pfx")) )
		params << "pfx="   << rv << "&";
	if( (rv = getrvx("sxop")) && strcmp(rv,"AND") != 0 )
		params << "sxop="  << rv << "&";
	if( (rv = getrvx("sort")) && strcmp(rv,"score") != 0 )
		params << "sort="  << rv << "&";
	if( (rv = getrvx("style")) && strcmp(rv,"html") != 0 )
		params << "style=" << rv << "&";

	if( param_zoom != 0 )
		params << "zoom="  << param_zoom << "&";
	if( param_n && param_n != RESULTSFORDISPLAY )
		params << "n=" << param_n << "&";

	if( (rv = getrvx("escindex")) )
		params << "index=" << rv;

	setrv("params",params.str().c_str());
}

int printGeneric(){
	sout
	<< "<P>" << getrv("num") << " "
	<< "<A HREF=\"" << getrv("escurl") << "\">" << getrv("title") << "</A>"
	<< "<small>"
	<< " " << getrv("author")
	<< "<BR>" << endl
	<< getrv("sdate") << " " << getrv("digest")
	<< " ... " << getrv("context")
	<< "</small>"
	<< endl;
	return 0;
}

static void
makeNext1(stringstream &next, const char *cgipath, const char *params,
          int disp, char *esckey, int from, char *label, char *icon=NULL){

    if(label[0] == 0){
        next << cgipath
             << "?"
             << params
             << "&from=" << from;
        return;
    }
    next << "<A "
         << "HREF=\"" << cgipath
         << "?"
         << params
         << "&from="  << from
         << "\">";
    if(isdigit(label[0])){
        next << label << "</A>" << endl;
    }else{
        if(icon){
        next << "<IMG SRC=" << icon
             << " ALT=\"" << label << "\" BORDER=0 ALIGN=middle></A>" << endl;
        }else{
        next << "[" << label << "]"
             << "</A>" << endl;
        }
    }
}

int makeNext(stringstream &next,
             const char *key, int resultNum, int disp, int from, int to){
    char *cgi = getrv("cgipath");
    char *params = getrv("params");

    int From1 = 0;
    int To1 = 0;
    char *escaped_key = escape_url(key);
    if(from > 0){
        int from1 = ((from - disp < 0) ? 0 : from - disp);
        From1 = from1;
    }
    int i, p;
    int i_from = from - 5 * disp;
    if(i_from < 0) i_from = 0;
    int i_to = i_from + 10 * disp;
    if(i_to > resultNum) i_to = resultNum;

    for(i = i_from,p = i / disp; i < i_to; i += disp, p++){
        if(i == from)
            next << "<B>" << p << "</B>" << endl;
        else{
            char label[32];
            sprintf(label,"%d",p);
            makeNext1(next,cgi,params,disp,escaped_key, i, label);
        }
    }
    if(to + 1 < resultNum){
        To1 = to + 1;
    }

    next << "&nbsp;";
    makeNext1(next,cgi,params,disp,escaped_key, 0, "[|<]", "tops.gif");
    int j_from = from - 10 * disp;
    if(j_from < 0) j_from = 0;
    makeNext1(next,cgi,params,disp,escaped_key, j_from, "[<<]", "prev2s.gif");
    next << "<B><BIG>";
    makeNext1(next,cgi,params,disp,escaped_key, From1, "[<-]", "prev1s.gif");
    next << "</BIG></B>";
    int prevh = from-disp/2;
    if(prevh < 0)
        prevh = 0;
    makeNext1(next,cgi,params,disp,escaped_key, prevh, "[<=]","prevhs.gif");
    int nexth = from+disp/2;
    if(resultNum < nexth)
        nexth = resultNum-1;
    makeNext1(next,cgi,params,disp,escaped_key, nexth, "[=>]","nexths.gif");
    next << "<B><BIG>";
    makeNext1(next,cgi,params,disp,escaped_key, To1, "[->]", "next1s.gif");
    next << "</BIG></B>";
    int j_to = from + 10 * disp;
    if(resultNum-disp < j_to) j_to = resultNum-disp;
    makeNext1(next,cgi,params,disp,escaped_key, j_to, "[>>]", "next2s.gif");
    makeNext1(next,cgi,params,disp,escaped_key, resultNum-disp,"[>|]","ends.gif");

    stringstream prev1;
    makeNext1(prev1,cgi,params,disp,escaped_key, From1, "");
    setrv("prev1",prev1.str().c_str());

    stringstream next1;
    makeNext1(next1,cgi,params,disp,escaped_key, To1, "");
    setrv("next1",next1.str().c_str());

    delete []escaped_key;
    return 0;
}


void
writelog1(int fd, const char *datebuf, const char *host, const char *user,
          const char *query, int resultNum){
    // total elapsed time field "searchTime/elapsedTime"
    char etime[32];
    sprintf(etime,"%0.3f",Time()-Start);

    // list of indexes
    for(char *sp = param_index; *sp; sp++)
        if(*sp == ' ')
            *sp = '+';

    stringstream line;
    line << datebuf << ' ' << host << ' ' << user << ' '
         << param_index << ' ' << param_style << ' '
         << resultNum << ' ' << param_from << ' ' << param_n << ' '
         << elapsedTimes << '/' << etime << ' '
         << query << '\n';

    string linestr = line.str();
    write(fd, linestr.c_str(), linestr.length());
}

char *setup_resultprint(char *title, char *query, char *uquery,
	int resultLen, int rto ){
        sprintf(elapsedTimes,"%0.3f",Time()-Start);

        stringstream stitle;
        stitle << "Result: " << query;
        stitle << " (" << resultLen << " documents found. #" << param_from;
        stitle << "--#" << rto << ')' << '\0'; // terminate explicitly
	strcpy(title,stitle.str().c_str());

        sprintf(resultFroms,"%d",param_from);
        sprintf(resultTos,"%d",rto);

        makeParams(query,0);
        stringstream Next;
        makeNext(Next,uquery,resultLen,param_n,param_from-1,rto-1);
	char *np = strdup(Next.str().c_str());
        makeParams(query,1);
	return np;
}

#include "SX.h"
char *kv_getval(KeyVals *kv, const char *key);
extern KeyVals *xkv;
char *xgetrv(const char *key);

KeyVals *scan_query(const char *query){
	const char *sp,*np;
	char ch;
	int qi;
	KeyVal kv[128];
	int ki = 0;

	char *cgipath = getenv("SCRIPT_NAME");
	if( cgipath ){
		kv[ki].key = "cgipath";
		kv[ki].val = cgipath;
		ki++;
	}

	sp = query;
	for( qi = 0; qi < 128 && *sp != 0; qi++ ){
		stringstream *key,*val;
		key = new stringstream;
		val = new stringstream;
		for( np = sp; (ch = *np) != 0; np++ ){
			if( *np == '=' ){
				for( np = np+1; (ch = *np); np++ ){
					if( ch == '&' )
						break;
					*val << ch;
				}
				break;
			}else{
				*key << ch;
			}
		}
		*key << '\0';
		*val << '\0';

		kv[ki].key = strdup(key->str().c_str());  
		kv[ki].val = strdup(val->str().c_str());  
		ki++;

		delete key;
		delete val;
		if( *np == 0 )
			break;
		sp = ++np;
	}
	KeyVals *keyval = new KeyVals;
	keyval->len = ki;
	keyval->keyval = new KeyVal[ki];
	memcpy(keyval->keyval,kv,sizeof(KeyVal)*ki);
	return keyval;
}

void add_indexlist(int add, char *lstfile,char *index,char *description);
void edit_indexlist(char *com){
	char *index;
	char lstpath[PATH_MAX];

	char *atype = getenv("AUTH_TYPE");
	char *auser = getenv("REMOTE_USER");
	if( atype == NULL || strcasecmp(atype,"Digest") != 0
	 || auser == NULL || *auser == 0 ){
		sout_line("<BR>*** You are not authorized ***<P>");
		putNotAuthorized();
		return;
	}

	strcpy(lstpath,"index.lst");

	if( strcasecmp(com,"Publish") == 0 )
	if( (index = xgetrv("addindex")) ){
		add_indexlist(1,lstpath,index,xgetrv("description"));
	}
	if( strcasecmp(com,"Shadow") == 0 )
	if( (index = xgetrv("delindex")) ){
		add_indexlist(0,lstpath,index,NULL);
	}
}

int condSubsts(char *line, int size, char *(*getrv)(const char *));
int scangen_html(const char *base, KeyVals *kv){
	char *com,file[1024],path[1024],line[0x10000];

	xkv = kv;
	com = kv_getval(kv,"com");
	if( *com ){
		edit_indexlist(com);
	}
/*
	if( *com )
		sprintf(file,"%s-%s.html",base,com);
	else	sprintf(file,"%s.html",base);
*/
	sprintf(file,"%s.html",base);
	if( 0 < SX_pfxPath("",file,path,sizeof(path)) ){
		Strncpy(file,path,sizeof(file));
	}

	Fstream html(file,"r");
	if( html.good() ){
		for(;;){
			if( html.gets(line,sizeof(line)) == NULL )
				break;
			scan_META(line,sizeof(line));
			condSubsts(line,sizeof(line),xgetrv);
			sout << line;
		}
	}
	return 0;
}

