// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Unifier.cc,v 0.9 1997-12-27 07:24:50+09 harada Exp $
//

#include "SX.h"
#include "Unifier.h"
#include "Word.h"
#include "kcnvlib.h"

struct pair_t {
    char *key;
    char *value;
};

static struct pair_t zenhanTbl[] = {
    {"", " "},    {"", ","},    {"", "."},    {"", ":"},
    {"", ";"},    {"", "?"},    {"", "!"},    {"", "^"},
    {"", "~"},    {"", "_"},    {"", "/"},    {"", "\\"},
    {"", "|"},    {"", "`"},    {"", "'"},    {"", "\""},
    {"", "\""},    {"", "("},    {"", ")"},    {"", "["},
    {"", "]"},    {"", "{"},    {"", "}"},    {"", "+"},
    {"", "-"},    {"", "="},    {"", "<"},    {"", ">"},
    {"", "\\"},    {"", "$"},    {"", "%"},    {"", "#"},
    {"", "&"},    {"", "*"},    {"", "@"},    {"", "0"},
    {"", "1"},    {"", "2"},    {"", "3"},    {"", "4"},
    {"", "5"},    {"", "6"},    {"", "7"},    {"", "8"},
    {"", "9"},    {"", "A"},    {"", "B"},    {"", "C"},
    {"", "D"},    {"", "E"},    {"", "F"},    {"", "G"},
    {"", "H"},    {"", "I"},    {"", "J"},    {"", "K"},
    {"", "L"},    {"", "M"},    {"", "N"},    {"", "O"},
    {"", "P"},    {"", "Q"},    {"", "R"},    {"", "S"},
    {"", "T"},    {"", "U"},    {"", "V"},    {"", "W"},
    {"", "X"},    {"", "Y"},    {"", "Z"},    {"", "a"},
    {"", "b"},    {"", "c"},    {"", "d"},    {"", "e"},
    {"", "f"},    {"", "g"},    {"", "h"},    {"", "i"},
    {"", "j"},    {"", "k"},    {"", "l"},    {"", "m"},
    {"", "n"},    {"", "o"},    {"", "p"},    {"", "q"},
    {"", "r"},    {"", "s"},    {"", "t"},    {"", "u"},
    {"", "v"},    {"", "w"},    {"", "x"},    {"", "y"},
    {"", "z"},
};

static struct pair_t hanzenTbl[] = {
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},    {"", ""},
    {"", ""},    {"", ""},    {"", ""},
};

static struct pair_t dakuonTbl[] = {
    {"", ""},   {"", ""},   {"", ""},   {"", ""},
    {"", ""},   {"", ""},   {"", ""},   {"", ""},
    {"", ""},   {"", ""},   {"", ""},   {"", ""},
    {"", ""},   {"", ""},   {"", ""},   {"", ""},
    {"", ""},   {"", ""},   {"", ""},   {"", ""},
    {"", ""},
};

static struct pair_t handakuonTbl[] = {
    {"", ""},   {"", ""},   {"", ""},   {"", ""},
    {"", ""},
};


HashBag<Word> *Unifier::zenhanMap = NULL;
HashBag<Word> *Unifier::hanzenMap = NULL;
HashBag<Word> *Unifier::dakuonMap = NULL;
HashBag<Word> *Unifier::handakuonMap = NULL;
int Unifier::refcount = 0;

Unifier::Unifier()
{
    if(zenhanMap){
        refcount++;
        return;
    }

    zenhanMap = new HashBag<Word>(sizeof(zenhanTbl)/sizeof(zenhanTbl[0]));
    hanzenMap = new HashBag<Word>(sizeof(hanzenTbl)/sizeof(hanzenTbl[0]));
    dakuonMap = new HashBag<Word>(sizeof(dakuonTbl)/sizeof(dakuonTbl[0]));
    handakuonMap = new HashBag<Word>(sizeof(handakuonTbl)/sizeof(handakuonTbl[0]));

    unsigned int i;
    for(i = 0; i < sizeof(zenhanTbl)/sizeof(zenhanTbl[0]); i++){
        Word key(zenhanTbl[i].key);
        zenhanMap->set(key, zenhanTbl[i].value);
    }
    for(i = 0; i < sizeof(hanzenTbl)/sizeof(hanzenTbl[0]); i++){
        Word key(hanzenTbl[i].key);
        hanzenMap->set(key, hanzenTbl[i].value);
    }
    for(i = 0; i < sizeof(dakuonTbl)/sizeof(dakuonTbl[0]); i++){
        Word key(dakuonTbl[i].key);
        dakuonMap->set(key, dakuonTbl[i].value);
    }
    for(i = 0; i < sizeof(handakuonTbl)/sizeof(handakuonTbl[0]); i++){
        Word key(handakuonTbl[i].key);
        handakuonMap->set(key, handakuonTbl[i].value);
    }
}

Unifier::~Unifier(){
    if(refcount == 0){
        delete zenhanMap;
        delete hanzenMap;
        delete dakuonMap;
        delete handakuonMap;
    }else
        refcount--;
}

int
Unifier::unify(char *buf, const char *str) const{
    const char *cp;
    char *dp, *pp;

    pp = NULL;  // pp points previous Kanji
    cp = str;  dp = buf;

    while(*cp){
        if(*cp == '-' && pp && (cp - 2) >= str &&
           *(unsigned char *)(cp - 2) == 0x8e){
            pp = dp;
            *dp++ = 0xa1; *dp++ = 0xbc; // EUC zenkaku chouon
            cp++;
            continue;
        }

        if(Is_dakuten_here(cp)){
            if(pp){
                Word key(pp, 2), value;
                if(dakuonMap->get(key, value)){
                    value.paste(dp - 2);
                    pp = dp - 2;
                    dp = pp + value.length();
                    cp += 2;
                    continue;
                }
            }
            pp = dp;
            *dp++ = 0xa1; *dp++ = 0xab; // EUC zenkaku dakuten
            cp += 2;
            continue;
        }

        if(Is_handakuten_here(cp)){
            if(pp){
                Word key(pp, 2), value;
                if(handakuonMap->get(key, value)){
                    value.paste(dp - 2);
                    pp = dp - 2;
                    dp = pp + value.length();
                    cp += 2;
                    continue;
                }
            }
            *dp++ = 0xa1; *dp++ = 0xac; // EUC zenkaku handakuten
            pp = NULL;
            cp += 2;
            continue;
        }

        if(Is_kanji_here(cp)){
            Word key(cp, 2), value;
            if(! zenhanMap->get(key, value)){
                pp = dp;
                *dp++ = *cp++; *dp++ = *cp++;
            }else{
                value.paste(dp);
                dp += value.length();
                cp += 2;
                pp = NULL;
            }
            continue;
        }

        if(*(cp + 1) && *(unsigned char *)cp == 0x8e){
            Word key(cp, 2), value;
            if(! hanzenMap->get(key, value)){ // This should not be true
                *dp++ = ' '; *dp++ = ' ';
                cp += 2;
                pp = NULL;
            }else{
                value.paste(dp);
                pp = dp;
                dp += value.length();
                cp += 2;
            }
            
            continue;
        }

        if((! iscntrl(*cp)) && isprint(*cp)){
            *dp++ = *cp++;
            pp = NULL;
        }else{
            *dp++ = ' ';
            cp++;
            pp = NULL;
        }
    }
    *dp = '\0';
    return dp - buf;
}

#ifdef UNIFIER_DEBUG
// % g++ -O -g -DUNIFIER_DEBUG Unifier.cc tHashBag.cc Word.cc
int main(){
    Unifier unif;
    char original[] = "-ˎ--ˎߎߎގޥҎޥҡ-͎ߎ";
    char test[128];
    unif.unify(test, original);
    cout << "orig = " << original << endl;
    cout << "test = " << test << endl;
    
    return 0;
}
#endif
// Local Variables:
//      file-coding-system: *euc-japan*
// End:

