// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: Word.h,v 0.9 1997-12-27 07:25:40+09 harada Exp $
//
#ifndef WORD_H_INCLUDED
#define WORD_H_INCLUDED

#include "HashKey.h"
#define STR_ISSTAT	1
#define STR_ISBANK	2

class Word : public HashKey {
    friend ostream& operator <<(ostream &stream, Word &w);
    friend bool operator ==(const Word &foo, const Word &bar);
    friend bool operator !=(const Word &foo, const Word &bar);
    friend bool operator >=(const Word &foo, const Word &bar);
    friend bool operator <=(const Word &foo, const Word &bar);
    friend bool operator >(const Word &foo, const Word &bar);
    friend bool operator <(const Word &foo, const Word &bar);
protected:
    char *buf;
/*
    int leng; // -staticLen-1 or dynamicLen+1
*/
    short leng; // -staticLen-1 or dynamicLen+1
    short flag;
    int cmp(const Word &that) const;
public:
/*
    Word() : buf(NULL){}
    Word() : buf(NULL), leng(0) {}
*/
    Word() : buf(NULL), leng(0), flag(0) {}
    Word(int len, const char *str){
	flag = STR_ISSTAT;
        leng = -len-1;
        buf = (char*)str;
    }
    Word(const char *str, int len, char *sbuf, int size){
	flag = STR_ISSTAT;
        leng = -len-1;
        buf = sbuf;
        if(size <= len){ cerr << "[Word buffer insufficient]" << endl; }
        for(int i = 0; i < len; i++) sbuf[i] = str[i];
        sbuf[len] = 0;
    }
    Word(const char *str){
        leng = strlen(str)+1;
        if(leng == 1){
            leng = -1;
            buf = "";
	}else
	if( stringBank ){
		buf = stringBank->newstring(str);
		leng = -(strlen(str)+1);
		flag |= STR_ISBANK;
        }else{
        buf = new char[leng];
/*
        buf = new char[strlen(str) + 1];
*/
        strcpy(buf, str);
        }
    }
    Word(const char *str, int strLen){
        if(strLen == 0){
            leng = -1;
            buf = "";
	}else
	if( stringBank ){
		buf = stringBank->newstring(str,strLen);
		leng = -(strLen+1);
        }else{
        leng = strLen+1;
        buf = new char[strLen + 1];
        strncpy(buf, str, strLen);
        buf[strLen] = '\0';
        }
    }
    ~Word(){
	if( buf != NULL && stringBank && (flag & STR_ISBANK) ){
		stringBank->freestring(buf);
		buf = 0;
	}
        if(0 <= leng)
        if(buf) delete[]buf;
    }
    const Word & operator =(const Word &that);
    operator const char *() const { return buf; }
    operator bool() const { return buf != NULL; }
    const char *toString() const { return buf ? buf : ""; }

    int operator ==(const HashKey &that) const {
        if( buf == that.toString() ){
             return 1;
        }
        return (0 == strcmp(buf, that.toString()));
    }
    int operator ==(const Word &that) const {
        return 0 == cmp(that);
    }
/*
    HashKey *clone() const { return new Word(buf); }
    HashKey *clone() const { HashKey *k=new Word(buf); k->kid = kid; return k; }
*/
    HashKey *clone() const {
	HashKey *k = new Word(buf);
	k->kid = kid;
	return k;
    }
    unsigned int hashCode() const;
/*
    int length() const { return strlen(buf); }
*/
    int length() const {
        if(0 <= leng)
            return  leng - 1;
        else
            return -leng - 1;
        /*
        return strlen(buf);
        */
    }
    int paste(char *to) const;
};

#endif
