// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: WordBreaker.cc,v 0.10 1998-01-08 01:49:56+09 harada Exp $
//

#include "SX.h"
#include "kcnvlib.h"
#include "Word.h"
#include "WordBreaker.h"
#include "valueSet.h"
#include "ConText.h"
int WordCapitalNum;
int WordCapitalLen;

WordBreaker::WordBreaker(Patricia &pat)
    : morphdic(pat)
{
    Ctx = 0;
}

WordBreaker::~WordBreaker(){
}

int
WordBreaker::is_valid_kanji(const char *kp) const{
    unsigned char valid[] = "";
    for(const unsigned char *vp = valid; *vp; vp += 2){
        if(       *vp == *(unsigned char *)kp
           && *(vp+1) == *(unsigned char *)(kp+1))
            return 1;
    }
    return Is_jis_level1(kp) || Is_jis_level2(kp)
        || Is_cyrillic(kp) || Is_greek(kp)
        || Is_hiragana_here(kp) || Is_katakana_here(kp);
}

int
WordBreaker::breakJapaneseText(HashBag<valueSet<pos_t> > &bag,
                               const char *text, ElemCtx &ectx, pos_t &offset){
    // cout << "breakJapaneseText:\"" << text << "\"\n";
    int from = bag.size();
    Pat_node *prefixes[MAXPREFIXES];
    const char *cp = text;
    const char *cpforward = cp; // ϿѤߤʬ
    int prevMatchNum = 0;

    while(*cp && *(cp + 1)){
        int matchNum = morphdic.search_prefix(cp, prefixes);
        char *kstr = 0;
        int klen = 0;
        if( 0 < matchNum ){
            kstr = prefixes[matchNum - 1]->key;
            klen = strlen(kstr);
        }

        ectx.c_words++;
        if(matchNum == 0
           || 0 < klen && klen <= 4 ){ // 4 ˤ٤?
            if(*(cp + 2) == '\0'){ // Ǹΰʸ
/*
                if(cpforward < cp + 2){
*/
                if(cpforward < cp + 2
		 || prevMatchNum == 0   // the prev. word is not in dict.
		 && cpforward == cp + 2 // single Japanese character at the tail
		){
                    char sbuf[3];
                    Word term(cp, 2, sbuf,sizeof(sbuf));
                    bag[term] += toPos(ectx,offset);
                    if(Ctx) Ctx->CtxPutpos(term,ectx,offset);
                    cpforward = cp + 2;
                }
                else{
                    ectx.c_words--;
                }
            }else{              // 2ʸ(4byte)Ͽ1ʸʤ
                char sbuf[5];
                Word term(cp, 4, sbuf,sizeof(sbuf));
                bag[term] += toPos(ectx,offset);
                if(Ctx) Ctx->CtxPutpos(term,ectx,offset);
                cpforward = cp + 4;
            }
            offset += 2;
            cp += 2;
        }else{
            // 3ʸ(6byte)ʾĹ줬ޥå
            Word term(klen,kstr);
            if(klen <= MaxKEYLENGTH - 1) // -1 ǰΤ
            {
                bag[term] += toPos(ectx,offset);
                if(Ctx) Ctx->CtxPutpos(term,ectx,offset);
            }
            else{
                ectx.c_words--;
            }
            cp += klen;
            cpforward = cp;
            offset += klen;
        }
        prevMatchNum = matchNum;
    }
    return bag.size() - from;
}


int
WordBreaker::entryWord(HashBag<valueSet<pos_t> > &bag, const char *word,
                       int wordLen, ElemCtx &ectx, pos_t offset, int method){
    char buf[MAXKEYLENGTH * 2];
    int allcapital = 1;
    int colcapital = 0;

    const char *cp = word;
    char *dp = buf;
    for(int i = 0; i < wordLen; i++){
        if(islower(*(unsigned char *)cp)){
            allcapital = 0;
            *dp++ = *cp++;	// just a copy
        }else{
            colcapital = i;
            if(isupper(*(unsigned char *)cp))
                *dp++ = *cp++ + 'a' - 'A'; // decapitalize
            else
                *dp++ = *cp++;	// just a copy
        }
    }
    *dp = '\0';
    if(!allcapital && MinCAPITALCOL && ((MinCAPITALCOL-1) <= colcapital)){
        allcapital = 1;
    }

    ectx.c_words++;
    if(method == FORINDEX){
        // ʸΤϿ
        // ٤ʸʤ餽Ͽ
        Word term(wordLen, buf);
        bag[term] += toPos(ectx,offset);
        if(allcapital && 0 != strcmp(word, buf)){
            Word origterm(wordLen, word);
            bag[origterm] += toPos(ectx,offset);
            if(Ctx) Ctx->CtxPutpos(origterm,ectx,offset);
            WordCapitalNum++;
            WordCapitalLen += wordLen;
        }
        else{
            if(Ctx) Ctx->CtxPutpos(term,ectx,offset);
        }
    }else{
        // ٤ʸʤ餽Ͽ
        // ǤʤȤϾʸΤϿ
        if(allcapital){
            Word origterm(wordLen, word);
            bag[origterm] += toPos(ectx,offset);
        }else{
            Word term(wordLen, buf);
            bag[term] += toPos(ectx,offset);
        }
    }
    return allcapital;
}

inline int symbols(const char *sp){
    if(*sp == ':'){
        if(strncmp(sp,"://",3) == 0) return 3;
    }
    return 0;
}

int
WordBreaker::breakText(HashBag<valueSet<pos_t> > &bag, const char *text,
                       ElemCtx &ectx, pos_t &offset, int method){
    char buf[MAXSTRINGLENGTH + 1];
    char *kp;
    int from = bag.size();
    
    for(const char *cp = text; *cp; ){
        if(Is_kanji_here(cp)){
            if(! is_valid_kanji(cp)){
                cp += 2;
                offset++;
            }else{
                kp = buf;
                do {
                    *kp++ = *cp++; *kp++ = *cp++;
                }while(Is_kanji_here(cp) && is_valid_kanji(cp));
                *kp = '\0';
                (void) breakJapaneseText(bag, buf, ectx, offset);
            }
        }else
        if(int termLen = symbols(cp)){
            Strncpy(buf,cp,termLen+1);
            entryWord(bag,buf,termLen,ectx,offset,method);
            cp += termLen;
            offset += termLen;
        }else if(is_valid_head(*cp)){ // ñ
            int bufLen = 0;
            char *bufp = buf;
            do {
                *bufp++ = *cp++;
                bufLen++;
            } while(is_valid_char(*cp));
            *bufp-- = '\0';
            // Ŭڤʸ ('.' ʤ)
            while(bufp > buf && (! is_valid_tail(*bufp)))
                *bufp-- = '\0';
            int termLen = bufp + 1 - buf;
            
            assert(termLen > 0);
            if(MinKEYLENGTH <= termLen)
            if(termLen <= MaxKEYLENGTH - 1)
                (void) entryWord(bag, buf, termLen, ectx, offset, method);
            offset += bufLen;
        }else if(Ishkana1(*cp) && Ishkana2(*(cp + 1))){
            // ȾѥʤskipbreakText()Ƥ
            cp += 2;
            offset += 2;
        }else{					// ¾skip
            if((*cp & 0x80)==0 && !isspace(*cp))
                ectx.c_delch = *cp; //delimiter to be put with next word
            cp++; offset++;
            while(isspace(*(unsigned char *)cp)) cp++;
/*
            continue;
*/
        }
    }
    
    return bag.size() - from;
}
