// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: findex.cc,v 0.10 1997-12-29 14:22:42+09 harada Exp $
//
#include "SX.h"
#include "Indexer.h"
#include "XMap.h"
/*
#include <fstream.h>
*/
#include <fstream>
void add_indexlist(int add, char *lstfile,char *index,char *description);

#include <signal.h>
void sigint(int sig){
	fprintf(stderr,"\n-- GOT SIGINT --\n");
	exit(0);
}

char *
mystrdup(const char *str){
    int l = strlen(str);
    char *s = new char[l + 1];
    strcpy(s, str);
    return s;
}

void
printusage(){
    cout << FREYABANNER << endl
         << "Usage: findex indexname [-cCONFIG] [source]" << endl;
    cout << "       findex [-d dictfile] [-o indexname] [source]" << endl;
    cout << "       find /www/path | any2fdif | findex indexname" << endl;
    cout << endl;
    cout << "  CONFIG parameters:" << endl;
    cout << "    -cMKL=n  max. key length to be indexed [64]" << endl;
    cout << "    -cNKL=n  min. key length to be indexed [1]" << endl;
    cout << "    -cNCC=n  min. column of capital char. in a key to be indexed [1]" << endl;
    cout << endl;
}

int
main(int argc, char *argv[]){
	signal(SIGINT,sigint);

    setAsyncIO();
    int optj = 0;
    LapReport(1,"",0,0,0,0);

    /*
    const char *dictpath = DEFAULT_MORPHDIC_PATH;
    const char *outputbase = "default";
    */
    char dpath[PATH_MAX];
/*
    char *dictpath = SX_defaultDict(dpath,sizeof(dpath));
*/
    char *dictpath = "";
    char bpath[PATH_MAX],fpath[PATH_MAX];
    char *outputbase = "default";
    int opta = 0;
    int optb = 1;
    int optd = 0;
    int optx = 1;
    extern char *optarg;
    extern int optind; // opterr, optopt;
    int numout = 0;

    if(argc == 1 && isaTty(fileno(stdin)) ){
        printusage();
        return 1;
    }

    int ch;
/*
    while((ch = getopt(argc, argv, "c:d:o:b:Sjv?")) != EOF){
*/
/*
    while((ch = getopt(argc, argv, "c:d:o:b:axSjv?")) != EOF){
*/
    while((ch = getopt(argc, argv, "c:d:o:b:aDxSjv?")) != EOF){
        switch(ch){
        case 'c':
            SX_scanCONF(optarg);
            break;
	case 'D':
	    dictpath = SX_defaultDict(dpath,sizeof(dpath));
            optd = 1;
	    break;
        case 'd':
            dictpath = mystrdup(optarg);
            optd = 1;
            break;
        case 'o':
            numout++;
            outputbase = mystrdup(optarg);
            optb = 0;
            break;
        case 'b':
            numout++;
            outputbase = mystrdup(optarg);
            optb = 1;
            break;
        case 'a':
            opta = 1;
            break;
        case 'x':
            optx = 0;
            break;
        case 'j':
            optj = 1;
            break;
        case 'S':
            opt_S = 1;
            break;
        case 'v':
            opt_v = 1;
            break;
        case '?':
        default:
            printusage();
            return 1;
        }
    }
    argc -= optind;
    argv += optind;

    if(numout == 0){ /* regard "-b arg" for first non optionn arg. */
        if(opta && argc == 1){
            /* -a arg.fdif == -a -b default arg.fdif */
        }else
        if(0 < argc){
            cerr << "*** -b " << argv[0] << endl;
            outputbase = mystrdup(argv[0]);
            optb = 1;
            argc--;
            argv++;
        }
    }
    if(argc == 0){
        if(!isaTty(fileno(stdin))){
	  int ch = getc(stdin);
	  if( ch == EOF ){
	    /* fix-140602a might be invoked from cron */
	    fprintf(stderr,"*** findex: EOF from stdin\r\n");
	  }else{
	    ungetc(ch,stdin);
            cerr << "*** stdin as FDIF" << endl;
            argv[0] = "-";
            argc++;
	  }
        }
    }

    if( *dictpath == 0 ){
    }else
    if(FileSize(dictpath) < 0)
    cerr << "*** WARNING: without -d Morphdic" << endl;
    else
    cerr << "*** -d " << dictpath << endl;

    // -b "path/name" for -o "path/name/name" "path/name/name.fdif"
    if( optb ){
        SX_createBase(outputbase,bpath,sizeof(bpath));
        outputbase = bpath;
        if(argc == 0){
            strcpy(fpath,bpath);
            strcat(fpath,".fdif");
            argv[0] = fpath;
            argc = 1;
	    fprintf(stderr,"*** findex: default FDIF: %s\r\n",fpath);
        }
        cerr << "*** -o " << outputbase << " " << argv[0] << endl;
    }

    if(argc != 1){
        printusage();
        return 1;
    }

char *seedbase = NULL;
if(opta)
seedbase = outputbase;

    char inputbase[PATH_MAX];
    strcpy(inputbase,outputbase);
    Files files(32,outputbase);
    char ob[PATH_MAX];
    sprintf(ob,"%s#",outputbase);
    outputbase = ob;

    const char *filepath = argv[0];
#ifndef NDEBUG
    if( *dictpath == 0 ){
    }else
    cout << "dictpath = " << dictpath << endl;
    cout << "filepath = " << filepath << endl;
    cout << "outputbase = " << outputbase << endl;
#endif

    // setup the word breaker
    Fstream morphdicfile(dictpath,"r");
    if( *dictpath == 0 && morphdicfile.good() ){
	fprintf(stderr,"??? succeeded to open empty filename...\n");
	morphdicfile.close();
	/* become !good() */
    }
    if(optd || 0 < FileSize(dictpath))
    if(!morphdicfile.good()){
        PRINTERRMSG("Error: cannot open morphdic.");
        cerr << "  dictpath = \"" << dictpath << "\"\n";
        if(optd)
        return 1;
    }
    Patricia morphdic(morphdicfile);
    morphdicfile.close();
/*
    if(! morphdic){
*/
    if(optd || 0 < FileSize(dictpath))
    if(!morphdic.good()){
        PRINTERRMSG("Error: invalid morphdic.");
        cerr << "  dictpath = \"" << dictpath << "\"\n";
        return 1;
    }
    LapReport(0,"loaded %d keys.",(int)morphdic,0,0,0);
    WordBreaker wbrk(morphdic);

    // setup the indexer
    const Unifier unif;
/*
    Indexer idxer(wbrk, unif);
*/

    // condensed text
    char ctxpath[PATH_MAX] = "";
    ConText *newctx = 0;
    if(optx){
        strcpy(ctxpath, outputbase); strcat(ctxpath, ".ctx");
        files.add("ConTextFile",".ctx");
        newctx = new ConText(ctxpath,"w");
    }
    if(opta){
        char seedctxpath[PATH_MAX];
        sprintf(seedctxpath,"%s.ctx",seedbase);
        newctx->setMap(FileSize(seedctxpath)/4);
	/* element of .ctx might be larger than 4 ... */
    }
    Indexer idxer(wbrk, unif, newctx);
    if(!optj)
        idxer.offUnifier();
    pos_t offset = 1;

    // prepare xmap
    char xmappath[PATH_MAX];
    strcpy(xmappath, outputbase); strcat(xmappath, ".map");
    files.add("XMapFile",".map");
    Fstream xmapfile(xmappath,"w");
    if(! xmapfile.good()){
        PRINTERRMSG("cannot create a translation map file.\n");
        cerr << "  xmappath = \"" << xmappath << "\"\n";
        return 2;
    }

    char path[PATH_MAX];
    sprintf(path,"%s.dsc",inputbase);
    Fstream idescfile(path,"r");
    if(idescfile.good()){
        idxer.idescf = &idescfile;
    }

    sprintf(path,"%s.dsc",outputbase);
    Fstream descfile(path,"w");
    idxer.descf = &descfile;
    files.add("DescFile",".dsc");

    if(opta){
        char path[PATH_MAX];
        sprintf(path,"%s.map",seedbase);
        Fstream sxmap(path,"r");
        if(!sxmap.good()){
            cerr << "cannot open XMapFile: " << path << endl;
            return 2;
        }
        int rcc;
        int size = 0;
        int ndocs = 0;
        XMapEntry entry;
        while( 0 < (rcc = sxmap.read((char*)&entry,sizeof(entry))) ){
            xmapfile.write((char*)&entry,rcc);
            ndocs++;
            size += rcc;
        }
        idxer.SeedDocs = ndocs;
        idxer.SeedWords = entry.last;

        // DescFile
        sprintf(path,"%s.dsc",seedbase);
        Fstream sdesc(path,"r");
        if(!sdesc.good()){
            cerr << "cannot open DescFile: " << path << endl;
            return 2;
        }
        char buf[256];
        while( 0 < (rcc = sdesc.read((char*)buf,sizeof(buf))) ){
            descfile.write((char*)buf,rcc);
        }
        descfile.flush();
    }
    if(errno == ENOENT)         // clear "No such file or directory"
        errno = 0;
    LapReport(0,"start indexing.",0,0,0,0);

    // Let's index!
    char tmpbase[PATH_MAX];
    strcpy(tmpbase, outputbase); strcat(tmpbase, "_");
/*
    int tmpfileNum;
*/
    int tmpfileNum = -1;
    if(0 == strcmp(filepath, "-")){
        tmpfileNum = idxer.makeIndices(cin, tmpbase, xmapfile, offset);
    }else{
        fstream source(filepath, ios::in);
        if(!source.good()){
            PRINTERRMSG("cannot open source text. ");
            cerr << "  filepath = \"" << filepath << "\"\n";
            return 3;
        }
        tmpfileNum = idxer.makeIndices(source, tmpbase, xmapfile, offset);
    }
    descfile.close();
    xmapfile.close();

    int ndocs = idxer.SeedDocs + idxer.NewDocs;
    IMask *mask = 0;

    if(opta){
        // create a mask indicating documents to be overridden
        // erase overridden documents in DescFile and XMap file
        mask = idxer.toberemoved(outputbase,ndocs);
    }

    idxer.mergeIndices(outputbase,tmpbase,tmpfileNum, mask,seedbase,files);
    createSortFiles(outputbase,files);

    if(newctx){
        char *inputs[3];
        char seedctxpath[PATH_MAX];

        newctx->close();
        if(opta){
            sprintf(seedctxpath,"%s.ctx",seedbase);
            inputs[0] = seedctxpath;
            inputs[1] = strdup(ctxpath);
            inputs[2] = 0;
            newctx->inputs(inputs,2);
        }else{
            newctx->open(ctxpath,"r");
        }
        SX_tempName(ctxpath);
        ConText *octx = new ConText(ctxpath,"w");
        newctx->adjustLex(octx,mask);
        newctx->close();
        octx->close();
        SX_realName(ctxpath);
    }
    idescfile.close();

    files.renames();

    char *iname,*dp,*ep,lpath[PATH_MAX];
    if((dp = strrchr(inputbase,'/'))){
        iname = dp + 1;
        strcpy(lpath,inputbase);
        dp = strrchr(lpath,'/');
        for(ep = dp-1; lpath < ep; ep--){
            if(*ep == '/'){
                dp = ep;
                break;
            }
        }
        strcpy(dp+1,"index.lst");
    }else{
        iname = inputbase;
        strcpy(lpath,"index.lst");
    }
fprintf(stderr,"#### [%s] %s\n",iname,lpath);
    add_indexlist(1,lpath,iname,"(no description)");
    return 0;
}
