// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: fmerge.cc,v 0.10 1998-01-09 22:50:42+09 harada Exp $
//
#include "SX.h"
#include "XMap.h"
#include "Indexer.h"
#include "DescFile.h"
#include "ConText.h"

char *
mystrdup(const char *str){
    int l = strlen(str);
    char *s = new char[l + 1];
    strcpy(s, str);
    return s;
}

void
printusage(){
    cout << FREYABANNER << endl
         << "Usage: fmerge newindex index [index ...]" << endl;
}

int
mergesum(Fstream &newsumfs, IMask *mask, char *sumfilenames[], int n){
    int olddoc = 0;
    for(int i = 0; i < n; i++){
        Fstream isum(sumfilenames[i],"r");
        if(!isum.good()){
            cerr << "Cannot open: " << sumfilenames[i] << endl;
            return -1;
        }
        char line[0x10000];
        while( isum.gets(line,sizeof(line)) != NULL ){
            if(mask == NULL || mask->get(olddoc) != -1){
                newsumfs.puts(line);
            }
        }
        olddoc++;
    }
    newsumfs.flush();
    return 0;
}
int
mergearray(const char *newpath, IMask *mask, char *srcfilenames[], int n,
           const char *ext, int elemsize){

    char path[PATH_MAX];
    sprintf(path,"%s%s",newpath,ext);
    Fstream out(path,"w");

    int olddoc = 0;
    char val[512];
    for(int i = 0; i < n; i++){
        sprintf(path,"%s%s",srcfilenames[i],ext);
        Fstream in(path,"r");
        if(!in.good()){
            // cerr << "Cannot open: " << path << endl;
            if( olddoc == 0 ){
                out.remove(1);
            }
            return -1;
        }
        while( in.read((char*)val,elemsize) == elemsize ){
            if(mask == NULL || mask->get(olddoc) != -1){
                out.write((char*)val,elemsize);
            }
        }
        olddoc++;
    }
    out.flush();
    return 0;
}

// merge .dsc files, and set the number of entries contained in each
int
mergedsc(Fstream &newdscfile, IMask *mask, char *dscfilenames[], int *ndocs, int n){
    cout << "merging descriptions " << flush;
    newdscfile.clear();
    newdscfile.seekp(0);
    int olddoc = 0;
    for(int i = 0; i < n; i++){
        Fstream dscfile(dscfilenames[i],"r");
        if(! dscfile.good()){
            cout << endl;
            PRINTERRMSG("cannot open .dsc file.");
            cerr << "  filename = \"" << dscfilenames[i] << '"' << endl;
            return i + 1;
        }
        off_t filesize = FileSize(dscfilenames[i]);
        ndocs[i] = filesize / DSCBLOCKSIZE;
        
        cout << '.' << flush;
        char buf[DSCBLOCKSIZE];
        dscfile.seekg(0);
        for(int d = 0; d < ndocs[i]; d++){
            dscfile.read(buf, DSCBLOCKSIZE);
            if(mask == NULL || mask->get(olddoc) != -1)
            newdscfile.write(buf, DSCBLOCKSIZE);
            olddoc++;
        }
        dscfile.close();
    }
    cout << "ok" << endl;

    return 0;
}

int countdocs(char *mapfilenames[], int n){
    int tdocs = 0;
    for(int i = 0; i < n; i++){
        int size = FileSize(mapfilenames[i]);
        tdocs += size / sizeof(XMapEntry);
    }
    return tdocs;
}

// merge .map files, and set the number of entries contained in each
int
mergemap(Fstream &newmapfile, IMask *mask, char *mapfilenames[],
         int *ndocs, pos_t *shifttbl, int n){
    cout << "merging translation maps " << flush;
    pos_t newlastoffset = 0;
    newmapfile.clear();
    newmapfile.seekp(0);

    const int entrysize = sizeof(XMapEntry);
    XMapEntry entry;
    int nwords = 0;
    int olddoc = 0;
    int newdoc = 0;
    for(int i = 0; i < n; i++){
        Fstream mapfile(mapfilenames[i],"r");
        if(! mapfile.good()){
            cout << endl;
            PRINTERRMSG("cannot open .map file.");
            cerr << "  filename = \"" << mapfilenames[i] << '"' << endl;
            return i + 1;
        }
        off_t filesize = FileSize(mapfilenames[i]);
        ndocs[i] = filesize / entrysize;
        
        cout << '.' << flush;
        mapfile.seekg(0);
        shifttbl[i] = newlastoffset;
        int last = 0;
        for(int d = 0; d < ndocs[i]; d++){
            mapfile.read((char*)&entry, entrysize);
            int size = entry.last - last;
            last = entry.last;
            if(mask == NULL || mask->get(olddoc) != -1){
/*
            nwords = entry.last = nwords + entry.size;
*/
                nwords += size;
                entry.last = nwords;
            newmapfile.write((char*)&entry, entrysize);
                newdoc++;
            }
            olddoc++;
        }
/*
SHIFT will be applied to the original Indexes
        newlastoffset = topos(newdoc,0,0);
*/
        newlastoffset = topos(olddoc,0,0);
    }
    cout << "ok" << endl;

    return 0;
}

extern int AllowDuplicated;

int
main(int argc, char *argv[]){
    setAsyncIO();
    extern int optind;
    extern char *optarg;
    int optb = 1;
    int optu = 0; // unify
    int optx = 0;
    int rem = -1;

    int ch;
    while((ch = getopt(argc, argv, "xr:oubSv?")) != EOF){
        switch(ch){
            case 'o':
                optb = 0;
                break;
            case 'b':
                optb = 1;
                break;
            case 'x':
                optx = 1;
                break;
            case 'r':
                rem = atoi(optarg);
                break;
            case 'u':
		AllowDuplicated = 1;
                optu = 1;
                break;
            case 'S':
                opt_S = 1;
                break;
            case 'v':
                printusage();
                break;
        }
    }
    argc -= optind;
    argv += optind;

    int i;
    if(argc < 2){
        printusage();
        return 1;
    }

    int srcNum = argc - 1;
    if(srcNum > 16){
        PRINTERRMSG("too many indices.");
        return srcNum;
    }

    char newpath[PATH_MAX];
    strcpy(newpath,argv[0]);
    if(optb){
        SX_searchBase(argv[0],newpath,sizeof(newpath));
        cerr << "[" << argv[0] << " -> " << newpath << "]" << endl;
    }
    Files files(32,newpath);
    strcat(newpath,"#");
    files.add("DescFile",".dsc");
    files.add("XMapFile",".map");

    typedef char *charp;
    char **srcfilenames = new charp[srcNum];
    char **lexfilenames = new charp[srcNum];
    char **idxfilenames = new charp[srcNum];
    char **ctxfilenames = new charp[srcNum];
    char **mapfilenames = new charp[srcNum];
    char **dscfilenames = new charp[srcNum];
    int *ndocsdsc = new int[srcNum];
    int *ndocsmap = new int[srcNum];       // this should be same as ndocsdsc
    pos_t *shifttbl = new pos_t[srcNum];
    typedef LexiconFile *LFP;
    LexiconFile **lexicons = new LFP[srcNum];
    typedef IndexFile *IFP;
    IndexFile **indices = new IFP[srcNum];

    for(i = 0; i < srcNum; i++){
        char path[PATH_MAX];
        SX_searchBase(argv[i+1],path,sizeof(path));
        cerr << "[" << argv[i+1] << " -> " << path << "]" << endl;

        srcfilenames[i] = new char[strlen(path)+1];
        strcpy(srcfilenames[i],path);

        lexfilenames[i] = new char[strlen(path) + 5];
        idxfilenames[i] = new char[strlen(path) + 5];
        ctxfilenames[i] = new char[strlen(path) + 5];
        mapfilenames[i] = new char[strlen(path) + 5];
        dscfilenames[i] = new char[strlen(path) + 5];

        strcpy(lexfilenames[i], path); strcat(lexfilenames[i], ".lex");
        strcpy(idxfilenames[i], path); strcat(idxfilenames[i], ".idx");
        strcpy(ctxfilenames[i], path); strcat(ctxfilenames[i], ".ctx");
        strcpy(mapfilenames[i], path); strcat(mapfilenames[i], ".map");
        strcpy(dscfilenames[i], path); strcat(dscfilenames[i], ".dsc");
    }
    
    char newlexfilename[PATH_MAX];
    char newidxfilename[PATH_MAX];
    char newctxfilename[PATH_MAX];
    char newmapfilename[PATH_MAX];
    char newdscfilename[PATH_MAX];

    strcpy(newlexfilename, newpath); strcat(newlexfilename, ".lex");
    strcpy(newidxfilename, newpath); strcat(newidxfilename, ".idx");
    strcpy(newctxfilename, newpath); strcat(newctxfilename, ".ctx");
    strcpy(newmapfilename, newpath); strcat(newmapfilename, ".map");
    strcpy(newdscfilename, newpath); strcat(newdscfilename, ".dsc");

/*
    Fstream newctxfs(newctxfilename,"w");
*/
    Fstream newmapfs(newmapfilename,"w");
    Fstream newdscfs(newdscfilename,"w");

    int tdocs = countdocs(mapfilenames,srcNum);
    cerr << tdocs << " documents in " << srcNum << " indices." << endl;

    IMask *mask = 0;
    if(0 <= rem){
        mask = new IMask(tdocs);
        if(0 <= rem && rem < tdocs){
            mask->set(rem,-1);
        }
        int docid = 0;
        int mi;
        for(mi = 0; mi < tdocs; mi++){
            if(mask->get(mi) != -1)
                mask->set(mi,docid++);
        }
        for(mi = 0; mi < tdocs; mi++){
            if(mask->get(mi) == -1){
                fprintf(stderr,"--[%d] %2d\n",mi,mask->get(mi));
            }
        }
    } 

/*
    files.add("SummaryFile",".sum");
    char *sumfilenames[srcNum];
    for(i = 0; i < srcNum; i++){
        sumfilenames[i] = new char[strlen(path) + 5];
        strcpy(sumfilenames[i], path); strcat(sumfilenames[i], ".sum");
    }
    char newsumfilename[PATH_MAX];
    strcpy(newsumfilename, newpath); strcat(newsumfilename, ".sum");
    Fstream newsumfs(newsumfilename,"w");
    if(0 != mergesum(newsumfs, mask, sumfilenames, srcNum)){
    }
    for(i = 0; i < srcNum; i++)
        delete sumfilenames[i];
*/

    if(0 != mergedsc(newdscfs, mask, dscfilenames, ndocsdsc, srcNum))
        return 2;
    newdscfs.close();

    if(0 != mergemap(newmapfs, mask, mapfilenames, ndocsmap, shifttbl, srcNum))
        return 2;
    newmapfs.close();

    for(i = 0; i < srcNum; i++){
        if(ndocsdsc[i] != ndocsmap[i]){
            PRINTERRMSG("inconsistency found between desc. and map.");
            cerr << "  description:\"" << dscfilenames[i] << "\", "
                 << ndocsdsc[i] << " entries." << endl;
            cerr << "  translation map:\"" << mapfilenames[i] << "\", "
                 << ndocsmap[i] << " entries." << endl;
            return i + 1;
        }
    }

    for(i = 0; i < srcNum; i++){
        lexicons[i] = new LexiconFile(lexfilenames[i],"r");
        indices[i] = new IndexFile(idxfilenames[i],"r");
        if(indices[i]->geterror()){
            PRINTERRMSG("cannot open index file.");
            cerr << "  path = \"" << idxfilenames[i] << '"' << endl;
            return i + 1;
        }
    }

  {
    LexiconFile newlexicon(newlexfilename,"w");
    IndexFile newindex(newidxfilename, "w");

    ConText *newctx = 0;
    if(optx == 0){
        newctx = new ConText();
        for(i = 0; i < srcNum; i++){
            newctx->setMap(FileSize(ctxfilenames[i])/4);
	    /* element of .ctx might be larger than 4 ... */
        }
    }
    Indexer idxer(newctx);

IMask *dmask = idxer.toberemoved(newpath, tdocs);
if(dmask){
    // must be added
    mask = dmask;
}

    cout << "merging lexicons and indexfiles ..." << flush;
    cout << endl;

/*
    Indexer::mergeFiles(&newlexicon, &newindex, srcNum,
                        lexicons, indices, files, mask, shifttbl);
*/
    idxer.mergeFiles(&newlexicon, &newindex, srcNum,
                        lexicons, indices, files, mask, shifttbl);

    if(newctx != NULL){
        cerr << "merging ConText..." << endl;
        ConText *octx = new ConText(newctxfilename,"w");
        newctx->inputs(ctxfilenames,srcNum);
/*
        newctx->adjustLex(octx,1,mask);
*/
        newctx->adjustLex(octx,mask);
        octx->close();
        newctx->close();
        files.add("ConTextFile",".ctx");
    }

    createSortFiles((const char *)newpath, files);
/*
    if(mergearray(newpath, mask, srcfilenames, srcNum, "+mtime.srt",4) == 0)
        files.add("SortFile","+mtime.srt");
    if(mergearray(newpath, mask, srcfilenames, srcNum, "+atime.srt",4) == 0)
        files.add("SortFile","+atime.srt");
*/
    if(mergearray(newpath, mask, srcfilenames, srcNum, ".vote.vot",4) == 0)
        files.add("VoteFile",".vote.vot");
  }
    cout << "ok" << endl;

    for(i = 0; i < srcNum; i++){
        delete lexicons[i];
        delete indices[i];
        delete lexfilenames[i];
        delete idxfilenames[i];
        delete mapfilenames[i];
        delete dscfilenames[i];
    }

    files.renames();
    return 0;
}
