// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// create morphdic file for the indexer
// $Id: fmkmorphdic.cc,v 0.9 1997-12-27 07:46:44+09 harada Exp $
//
#include "SX.h"
#include "jlib.h"
#include "patricia.h"

void
printusage(){
    cout << FREYABANNER << endl
         << "Usage: fmkmorphdic <sourcetext> <morphdicfile>" << endl << endl
         << "example: " << endl
         << "  % bin/mkmorphdic dic/morph.txt dic/morph.dic" << endl
         << "  % nkf -e words.txt | bin/mkmorphdic - dic/morph.dic" << endl;
}

int
buildDictionary(Patricia &pat, istream &source){
    char line[LINE_MAX];
    char prevline[LINE_MAX];
    line[0] = '\0';

    cout << "building dictionary ... ";
    int lineno;
    for(lineno = 0; ; lineno++){
		source >> line;
        if(source.eof()) break;
        int match = jprefixmatch(prevline, line);
		if(Patricia::EDUPLICATE == pat.insert(line, match))
			cerr << "Warning: already stored? : " << line << endl;
        strcpy(prevline, line);
        // cout << lineno << " : " << line << " " << match << endl;
	}
    cout << lineno << " words found." << endl;
    if(lineno <= 0){
        cerr << "No valid words found?\nabort." << endl;
        return -1;
    }

    return 0;                   // succeeded
}

int
main(int argc, char *argv[]){
    if(argc != 3){
        printusage();
        return 1;
    }

    Patricia pat;               // empty
    if(0 == strcmp(argv[1], "-")){
        buildDictionary(pat, cin);
    }else{
        Fstream source(argv[1],"r");
        if(! source.good()){
            PRINTERRMSG("cannot open source");
            return 1;
        }
        buildDictionary(pat, source);
        source.close();
    }

    Fstream patfile(argv[2], "w");
    cout << "saving dictionary as file: " << argv[2] << endl;
	int size = pat.save(patfile);
    if(patfile.good()){
        cout << "write: " << size << " bytes" << endl;
    }else{
        PRINTERRMSG("Error occured while writing dictionary?");
    }
    patfile.close();

    return 0;
}
