// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: fsearch.cc,v 0.12 1998-01-24 23:48:10+09 harada Exp $
//
#include "SX.h"
#include "Retriever.h"
#include "Unifier.h"
#include "Expression.h"
#include "Evaluator.h"
#include "Result.h"
#include "DescFile.h"
#include "htmlutil.h"

void
printusage(){
    cout << FREYABANNER << endl
         << "Usage: fsearch [-s mail|html] [-f n] [-t n] -i index key [key ..]" << endl;
    cout << "       fsearch [-f n] [-t n] -b index key [key ..]" << endl;
}

char *
mystrdup(const char *str){
    int l = strlen(str);
    char *s = new char[l + 1];
    strcpy(s, str);
    return s;
}

void
printgeneric(DescRecord &rec, int score){
    cout << "score = " << score << endl;
    cout << rec << flush;
}

void
printmail(DescRecord &rec, int score){
    const char *title = rec.getTitle();
    const char *path = rec.getUrl();
    const char *author = rec.getAuthor();
    const char *digest = rec.getDigest();
    time_t date = rec.getDate();
    char datebuf[128];

    if(path) cout << '[' << path << ']';
    cout << " (score: " << score << ')' << endl;
    cout << "  Date: " << rfc822date(datebuf, sizeof(datebuf), date) << endl;
    cout << "  Length: " << rec.getSize() << endl;
    if(author && author[0])
        cout << "  From: " << author << endl;
    if(title && title[0])
        cout << "  Subject:" << title << endl;
    if(digest && digest[0])
        cout << ">> " << digest << ".." << endl;
    else
        cout << " (no summary)" << endl;
}

void
printhtml(DescRecord &rec, int score){
    const char *title = rec.getTitle();
    const char *path = rec.getUrl();
    const char *author = rec.getAuthor();
    const char *digest = rec.getDigest();
    char datebuf[128];
    time_t lastmod = rec.getLastModified();

    if(title && title[0])
        cout << '"' << title << '"';
    cout << " (score: " << score;
    cout << ", size: " << rec.getSize() << " bytes)" << endl;
    if(path && path[0])
        cout << "  URL: " << path << endl;
    cout << "  Date: " << rfc822date(datebuf,sizeof(datebuf),lastmod) << endl;
    if(author && author[0])
        cout << "  Author: " << author << endl;
    if(digest && digest[0])
        cout << ">> " << digest << ".." << endl;
    else
        cout << " (no summary)" << endl;
}

void
buildquery(char *q, int argc, char *argv[]){
    static Unifier unif;
    char buf[MAXKEYLENGTH + 1];
    char unified[MAXKEYLENGTH + 16 + 1];
    char *p = q;
    for(int i = 0; i < argc; i++){
        strncpy(buf, argv[i], MAXKEYLENGTH);
        unif.unify(unified, buf);
        char *cp = unified;
        while((*p = *cp++) && p - q < MAXQUERYLENGTH)
            p++;
        *p++ = ' ';
    }
    if(*p == ' ' && p > q) p--;
    *p = '\0';
}

int
main(int argc, char *argv[]){
    setAsyncIO();
    char *indexbase = "default";
    char *style = NULL;
    extern char *optarg;
    extern int optind; // opterr, optopt;
    int rfrom = 1;
    int rto = RESULTSFORDISPLAY;
    int optb = 1;
    int numout = 0;

    int ch;
    /*
    while((ch = getopt(argc, argv, "f:t:s:i:v?")) != EOF){
    */
    while((ch = getopt(argc, argv, "f:t:s:i:b:v?")) != EOF){
        switch(ch){
        case 's':
            style = mystrdup(optarg);
            break;
        case 'i':
            numout++;
            optb = 0;
            indexbase = mystrdup(optarg);
            break;
        case 'b':
            numout++;
            optb = 1;
            indexbase = mystrdup(optarg);
            break;
        case 'f':
            rfrom = atoi(optarg);
            break;
        case 't':
            rto = atoi(optarg);
            break;
        case 'v':
            opt_v = 1;
            break;
        case '?':
        default:
            printusage();
            return 1;
        }
    }
    argc -= optind;
    argv += optind;

    if(numout == 0){
        if(0 < argc){
            indexbase = mystrdup(argv[0]);
            optb = 1;
            argc--;
            argv++;
        }
    }

    if(! argv[0]){
        printusage();
        return 1;
    }
    if(NULL == style){
        style = "generic";
    }else if(0 != strcmp(style, "mail")
             && 0 != strcmp(style, "html")
             && 0 != strcmp(style, "generic")){
        PRINTERRMSG("unknown style(-s) specified.");
        cerr << "  style = \"" << style << '"' << endl;
        return 1;
    }

    // rto == 0 stands for `all result'
    if((rto > 0 && rto < rfrom) || rfrom <= 0 || rto < 0){
        PRINTERRMSG("inappropriate range specified.");
        printusage();
        return 1;
    }

    char query[MAXQUERYLENGTH + 1];
    buildquery(query, argc, argv);

    char indexbank[PATH_MAX];
    if( optb ){
    SX_searchBase(indexbase,indexbank,sizeof(indexbank));
    indexbase = indexbank;
    }
    // setup files ...
    char indexpath[PATH_MAX];
    char lexpath[PATH_MAX];
    char xmappath[PATH_MAX];
    strcpy(indexpath, indexbase); strcat(indexpath, ".idx");
    strcpy(lexpath, indexbase); strcat(lexpath, ".lex");
    strcpy(xmappath, indexbase); strcat(xmappath, ".map");

    IndexFile idxfile(indexpath,"r");
    if(idxfile.geterror()){
        PRINTERRMSG("bad index file.");
        cerr << "  indexpath = \"" << indexpath << "\"\n";
        return 1;
    }
    LexiconFile lexfile(lexpath,"r");

    int fd = open(xmappath, O_RDONLY, 0666);
    if(fd == -1){
        PRINTERRMSG("cannot open map file");
        cerr << "  xmappath = \"" << xmappath << "\"\n";
        return 1;
    }

    // setup retriever object
    XMap map(fd);
    Retriever retr(lexfile, idxfile, map);
    Evaluator evaluator;        // use standard ranking policy

    ComplexExpression cexp(query);
    if(cexp.error()){
        cout << "Error: " << cexp.errmsgs[cexp.error()] << endl;
        return 2;
    }

    // do search!
    int resultLen = 0;
    Result *result = cexp.eval(retr, evaluator, resultLen);
    if(resultLen > 0){
        evaluator.rank(result, resultLen);

        char descpath[PATH_MAX];
        strcpy(descpath, indexbase); strcat(descpath, ".dsc");
        Fstream descfile(descpath,"r");
        if(! descfile.good()){
            PRINTERRMSG("cannot open description file.");
            cerr << " descpath = \"" << descpath << "\"\n";
            return 3;
        }

        DescFile descf(descfile);
        void (*printfunc)(DescRecord &, int);

        if(0 == strcmp(style, "mail"))
            printfunc = printmail;
        else if(0 == strcmp(style, "html"))
            printfunc = printhtml;
        else
            printfunc = printgeneric;

        cout << "Result: " << resultLen << " documents found.\n\n";
        if(rto == 0)
            rto = resultLen;
        else
            rto = (resultLen > rto) ? rto : resultLen;
        for(int r = rfrom - 1; r <= rto - 1; r++){
            DescRecord rec;
            descf.getRecord(rec, result[r].docid);
            int docid = result[r].docid;
            if( opt_v ){
                cout << "docid = " << result[r].docid << ", ";
/*
                cout << "docsize = " << result[r].docsize << ", ";
*/
                cout << "docsize = " << map.getdocsize(docid) << ", ";
                cout << "ntimes = " << (int)result[r].ntimes << ", ";
            }
            (*printfunc)(rec, (int)result[r].weight);
            cout << endl;
        }
        cout << endl;
        

#if 0
        Description *desc;
        if(0 == strcmp(style, "mail"))
            desc = new MailDescription(descfile);
        else if(0 == strcmp(style, "html"))
            desc = new HtmlDescription(descfile);
        else
            desc = new Description(descfile);
            
        cout << "Result: " << resultLen << " documents found.\n\n";
        if(rto == 0)
            rto = resultLen;
        else
            rto = (resultLen > rto) ? rto : resultLen;
        for(int r = rfrom - 1; r <= rto - 1; r++){
            desc->print(cout, result[r]);
        }
        cout << endl;
        delete desc;
#endif
    }else{
        cout << "nothing found.\n";
#ifndef NDEBUG
        cexp.printDebug();
#endif
    }
    if(result) delete []result;

    return 0;
}
