// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// $Id: fsearchcgi.cc,v 0.15 1998-01-27 12:30:30+09 harada Exp $
//
#include "SX.h"

#include "Retriever.h"
#include "Unifier.h"
#include "Expression.h"
#include "Evaluator.h"
#include "Result.h"
#include "DescFile.h"
#include "ConText.h"
#include "kcnvlib.h"
#include "htmlutil.h"

void sout_flush();
void sout_putline(const char *linebuf);
void makeParams(const char *query, int from);
char *setup_resultprint(char *t, char *q, char *uq, int resultLen, int rto);
void Exp2string(Expression &exp, char *query, int size);
void writelog1(int fd, const char *datebuf, const char *host,
             const char *user, const char *query, int resultNum);
void sout_line(char *line);
void putMoved(char *url);
int replace_string(char *buf, const char *keyword, const char *str);
int printGeneric();
int SX_include(char *name, int nsiz, char *linebuf);
void scan_META(char *line, int size);
extern double Start;

char param_key[MAXQUERYLENGTH + 1];
char param_hidden_key[MAXQUERYLENGTH + 1];
extern char param_index[NAME_MAX + 1];
extern char param_style[16];
int param_from;
int param_n;

#ifdef _MSC_VER
extern "C" {
int EXlockB(int fd);
int UNlock(int fd);
}
int takeLock(int fd){ return EXlockB(fd); }
int releaseLock(int fd){ return UNlock(fd); }
#else
static struct flock thelock;
int
takeLock(int fd){
    thelock.l_whence = SEEK_SET;
    thelock.l_start = 0;
    thelock.l_len = 0;             // until end of the file
    thelock.l_type = F_WRLCK;      // exclusive/write
    thelock.l_pid = 0;
    int r;
    while ((r = fcntl(fd, F_SETLKW, &thelock)) < 0 && errno == EINTR)
        continue;
    return r < 0 ? r : 0;
}

int
releaseLock(int fd){
    thelock.l_whence = SEEK_SET;
    thelock.l_start = 0;
    thelock.l_len = 0;             // until end of the file
    thelock.l_type = F_UNLCK;      // unlock
    thelock.l_pid = 0;
    int r;
    while ((r = fcntl(fd, F_SETLKW, &thelock)) < 0 && errno == EINTR)
        continue;
    return r < 0 ? r : 0;
}
#endif

#define FSEARCHCGI_LOGPATH "fsearch.log"

#ifdef DISABLE_LOGGING
#define writelog(q,r) ;
#else
// log some information into file
// format is:
//   DATE REMOTE_HOST REMOTE_USER INDEXNAME STYLE resultNum from n query\n
int
writelog(const char *query, int resultNum){
    char *host = getenv("REMOTE_HOST");
    if(NULL == host) host = getenv("REMOTE_ADDR");
    if(NULL == host || '\0' == host[0]) host = "-";

    char *user = getenv("REMOTE_USER");
    if(NULL == user || '\0' == user[0]) user = "-";

    char datebuf[64];
    time_t now = time(NULL);
    strftime(datebuf, sizeof(datebuf) - 1,
             "%d %b %Y %H:%M:%S %Z", localtime(&now));

    if(NULL == query || '\0' == query[0]) query = "-";

    int fd = open(FSEARCHCGI_LOGPATH, O_WRONLY|O_CREAT, 0666);
    if(-1 == fd) return -1;     // exit silently (without logging)
    errno = 0;
    if(0 != takeLock(fd)) return -1;
    lseek(fd, 0, SEEK_END);
    writelog1(fd,datebuf,host,user,query,resultNum);
    releaseLock(fd);
    close(fd);

    return errno;
}
#endif

//-FX - EXTENSTIONS for FreyaSX -------------------------------------------
//-FX
//-FX
#define MAX_IDX 16
char param_sxop[16];       // "sxop=AND|OR"
char param_sort[16];       // "sort=date|date-reverse|score|atime|hits"
char param_url[LINE_MAX];  // "url=fsearch.html"
extern char param_pfx[16]; // "pfx=ja|en"
char param_dbg[64];        // "dbg=nocache,sync" 
char param_docid[64];      // "docid=12345" supercedes "from=n"
char param_pc[64];         // panel control options
char param_jump[LINE_MAX]; // jump=xid.docid.URL for counter
char param_find[LINE_MAX]; // find=URL -> move to the page with index&docid
int param_ssk;             // "ssk=42" text box size for search-key
extern int param_zoom;     // "zoom=1" for <big> and "zoom=-1" for <small>
char param_vote[64];       // "vote=xid.did.N|-N"
char param_frame[64];      // "frame=main|panel|body"
char param_comm[LINE_MAX]; // comment

extern int SXopAND;
#define SXsdate    1
#define SXsdocid   2
#define SXext      3
#define SXsurl     4
#define SXssize    5
#define SXsvote    6
#define SXscount   7
#define SXsatime   8
#define SXsauthor  9
#define SXsint     100
#define SXsstr     101
static int SXsort;
static char SXsrt[32];

static int scan_sort(char *sort){
  int SXsort = 0;
  if(sort[0]){
    if(*sort == 0 || 0 == strcmp(sort,"score"))
       SXsort = 0;
    else
    if(0 == strcmp(sort,"date"))
       SXsort = SXsdate;
    else
    if(0 == strcmp(sort,"date-reverse"))
       SXsort = -SXsdate;
    else
    if(0 == strcmp(sort,"url"))
       SXsort = SXsurl;
    else
    if(0 == strcmp(sort,"author"))
       SXsort = SXsauthor;
    else
    if(0 == strcmp(sort,"size"))
       SXsort = SXssize;
    else
    if(0 == strcmp(sort,"scount"))
       SXsort = SXscount;
    else
    if(0 == strcmp(sort,"atime"))
       SXsort = SXsatime;
    else
    if(0 == strcmp(sort,"vote"))
       SXsort = SXsvote;
    else
    if(0 == strcmp(sort,"docid"))
       SXsort = -SXsdocid;
    else
    if(0 == strcmp(sort,"docid-reverse"))
       SXsort = SXsdocid;
    else{
       SXsort = SXext;
       strncpy(SXsrt,sort,sizeof(SXsrt)-1);
    }
  }
  return SXsort;
}
class XResult {
public:
    short indexid;   //
    int oresx;       // offset in oresult
    int docid;       // document ID
    float weight;    // original weight to be printed
    char *Xweight;   // string value as URL 
    char *Xurl;      // URL as the last comparation value /* new-140619a */
    XResult() : indexid(0), docid(0), weight(0), Xweight(0) { }
};
#define iweight(rp) *(int*)(&(rp)->Xweight)
#define sweight Xweight

static int comp_score(XResult **r1, XResult **r2){
    return ((*r2)->weight > (*r1)->weight) ? 1 : -1;
}
inline int comp_docid(XResult *r1, XResult *r2){
    int diff;
    if( (diff = r1->indexid - r2->indexid) )
        return diff;
    return r1->docid - r2->docid;
}
static int comp_str(XResult **r1, XResult **r2){
    int diff;
    if( (diff = strcmp((*r1)->Xweight,(*r2)->Xweight)) )
        return diff;
    return comp_docid(*r1,*r2);
}
static int comp_forw(XResult **r1, XResult **r2){
    int diff;
    if( (diff = iweight(*r2) - iweight(*r1)) )
        return diff;
    if( (diff = strcmp((*r2)->Xurl,(*r1)->Xurl)) ) /* new-140619a */
        return diff;
    return comp_docid(*r1,*r2);
}
static int comp_rev(XResult **r1, XResult **r2){
    int diff;
    if( (diff = iweight(*r1) - iweight(*r2)) )
        return diff;
    if( (diff = strcmp((*r1)->Xurl,(*r2)->Xurl)) ) /* new-140619a */
        return diff;
    return comp_docid(*r1,*r2);
}
static XResult **SXsortXResults(XResult *result, int len){
    int (*func)(XResult**,XResult**);

    if(SXsort == 0)
        func = comp_score;
    else
    if(SXsort == SXsstr)
        func = comp_str;
    else
    if(0 < SXsort)
        func = comp_forw;
    else
        func = comp_rev;

    XResult **resultp;
    typedef XResult *XRP;
    resultp = new XRP[len];
    for(int i = 0; i < len; i++)
        resultp[i] = &result[i];

    qsort(resultp, len, sizeof(XResult*),
        (int(*)(const void*,const void*))func);

    return resultp;
}

int SX_pfxPath(char *pfx, const char *filepath, char *xpath, int xsize);
const char *shortDate(char *buf, size_t bufsiz, time_t &t);
int fileMtime(char*);

typedef struct {
    char     *name;      // name of the index
    int       Idxid;
    DescFile *descF;     // .dsc file
    int       mtime;     // modified time of .dsc file
    Result   *oresult;   // original Result info.
    XResult  *result;    // extended Result info. based on .dsc
    int       resultLen; // the number of results
    int       cached;    // oresult is loaded from cache
    int       sortsiz;   // from .srt file
    int      *sortval;   // from .srt file
 LexiconFile *Lex;       // .lex file
    ConText  *Ctx;       // .ctx file
    XMap     *Xmap;      // .map file
 CounterElem *CntShared; // .vot file for shared counter by URL /* new-140616j */
    int       CntSharedLen; // returned from SX_loadCounts()
 CounterElem *Cnt;       // .vot file for vote counter
    int       CntLen;
    int       CntCachex; // updated in this query
    int       CntCached; // updated in this query
 CounterElem  CntCache;  // updated in this query
    NumFile  *Nums;
} Index;
static Index Indexes[MAX_IDX];

/* new-140616j */
#define SHARED_COUNTER_SIZE (1024*1024)
static int sharedDid(const char *url){
    unsigned int crc32; /* fix-140621b unsigned not to make minus value of did */
    int did;

    crc32 = strCRC32(0,url);
    did = crc32 % SHARED_COUNTER_SIZE;
    return did;
}
static char *sharedBase(const char *name,char *base){
    char *dp;

    strcpy(base,name);
    if(dp = strrchr(base,'/'))
        sprintf(dp+1,"../_shared/byurl-%04dK",SHARED_COUNTER_SIZE/1024); /* mod-140621c */
    return base;
}

char *getUrl(Index *X1, int docid);
int getCnt(Index *X1, int docid){
    if(X1->CntShared){ /* new-140616j */
        int vote = 0;
        int xdocid;
        const char *url;
        url = getUrl(X1,docid);
        xdocid = sharedDid(url);
        if(0 <= xdocid && xdocid < X1->CntSharedLen){
            vote = X1->CntShared[xdocid].c_count;
        }
        return vote;
    }
    if(X1->Cnt == NULL)
        return 0;
    if(X1->CntCached == docid)
        return X1->CntCache.c_count;
    if(X1->CntLen < docid)
        return 0;
    return X1->Cnt[docid].c_count;
}
int getSCnt(Index *X1, int docid){
    if(X1->CntShared){ /* new-140616j */
        int hits = 0;
        int xdocid;
        const char *url;
        url = getUrl(X1,docid);
        xdocid = sharedDid(url);
        if(0 <= xdocid && xdocid < X1->CntSharedLen){
            hits = X1->CntShared[xdocid].c_acount;
        }
        return hits;
    }
    if(X1->Cnt == NULL)
        return 0;
    if(X1->CntLen < docid)
        return 0;
    return X1->Cnt[docid].c_acount;
}
int getAtime(Index *X1, int docid){
    if(X1->Cnt == NULL)
        return 0;
    if(X1->CntLen < docid)
        return 0;
    return X1->Cnt[docid].c_lastmod;
}

// cache must be refreshed if it is older than the executable
static int Mytime; // the time when this executable is created
static int refreshTime(char *path){
    int mtime = fileMtime(path);
    if(mtime < Mytime)
        mtime = Mytime;
    return mtime;
}

//-FX - scan index list in URL as "index=idx1 idx2"
//-FX - add increment indexes named as "idx+N"
//-FX - should suport "index=name*"
//-FX
static int Idxid;
static void initIndex(Index *X1, char *index1){
    char index1f[PATH_MAX];
    char numpath[PATH_MAX];
    char sharedcounter[PATH_MAX];

    X1->Ctx = 0;
    X1->Lex = 0;
    X1->Idxid = Idxid++;
    X1->name = strdup(index1);
    sprintf(index1f,"%s.idx",index1);
    X1->mtime = refreshTime(index1f);
    X1->CntCached = -1;
    X1->Cnt = SX_loadCounts(X1->name,"vote",X1->CntLen);
    X1->CntShared = SX_loadCounts(sharedBase(X1->name,sharedcounter),"vote",X1->CntSharedLen);
    sprintf(numpath,"%s%s",index1,NUM_EXT);
    X1->Nums = new NumFile(numpath);
}

static int scan_indexes(char *xlist, Index *Xlist){
    int nix = 0;
    char *indexp = xlist;
    for( int ix = 0; indexp && nix < MAX_IDX; ix++ ){
        char index1[PATH_MAX],*dp = index1;
        char *dx = &index1[sizeof(index1)/2-16];
        int ch;
        for(; (ch = *indexp); indexp++ ){
            if( ch == ' ' || ch == ',' ){
                indexp++;
                break;
            }
            if( dp < dx )
            *dp++ = ch;
        }
        *dp = 0;
        if( dp == index1 )
            break;

        SX_searchBase(index1,index1,sizeof(index1));
        initIndex(&Xlist[nix++],index1);
  
        for( int ax = 1; nix < MAX_IDX; ax++ ){
            char indexa[PATH_MAX],indexaf[PATH_MAX];
            sprintf(indexa,"%s+%d",index1,ax);
            sprintf(indexaf,"%s.idx",indexa);
            int fd;
            if( 0 <= (fd = open(indexaf,0)) ){
                close(fd);
                initIndex(&Xlist[nix++],indexa);
            }else{
                break;
            }
        }
    }
    return nix;
}

char *getAuthor(Index *X1, int docid){
    return X1->descF->GetAuthor(docid);
}
char *getUrl(Index *X1, int docid){
    return X1->descF->GetUrl(docid);
}
int getLastMod(Index *X1, int docid){
    if(X1->Nums){
        return X1->Nums->get(docid,N_LASTMOD);
    }
    return X1->descF->GetLastModified(docid);
}
int getSortVal(Index *X1, int docid){
    if(X1->sortsiz == 0){
        X1->sortsiz = -1;
        char path[PATH_MAX];
        int *mm;
        sprintf(path,"%s.%s.srt",X1->name,SXsrt);
        int size = SX_arrayMap(path,&mm);
        if( 0 < size && (int)sizeof(int) <= size ){
            X1->sortsiz = size / sizeof(int);
            X1->sortval = mm;
        }
    }
    if(0 <= docid && docid < X1->sortsiz){
        return ntohl(X1->sortval[docid]);
    }
    return 0;
}

//-SX - cache
//-SX
int SX_openCache(int wr, const char *Key, int Mtime, Fstream &cache);
int openCache(int wr, Index *X1, char *query, Fstream &cache){
    char Key[MAXQUERYLENGTH*2];
    sprintf(Key,"%s\n%x,%x%s,%s",query,SXopAND,SXsort,SXsrt,X1->name);
    return SX_openCache(wr,Key,X1->mtime,cache);
}
int lookupCache(Index *Xlist, int ix, char *query){
    Index *X1 = &Xlist[ix];
    int size;
    Fstream cache;

    X1->resultLen = -1;
    if(openCache(0,X1, query, cache) < 0)
        return 0;
    size = cache.size();
    if(size < 0)
        return 0;

    X1->oresult = new Result[size/sizeof(Result)];
    cache.read((char*)X1->oresult,size*sizeof(Result));
    X1->resultLen = size / sizeof(Result);
    X1->cached = 1;
    return 1;
}
//
//-SX - store result cache in sorted order
//
void storeCache(Index *Indexes, int nix, char *query, XResult **xres, int xlen){
    Fstream cache;
    for(int ix = 0; ix < nix; ix++){
        Index *X1 = &Indexes[ix];
        if(X1->cached)
            continue;
        if(openCache(1,X1, query, cache) == 0){
            int n = 0;
            for(int r = 0; r < xlen; r++){
                if(xres[r]->indexid == ix){
                    int oresx = xres[r]->oresx;
                    cache.write((char*)&X1->oresult[oresx],sizeof(Result));
                    n++;
                }
            }
            cache.close();
        }
    }
}

static void setFiles(Index *X1){
	// Lex and Ctx set in Retriever should be reused ...
    if(X1->Lex == NULL){
        char lexpath[PATH_MAX];
        strcpy(lexpath,X1->name);
        strcat(lexpath,".lex");
        X1->Lex = new LexiconFile(lexpath,"r");

        char ctxpath[PATH_MAX];
        strcpy(ctxpath,X1->name);
        strcat(ctxpath,".ctx");
        X1->Ctx = new ConText(ctxpath,"r");
	X1->Ctx->Idxid = X1->Idxid;

        char xmappath[PATH_MAX];
        strcpy(xmappath,X1->name);
        strcat(xmappath,".map");
        int fd = open(xmappath, O_RDONLY, 0666);
        if(0 <= fd)
            X1->Xmap = new XMap(fd);
        else
            X1->Xmap = 0;
    }
}
static int getDescList(Index *Xlist, int ix, int nix){
    Index *X1 = &Xlist[ix];
    char descpath[PATH_MAX];
    strcpy(descpath, X1->name); strcat(descpath, ".dsc");
    Fstream descfile(descpath,"r");
    if(! descfile.good())
        return -1;
  
    Fstream *descf = new Fstream(descpath,"r");
    X1->descF = new DescFile(*descf);

    int *mm;
    int size = SX_arrayMap(descpath,&mm);
    if( 0 < size && (int)sizeof(int) <= size ){
        X1->descF->setmmap(mm,size/sizeof(int));
    }

    X1->result = new XResult[X1->resultLen];

    XResult *xresult = X1->result;
    Result *result = X1->oresult;
    for(int r = 0; r < X1->resultLen; r++){
        xresult[r].oresx = r;
        xresult[r].indexid = ix;
        xresult[r].docid = result[r].docid;
        xresult[r].weight = result[r].weight;
    }

    if(SXsort){
        setFiles(X1);

        for(int r = 0; r < X1->resultLen; r++){
            int docid = result[r].docid;
            xresult[r].Xurl = getUrl(X1,docid); /* new-140619a */
            switch(SXsort){
                case  SXsdocid:
                case -SXsdocid:
                    iweight(&xresult[r]) = (ix<<26) | result[r].docid;
                    break;
                case  SXsauthor:
                    xresult[r].sweight = getAuthor(X1,docid);
                    break;
                case  SXsurl:
                    if(nix == 1){
                        iweight(&xresult[r]) = X1->Nums->get(docid,N_USORT);
                    }else{
                        xresult[r].sweight = getUrl(X1,docid);
                    }
                    break;
                case  SXsvote:
                    iweight(&xresult[r]) = getCnt(X1,docid);
                    break;
                case  SXsatime:
                    iweight(&xresult[r]) = getAtime(X1,docid);
                    break;
                case  SXscount:
                    iweight(&xresult[r]) = getSCnt(X1,docid);
                    break;
                case  SXsdate:
                case -SXsdate:
                    iweight(&xresult[r]) = getLastMod(X1,result[r].docid);
                    break;
                case  SXssize:
                    iweight(&xresult[r]) = X1->Nums->get(docid,N_SIZE);
/*
                    if(X1->Xmap)
                    iweight(&xresult[r]) = X1->Xmap->getdocsize(docid);
*/
                    break;
                case SXext:
                    iweight(&xresult[r]) = getSortVal(X1,result[r].docid);
                    break;
            }
        }
    }
    return 0;
}
//-FX - merge results
//-FX - sort by information in .dsc
//-FX - normarlize score
//-FX
static XResult *mergeResult(Index *Xlist, int nix, int ixn, int &resultLenp){
    if(ixn <= 0)
        return NULL;
    XResult *result = NULL;
    int resLen = 0;
    for( int ix = 0; ix < nix; ix++ ){
        if( Xlist[ix].resultLen <= 0 )
            continue;
        if(getDescList(Xlist,ix,nix) != 0){
            return NULL;
        }
        resLen += Xlist[ix].resultLen;
    }
    if(SXsort == SXsurl){
        if(nix == 1){
            SXsort = SXsint;
        }else{
            SXsort = SXsstr;
        }
    }
    if(SXsort == SXsauthor){
        SXsort = SXsstr;
    }
    if(ixn == 1){
        for( int ix = 0; ix < nix; ix++ ){
            if( 0 < Xlist[ix].resultLen ){
                result = Xlist[ix].result;
                break;
            }
        }
    }else{
        result = new XResult[resLen];
        int ro = 0;
        for( int ix = 0; ix < nix; ix++ ){
            if( Xlist[ix].resultLen <= 0 )
                continue;
            for( int r = 0; r < Xlist[ix].resultLen; r++ ){
                result[ro++] = Xlist[ix].result[r];
            }
        }
    }
    float maxweight = 0;
    int r;
    for(r = 0; r < resLen; r++){
        if(maxweight < result[r].weight)
            maxweight = result[r].weight;
    } 
    double mag = 999 / maxweight;
    for(r = 0; r < resLen; r++){
        result[r].weight *= mag;
    }
    resultLenp = resLen;
    return result;
}
static void findDocid(XResult **result, int resultLen){
    if(param_docid[0]){
        // indexid.docid
        int idxid = 0,docid = 0;
        sscanf(param_docid,"%d.%d",&idxid,&docid);
        for(int r = 0; r < resultLen; r++){
            if(result[r]->indexid == idxid && result[r]->docid == docid){
                param_from = r+1;
/*
                param_from = r;
*/
                break;
            }
        }
    }
}

////////////////////////////////////////////////////////////////
int putComment(const char *url, const char *from, const char *comment);
int getComment(const char *url, const char *from, char *comment, int size);

char resultFroms[16];
char resultTos[16];

typedef struct {
    char *Nexts;
    char Next1[256];
    char Prev1[256];
    char Indexes[NAME_MAX*3];
    char Cgipath[128];
    char Pfx[128];
    char Sxop[128];
    char Dispnum[16];
    char Sort[128];
    char Style[128];
    char Zoom[16];
    char Ssk[16];
    char Params[128];
    char Tidxs[16];
    char Tdocs[16];
    char Twords[16];

    char Even[2];
    char Docid[16];
    char Idxname[64];
    char Checked[16];
    char Serial[16];
    char Score[16];
    char ScoreA[8];
    char Size[16];
    char Vote[32];
    char Scount[32];
    char URL[DSCBLOCKSIZE + 128];
    char Author[DSCBLOCKSIZE + 128];
    char Title[DSCBLOCKSIZE + 128];
    char Digest[1024];
/*
    char Context[1024];
*/
    char Context[4096];
    char Date[64];
    char Lastmod[64];
    char ShortDate[32];

    Index *X1;
    int docid;
    Result res;
} ResultDisp;
static ResultDisp resultDisp;

/* introduced in 0.99.17 */
static char *baseof(const char *url){
	char base[1024],*p;
	strncpy(base,url,sizeof(base));
	if( (p = strrchr(base,'/')) != 0  )
		*p = 0;
	return strdup(base);
}
char *getresbuf(const char *name){
    if( strcmp(name,"zoom") == 0 )     return resultDisp.Zoom;
    if( strcmp(name,"ssk") == 0 )      return resultDisp.Ssk;
    if( strcmp(name,"style") == 0 )    return resultDisp.Style;
/*
    if( strcmp(name,"hidden_key") == 0)return hidden_key;
*/
    if( strcmp(name,"even") == 0 )     return resultDisp.Even;
    if( strcmp(name,"from") == 0 )     return resultFroms;
    if( strcmp(name,"checked") == 0 )  return resultDisp.Checked;
    if( strcmp(name,"vote") == 0 )     return resultDisp.Vote;
    if( strcmp(name,"scount") == 0 )   return resultDisp.Scount;
    if( strcmp(name,"date") == 0 )     return resultDisp.Date;
    if( strcmp(name,"sdate") == 0 )    return resultDisp.ShortDate;
    if( strcmp(name,"lastmod") == 0 )  return resultDisp.Lastmod;
    if( strcmp(name,"author") == 0 )   return resultDisp.Author;
    if( strcmp(name,"digest") == 0 )   return resultDisp.Digest;
    if( strcmp(name,"context") == 0 )  return resultDisp.Context;
    if( strcmp(name,"title") == 0 )    return resultDisp.Title;
    if( strcmp(name,"escindex") == 0 ) return resultDisp.Indexes;
    if( strcmp(name,"cgibase") == 0 ){
    	/* fix-140527i this code is not used for substitution of $cgibase */
					return baseof(resultDisp.Cgipath);
    }
    if( strcmp(name,"cgipath") == 0 )  return resultDisp.Cgipath;
    if( strcmp(name,"sxop") == 0 )     return resultDisp.Sxop;
    if( strcmp(name,"sort") == 0 )     return resultDisp.Sort;
    if( strcmp(name,"num") == 0)       return resultDisp.Serial;
    if( strcmp(name,"dispnum") == 0)   return resultDisp.Dispnum;
    if( strcmp(name,"n") == 0 )        return resultDisp.Dispnum;
    if( strcmp(name,"pfx") == 0 )      return resultDisp.Pfx;
    if( strcmp(name,"docid") == 0 )    return resultDisp.Docid;
    if( strcmp(name,"escurl") == 0 )   return resultDisp.URL;
    if( strcmp(name,"params") == 0 )   return resultDisp.Params;
    if( strcmp(name,"next1") == 0 )    return resultDisp.Next1;
    if( strcmp(name,"prev1") == 0 )    return resultDisp.Prev1;
    return 0;
}
void setrv(const char *name, const char *val){
    char *valbuf;
    if( (valbuf = getresbuf(name)) ){
        strcpy(valbuf,val);
    }
}
char *getrvx(const char *name){
    char *val;
    if( (val = getresbuf(name)) )
       if(*val != 0)
          return val;
    return 0;
}
char *getrv(const char *name)
{   char *val;
    if( (val = getresbuf(name)) ) return val;
    if( strcmp(name,"index") == 0 )    return param_index;
    if( strcmp(name,"pc") == 0 )       return param_pc;
    if( strcmp(name,"next") == 0 )
        if( resultDisp.Nexts != NULL )  return "1";
    return "";
}

static void setResultDisp(int n, DescRecord &rec, int score){
    resultDisp.Even[0] = ((n%2) == 0) ? ' ' : 0;
    sprintf(resultDisp.Serial,"%02d",n);
    sprintf(resultDisp.Score,"%3d",score);
    if(score < 0 )
        score = 0;
    if(1000 <= score)
        score = 999;
    if(score == 999)
        strcpy(resultDisp.ScoreA,"A");
    else
    sprintf(resultDisp.ScoreA,"%d",score/100);

    resultDisp.Size[0] = 0;
    sprintf(resultDisp.Size,"%d",(int)rec.getSize());

    resultDisp.Author[0] = 0;
    if( const char *author = rec.getAuthor() )
        html_entitize(resultDisp.Author,author);

    resultDisp.Digest[0] = 0;
    if( const char *digest = rec.getDigest() )
        html_entitize(resultDisp.Digest,digest);

    resultDisp.Title[0] = 0;
    if( const char *title = rec.getTitle() )
        html_entitize(resultDisp.Title,title);

    resultDisp.Date[0] = 0;
    if( time_t date = rec.getDate() ){
        rfc822date(resultDisp.Date, sizeof(resultDisp.Date)-1, date);
        shortDate(resultDisp.ShortDate, 32, date);
    }
    resultDisp.Lastmod[0] = 0;
    resultDisp.ShortDate[0] = 0;
    if( time_t lastmod = rec.getLastModified() ){
        rfc822date(resultDisp.Lastmod, sizeof(resultDisp.Lastmod)-1, lastmod);
        shortDate(resultDisp.ShortDate, 32, lastmod);
    }

    if( const char *upath = rec.getUrl() ){
        char *escurl = escape_url(upath);
        strcpy(resultDisp.URL,escurl);
        delete []escurl;
    }else
        resultDisp.URL[0] = 0;
}
static void
getContext(Index *X1,int docid,Result &res, char *buf,int siz, int win,char *del){
    *buf = 0;
/*
    if(param_key[0]==0 || param_key[0]!=' '){
        return;
    }
*/
    if(X1->Lex != NULL && X1->Ctx->good() ){
/*
         int base = 0;
*/
         off_t base = 0;
         if(X1->Xmap != NULL){
              base = X1->Xmap->getdocbase(docid);
         }
         int nelem;
         pos_t list[8];
         nelem = sgetintvl(res.list,sizeof(res.list),list,8);
         if(nelem == 0)
             return;

         for(int i = 0; i < nelem; i++){
             list[i] += base;
         }
         int winsize;
         winsize = win / nelem + 2;
         X1->Ctx->getContext(X1->Lex,list,nelem, buf,siz,winsize,del);
    }
}
static void
setContext(Index *X1, int docid, Result &res){
    resultDisp.Context[0] = 0;
    if(res.list[0] == 0)
        return;
    setFiles(X1);
    getContext(X1,docid,res,resultDisp.Context,sizeof(resultDisp.Context),10," .. ");
}


int condSubsts(char *line, int size, char *(*getrv)(const char *));
extern char ContentType[64];
void SX_display(char *line, int size){
    if(strchr(line,'$') == NULL)
        return;
    char buf[1024],xbuf[4*1024];

    condSubsts(line,LINE_MAX,getrv);
    replace_string(line, "$version", FREYABANNER);
    replace_string(line, "$num",     resultDisp.Serial);
    replace_string(line, "$scoreA",  resultDisp.ScoreA);
    replace_string(line, "$score",   resultDisp.Score);
    replace_string(line, "$escurl",  resultDisp.URL);
    replace_string(line, "$size",    resultDisp.Size);

    replace_string(line, "$escindex",resultDisp.Indexes);
    replace_string(line, "$cgipath", resultDisp.Cgipath);

    /* fix-140527i */
    replace_string(line, "$cgibase", baseof(resultDisp.Cgipath));
    replace_string(line, "$prev1",   resultDisp.Prev1);
    replace_string(line, "$next1",   resultDisp.Next1);

    if( resultDisp.Nexts != NULL )
    replace_string(line, "$next",    resultDisp.Nexts);

    if( strstr(line,"$etime") ){
        char etime[64];
        sprintf(etime,"%0.3f",Time()-Start);
        replace_string(line, "$etime", etime);
    }
    replace_string(line, "$docid",   resultDisp.Docid);
    replace_string(line, "$idxname", resultDisp.Idxname);
    replace_string(line, "$from",    resultFroms);
    replace_string(line, "$to",      resultTos);
    replace_string(line, "$pfx",     resultDisp.Pfx);
    replace_string(line, "$sort",    resultDisp.Sort);
    replace_string(line, "$sxop",    resultDisp.Sxop);
    replace_string(line, "$style",   resultDisp.Style);
    replace_string(line, "$zoom",    resultDisp.Zoom);
    replace_string(line, "$ssk",     resultDisp.Ssk);
    if(strstr(line,"$params1")){
        sprintf(xbuf,"%s&%s=%s",resultDisp.Params,"docid",resultDisp.Docid);
        replace_string(line, "$params1", xbuf);
    }
    replace_string(line, "$params",  resultDisp.Params);
    replace_string(line, "$tidxs",   resultDisp.Tidxs);
    replace_string(line, "$tdocs",   resultDisp.Tdocs);
    replace_string(line, "$twords",  resultDisp.Twords);

    replace_string(line, "$vote",    resultDisp.Vote);
    replace_string(line, "$scount",  resultDisp.Scount);
    replace_string(line, "$date",    resultDisp.Date);
    replace_string(line, "$sdate",   resultDisp.ShortDate);
    replace_string(line, "$lastmod", resultDisp.Lastmod);
    replace_string(line, "$author",  resultDisp.Author);
    replace_string(line, "$title",   resultDisp.Title);
    replace_string(line, "$digest",  resultDisp.Digest);

    if(strstr(line,"$contextx") != 0){
        getContext(resultDisp.X1,resultDisp.docid,resultDisp.res, xbuf,sizeof(xbuf),30," .. <BR> ..\n");
        replace_string(line, "$contextx", xbuf);
    }
    replace_string(line, "$context", resultDisp.Context);

    if(strstr(line,"$atime") != 0){
        time_t t;
        t = getAtime(resultDisp.X1,resultDisp.docid);
        if(t == 0){
            strcpy(buf,"-");
        }else{
            strftime(buf,sizeof(buf),"%a, %d %b %Y %H:%M:%S %Z",localtime(&t));
        }
        replace_string(line,"$atime",buf);
    }
    if(strstr(line,"$comment") != 0){
        getComment(resultDisp.URL,"",buf,sizeof(buf));
        html_entitize(xbuf,buf);
        replace_string(line, "$comment", xbuf);
    }
}

//-FX - EXTENSTIONS for FreyaSX -------------------------------------------

int h_hilight(char *text,int size);
int StopDisplay = 0;

int remove_tailkey(char *key, const char *xkey){
	int xklen = strlen(xkey);
	char *kp;
	int ex = 0;

	for( kp = key; *kp; kp++ ){
		if( *kp == ' ' && strncmp(kp+1,xkey,xklen) == 0 && kp[1+xklen] == 0 ){
			*kp = 0;
			ex = 1;
			break;
		}else
		if( strncmp(kp,xkey,xklen) == 0 && kp[xklen] == 0 ){
			*kp = 0;
			ex = 1;
			break;
		}
	}
	return ex;
}

int
display_html(const char *filepath,
             const char *title, const char *query, const char *errormsg,
             int resultNum, int lev=0 ){

	if(8 < ++lev){
		fprintf(stderr,"display_html: too deep inclusion (%d)[%s]\n",
			lev,filepath);
		return 0;
	}
	if(StopDisplay)
		return 0;

	char xpath[PATH_MAX];
	if( 0 < SX_pfxPath(param_pfx,filepath,xpath,sizeof(xpath)) )
		filepath = xpath;

//
// should try "pub-filepath" if "filepath" not found
//

    FILE *htmlfile = fopen(filepath,"r");
    if( htmlfile == NULL )
        return -1;

    char qtitle[LINE_MAX];
    char qerrormsg[LINE_MAX];
    char qquery[LINE_MAX];
    char qindex[LINE_MAX];
    html_entitize(qtitle, title);
    html_entitize(qerrormsg, errormsg);
    while(*query == ' ')
      query++;
    html_entitize(qquery, query);
    remove_tailkey(qquery,param_hidden_key); /* new-140601r */
    html_entitize(qindex, param_index);

    char resultnum[32];
    sprintf(resultnum," %d",resultNum);
    char dispnum[32];
    sprintf(dispnum,"%d",param_n);

    h_hilight(resultDisp.Context,sizeof(resultDisp.Context));
/* -- h_hilight must see the buf. size ... */
    h_hilight(resultDisp.Digest,sizeof(resultDisp.Digest));

/*
    char linebuf[0x8000]; // must be boundary checked in replace_sring()
*/
    char linebuf[0x20000]; // must be boundary checked in replace_sring()
                          // it can be overflow on substution of $next or so

    for(;;){
        if( fgets(linebuf,LINE_MAX,htmlfile) == NULL )
            break;

        char *dp = strpbrk(linebuf,"\r\n");
        if( dp ) *dp = 0;

        // disable conversion to user-origined data
        int noconv = strchr(linebuf,'$')
                 && (strstr(linebuf,"$title") != NULL
                  || strstr(linebuf,"${title") != NULL
                  || strstr(linebuf,"$digest") != NULL
                  || strstr(linebuf,"$context") != NULL);

        replace_string(linebuf, "$title", qtitle);
        replace_string(linebuf, "$errormsg", qerrormsg);
        replace_string(linebuf, "$query", qquery);
        replace_string(linebuf, "$hidden_key", param_hidden_key);
        replace_string(linebuf, "$index", qindex);
        replace_string(linebuf, "$result", resultnum);
        replace_string(linebuf, "$dispnum", dispnum);

        char name[LINE_MAX];
        if(SX_include(name,sizeof(name),linebuf)){
            if(display_html(name,title,query,errormsg,resultNum,lev)!=0)
                sout_line("<!-- ## Failed loading a file ## -->");
            if( linebuf[0] == 0 )
                continue;
        }
        scan_META(linebuf,sizeof(linebuf));
        SX_display(linebuf,sizeof(linebuf));

        if(!noconv){
          if(SX_include(name,sizeof(name),linebuf)){
            if(display_html(name,title,query,errormsg,resultNum,lev)!=0)
                sout_line("<!-- ## Failed loading a file ## -->");
            if( linebuf[0] == 0 )
                continue;
          }
          SX_enConv(replace_string,linebuf);
        }
        sout_line(linebuf);

        if(strcmp(linebuf,"</HTML>") == 0){
            StopDisplay = 1;
            break;
        }
    }
    fclose(htmlfile);
    return 0;
}

void putError(const char *msg1, const char *msg2);

void
cgi_error(const char *msg1, const char *msg2 = NULL,
          const char *query = NULL, int log = 1){
    if(query == NULL) query = "";
    if(msg2 == NULL) msg2 = "";
    if(-1 == display_html("error.html", msg1, query, msg2, 0)){
        putError(msg1,msg2);
    }
    if(log) writelog(query, -1);
}

static void set_cgipath(){
    char *cgipath = getenv("SCRIPT_NAME");
    if(NULL == cgipath || '\0' == cgipath[0]){
        // We couldn't get the path (Why?), try another way.
        cgipath = getenv("REQUEST_URI");
        if(NULL == cgipath || '\0' == cgipath[0])
            cgipath = "fsearch.cgi";
        else{
            // in order to get CGI's path, we replace the first '?' with NUL
            char *cp = cgipath;
            while(*cp){
                if(*cp == '?'){ *cp = '\0'; break; }
                cp++;
            }
        }
    }
    strncpy(resultDisp.Cgipath,cgipath,sizeof(resultDisp.Cgipath)-1);
}

int wordisin(char *str, char *word){
    int len = strlen(word);
    for(char *sp = str; *sp;){
        while(isspace(*sp))
            sp++;
        if(*sp == 0)
            break;
        if(strncmp(sp,word,len) == 0){
            if(sp[len] == 0 || isspace(sp[len]))
                return 1;
        }
        /*
        while(!isspace(*sp))
        */
        while(*sp && !isspace(*sp))
            sp++;
    }
    return 0;
}

int
get_query(char *key){
    char *method = getenv("REQUEST_METHOD");
    if(! (method && 0 == strcmp(method, "GET"))) {
        if(method){
            cgi_error("Error", "unsupported method");
        }else{
            cgi_error("Error", "no method provided");
        }
        return 1;
    }
    char *qs = getenv("QUERY_STRING");
    if(NULL == qs || '\0' == qs[0]){
        cgi_error("Error", "No query information to decode.");
        return 1;
    }
    qs = strdup(qs); // for repetitive get_query() after setConv()
    char *k, *v;
    char *idx = param_index;
    while((k = strtok(qs, "&"))){
        qs = NULL;
        v = strchr(k, '=');
        if(v != NULL){
            *v++ = '\0';
            unescape_url(v);

            if(strcmp(k,"key") != 0)
            if(strcmp(k,"hidden_key") != 0)
            SX_deConv(replace_string,k,v);
            if(0 == strcmp(k, "sort")){
                strncpy(param_sort, v, sizeof(param_sort) - 1);
            }else if(0 == strcmp(k, "sxop")){
                strncpy(param_sxop, v, sizeof(param_sxop) - 1);
            }else if(0 == strcmp(k, "url")){
                strncpy(param_url, v, sizeof(param_url) - 1);
            }else if(0 == strcmp(k, "pfx")){
                strncpy(param_pfx, v, sizeof(param_pfx) - 1);
            }else if(0 == strcmp(k, "pc")){
                strncpy(param_pc, v, sizeof(param_pc) - 1);
            }else if(0 == strcmp(k, "dbg")){
                strncpy(param_dbg, v, sizeof(param_dbg) - 1);
            }
            else if(0 == strcmp(k, "docid")){
                strncpy(param_docid, v, sizeof(param_docid) - 1);
            }
            else if(0 == strcmp(k, "jump")){
                strncpy(param_jump, v, sizeof(param_jump) -1);
            }
            else if(0 == strcmp(k, "find")){
                strncpy(param_find, v, sizeof(param_find) -1);
            }
            else if(0 == strcmp(k, "vote")){
                strncpy(param_vote, v, sizeof(param_vote) - 1);
            }
            else if(0 == strcmp(k, "ssk")){
                param_ssk = atoi(v);
            }
            else if(0 == strcmp(k, "frame")){
                strncpy(param_frame, v, sizeof(param_frame) - 1);
            }
            else if(0 == strcmp(k, "comm")){
                strncpy(param_comm, v, sizeof(param_comm) - 1);
            }
            else if(0 == strcmp(k, "zoom")){
                if(strcmp(v,"small")==0) param_zoom = -1; else
                if(strcmp(v,"big"  )==0) param_zoom =  1; else
                param_zoom = atoi(v);
            }

            if(0 == strcmp(k, "key")){
                if(NULL == conv2euc(key, v, CODE_UNKNOWN)){
                    strncpy(key, v, MAXQUERYLENGTH);
                }
	    }else
            if(0 == strcmp(k, "hidden_key")){ /* new-140601r */
		char *kp;
		/*
		 - should not be depend the order of key and hidden_key
		 - hidden_key should be normalized
		 */
                if(NULL == conv2euc(param_hidden_key, v, CODE_UNKNOWN)){
		    strncpy(param_hidden_key, v, MAXQUERYLENGTH);
		}
		kp = key + strlen(key);
		if( (kp - key)+1+strlen(param_hidden_key) < MAXQUERYLENGTH ){
		    if( key < kp )
		    	*kp++ = ' ';
                    strncpy(kp, v, MAXQUERYLENGTH-(kp-key));
                }
            }else if(0 == strcmp(k, "from")){
                param_from = atoi(v);
            }else if(0 == strcmp(k, "index")){
                if(param_index < idx){
                 if(wordisin(param_index,v) || *v == 0){
                 }else{
                   int len = idx - param_index;
                   int rem = sizeof(param_index) - len;
                   *idx++ = ' ';
                   strncpy(idx, v, rem - 2);
                   idx += strlen(idx);
                 }
                }else{
                strncpy(param_index, v, sizeof(param_index) - 1);
                idx = param_index + strlen(param_index);
                }
            }else if(0 == strcmp(k, "style")){
                strncpy(param_style, v, sizeof(param_style) - 1);
            }else if(0 == strcmp(k, "n")){
                param_n = atoi(v);
            }
        }
    }

    return 0;                   // success
}

int
printtemplate(const char *style, int n, DescRecord &rec, const XResult *result){
    if(strcmp(param_docid,resultDisp.Docid) == 0)
        strcpy(resultDisp.Checked,"1");
    else
        strcpy(resultDisp.Checked,"");
    int score = (int)result->weight;

Index *X1 = &Indexes[result->indexid];
int rank = 0;
    if(X1->CntShared){ /* new-140616j */
        rank = getCnt(X1,result->docid);
    }else
if(result->indexid == X1->CntCachex && result->docid == X1->CntCached){
    rank = X1->CntCache.c_count;
}else{
  if(result->docid < X1->CntLen){
    rank = X1->Cnt[result->docid].c_count;
  }
}
sprintf(resultDisp.Vote,"%s%d",0<rank?"+":"",rank);
strcpy(resultDisp.Scount,"0");

    if(X1->CntShared){ /* new-140616j */
        int hits = getSCnt(X1,result->docid);
        sprintf(resultDisp.Scount,"%s%d",0<hits?"*":"",hits);
    }else
if(result->docid < X1->CntLen){
    int hits = X1->Cnt[result->docid].c_acount;
    sprintf(resultDisp.Scount,"%s%d",0<hits?"*":"",hits);
}

    resultDisp.X1 = X1;
    resultDisp.docid = result->docid;
    resultDisp.res = X1->oresult[result->oresx];
    setResultDisp(n,rec,score);
    const char *title = rec.getTitle();
    if( title == 0 ) title = "";

    char *name = strrchr(X1->name,'/');
    name = name != NULL ? name + 1 : X1->name;
    html_entitize(resultDisp.Idxname,name);

    char path[PATH_MAX];
    sprintf(path,"style-%s.html",style);
    int code = display_html(path, title, param_key, "(no error)", 0);

    if(code == 0)
        return 0;
    else
        return -1;
}

int
print_results(Index *Xlist, XResult **result, const char *key,
              int resultNum, int disp, int from, int to){
    assert(to >= from);
    assert(to - from < resultNum);

    int r;
    unsigned int prevsize = 0;
    int prevcrc32 = -1;
    for(r = from; r <= to; r++){
        Index *X1 = &Xlist[result[r]->indexid];
        int oresx = result[r]->oresx;
        setContext(X1,result[r]->docid,X1->oresult[oresx]);
        DescRecord rec;
        Xlist[result[r]->indexid].descF->getRecord(rec, result[r]->docid);
        /*
        descf.getRecord(rec, result[r].docid);
        */

/*
        if(rec.getSize() == prevsize && rec.getCRC32() == prevcrc32){
        sout << "<!-- DUPLICATE CRC32=" << rec.getCRC32() << "-->" << endl;
            continue;
        }
*/
        prevsize = rec.getSize();
        prevcrc32 = rec.getCRC32();

        char line[LINE_MAX];
        sprintf(resultDisp.Docid,"%d.%d",result[r]->indexid,result[r]->docid);
        sprintf(line,"<!-- id: %s -->",resultDisp.Docid);
        sout_line(line);
        if( printtemplate(param_style,r+1,rec,result[r]) != 0 ){
           printGeneric();
        }
    }
    
    //
    // print footer references.
    //
    return 0;
}

static int countup(int xid, int did, int inc, const char *name){
    Index *X1 = &Indexes[xid];
    char crc8,*addr;
    crc8 = inc;
    if((addr = getenv("REMOTE_ADDR")))
        crc8 = strCRC8(crc8,addr,strlen(addr));
    CounterElem *ce = SX_incCount(X1->name,name,did,inc,crc8);
    if(ce != NULL){
         X1->CntCachex = xid;
         X1->CntCached = did;
         X1->CntCache = *ce;
    }
    return 0;
}

/* new-140616j */
static int countupByURL(int xid, const char *url, int inc, const char *name){
    Index *X1 = &Indexes[xid];
    char crc8,*addr;
    int did;
    char base[1024];

    crc8 = inc;
    if((addr = getenv("REMOTE_ADDR")))
        crc8 = strCRC8(crc8,addr,strlen(addr));

    did = sharedDid(url);
    sharedBase(X1->name,base);
    CounterElem *ce = SX_incCount(base,name,did,inc,crc8);
    return 0;
}

int mainx(int argc, char *argv[]);
int
main(int argc, char *argv[]){
    double now = Time();
    int rands = (int)(0x1000*(now-(int)now)) % 0x1000;
#ifdef _MSC_VER
    char *randstack = (char*)alloca(4*rands);
#else
    int randstack[rands];
#endif
    randstack[0] = 0;
    strcpy(ContentType,"text/html");
    mainx(argc,argv);
    sout_flush();
    return 0;
}
int
mainx(int argc, char *argv[]){
    Start = Time();
    param_from = 0;
    param_n = RESULTSFORDISPLAY;
    strcpy(param_index, "default");
    strcpy(param_style, "html");
    strcpy(param_sort, "score");
    strcpy(param_sxop, "AND");

    memset(&resultDisp,0,sizeof(resultDisp));
    set_cgipath(); // before any (error) response is generated using $cgipath
    makeParams("",0); // for error message

    param_key[0] = '\0';
    if(0 != get_query(param_key)){ // error
        return 0;
    }
    delSpaces(param_index);

    // to make cout efficient
    if(strstr(param_dbg,"sync") == NULL){
        setAsyncIO();
    }
    int reload = 0;
    if(strstr(param_dbg,"nocache") != NULL)
        reload |= 1;
    char *pragma = getenv("HTTP_PRAGMA");
    if(pragma != NULL && strstr(pragma,"no-cache") != NULL)
        reload |= 2;

    char ibuf[LINE_MAX], *e = ibuf; 
    for(char *p = param_index; *p; p++)
        *e++ = *p == ' ' ? '+' : *p;
    *e = 0;
    html_entitize(resultDisp.Indexes,ibuf);
    html_entitize(resultDisp.Pfx,param_pfx);
    html_entitize(resultDisp.Sxop,param_sxop);
    html_entitize(resultDisp.Sort,param_sort);
    html_entitize(resultDisp.Style,param_style);
    sprintf(resultDisp.Dispnum,"%d",param_n);
    resultDisp.Zoom[0] = 0;
    if(0 < param_ssk){
        sprintf(resultDisp.Ssk,"%d",param_ssk);
    }
    if(param_zoom < 0){
        strcpy(resultDisp.Zoom,"small");
    }else
    if(0 < param_zoom){
        strcpy(resultDisp.Zoom,"big");
    }
    makeParams(param_key,0); // for error message

    if(param_frame[0]){
        if(strcmp(param_frame,"main") == 0){
            if(display_html("frames.html", "", param_key, "(no error)", 0) == 0){
                return 0;
            }
        }
    }
    if(param_comm[0]){
        putComment("",getenv("REMOTE_HOST"),param_comm);
    }
    if(param_pfx[0]){
        SX_setConv(param_pfx);// setup text conversion.
        param_key[0] = 0;
        param_hidden_key[0] = 0;
        get_query(param_key); // re-scan query with the conversion.
        delSpaces(param_index);
    }
    if(param_url[0]){
        char path[4+sizeof(param_url)];
        strcpy(path,"pub-");
        sscanf(param_url,"%32[-_a-zA-Z0-9.]",path+strlen(path));
        if(display_html(path, "", param_key, "(no error)", 0) == 0)
          return 0;
    }
    SXsort = scan_sort(param_sort);
    if(0 == strcmp(param_sxop,"OR"))
        SXopAND = 0;

    param_from++;
/*
    if(param_key[0] == '\0'){
        if(0 == strcmp(param_pfx,"ja"))
        cgi_error("顼: Ĥޤ",
                  "ϤƤ", "");
        else
        cgi_error("Error: empty query","input a query", "");
        return 0;
    }
*/
    if(param_n < 0 || param_n > MAXRESULTSFORDISPLAY){
        param_n = RESULTSFORDISPLAY;
        if(0 == strcmp(param_pfx,"ja"))
        cgi_error("顼: ϰϤǤ",
                  "n=...ŬڤͤˤƤ", "");
        else
        cgi_error("Error: Invalid range of output",
                  "n=...must be a proper value", "");
        return 0;
    }

    //
    // search
    //
    char uquery[MAXQUERYLENGTH + 1];
    const Unifier unif;
    unif.unify(uquery, param_key);
    ComplexExpression cexp(uquery);
    if(param_key[0] != 0)
    if(cexp.error()){
        if(0 == strcmp(param_pfx,"ja"))
        cgi_error("顼: Ǥ",
                  cexp.errmsgs[cexp.error()], uquery);
        else
        cgi_error("Error: Invalid query expression",
                  cexp.errmsgs[cexp.error()], uquery);
        return 2;
    }

  Mytime = fileMtime(argv[0]); // must be set before cache scan by scan_indexes()
  /////////////////////////////////////
  int resultLen = -1;
  XResult *result = NULL;
  int nix = scan_indexes(param_index, Indexes);
  int ixn = 0; // indexes with non-empty result

  if(param_find[0]){
    int ix;
    int did = -1;
    for(ix = 0; ix < nix; ix++){
        Index *X1 = &Indexes[ix];
        if(0 <= (did = findUrl(*X1->Nums,param_find)))
            break;
    }

    char url[LINE_MAX];
    sprintf(url,"%s?%s&docid=%d.%d",resultDisp.Cgipath,
        resultDisp.Params,ix,did);
    putMoved(url);
    return 0;
  }

  // docid=url.URL -> docid=xid.did
  int docid_xid = -1;
  int docid_did = -1;
  if(param_docid[0]){
    if(strncmp(param_docid,"url.",4) == 0){
      char *url = param_docid + 4;
      for(int ix = 0; ix < nix; ix++){
          Index *X1 = &Indexes[ix];
          if(0 <= (docid_did = findUrl(*X1->Nums,url))){
              docid_xid = ix;
              break;
          }
      }
      if(0 <= docid_did){
        sprintf(param_docid,"%d.%d",docid_xid,docid_did);
      }
    }
  }
  if(param_jump[0]){
    int xid,did;
    char *url = param_jump;
    if( sscanf(param_jump,"%d.%d",&xid,&did) == 2 ){
        docid_xid = xid;
        docid_did = did;
        char *dp;
        if((dp = strchr(param_jump,'.')) != 0)
            if((dp = strchr(dp+1,'.')) != 0)
                url = dp+1;
    }
    if(0 <= docid_xid && 0 <= docid_did){
        countup(docid_xid,docid_did,0,"vote");
        countupByURL(docid_xid,url,0,"vote"); /* new-140616j */
    }
    putMoved(url);
    return 0;
  }
  if(param_vote[0]){
    char name[32];
    char urlb[LINE_MAX];
    int xid,did,inc;
    strcpy(name,"vote");
    strcpy(urlb,"");
    /*
    if(isdigit(*param_vote)&&sscanf(param_vote,"%d.%d.%d",&xid,&did,&inc) == 3
     || sscanf(param_vote,"%20[^.].%d.%d.%d",name,&xid,&did,&inc) == 4){
    */
    if(isdigit(*param_vote)&&sscanf(param_vote,"%d.%d.%d.%s",&xid,&did,&inc,urlb) >= 3
     || sscanf(param_vote,"%20[^.].%d.%d.%d.%s",name,&xid,&did,&inc,urlb) >= 4){
        if(0 <= xid && xid < nix){
            countup(xid,did,inc,name);
            if(urlb[0]){
                countupByURL(xid,urlb,inc,name); /* new-140616j */
            }
        }
        char url[LINE_MAX];
        sprintf(url,"%s?%s&docid=%d.%d",
            resultDisp.Cgipath,resultDisp.Params,xid,did);
        putMoved(url);
        return 0;
    }
  }
  if( reload == 0 ){
    for( int ix = 0; ix < nix; ix++ ){
      lookupCache(Indexes,ix,uquery);
    }
  }

  for( int ix = 0; ix < nix; ix++ ){
    char *param_index = Indexes[ix].name;

    int resultLen = -1;
    Result *result = NULL;
    // result = checkCache(uquery, resultLen);
    if( Indexes[ix].cached ){
        resultLen = Indexes[ix].resultLen;
        result = Indexes[ix].oresult;
    }
    if(-1 == resultLen){        // it didn't hit cache
        char indexpath[PATH_MAX];
        char lexpath[PATH_MAX];
        char xmappath[PATH_MAX];
        strcpy(indexpath, param_index); strcat(indexpath, ".idx");
        strcpy(lexpath, param_index); strcat(lexpath, ".lex");
        strcpy(xmappath, param_index); strcat(xmappath, ".map");

        IndexFile idxfile(indexpath,"r");
        if(idxfile.geterror()){
            cgi_error("Error", "bad index file.", param_key);
            return 1;
        }
        LexiconFile lexfile(lexpath,"r");

        int fd = open(xmappath, O_RDONLY, 0666);
        if(fd == -1){
            cgi_error("Error", "cannot open map file", param_key);
            return 1;
        }

        // setup retriever object
        XMap map(fd);
        Retriever retr(lexfile, idxfile, map);
        Evaluator evaluator;           // use standard ranking policy

        retr.Idxid = ix;
        char ctxpath[PATH_MAX];
        snprintf(ctxpath,sizeof(ctxpath),"%s.ctx",param_index);
        retr.Ctx = new ConText(ctxpath,"r");
	retr.Ctx->lexf = &lexfile;

        result = cexp.eval(retr, evaluator, resultLen);
/*
normalize "weight" after all indexes are searched and merged.
        if(resultLen > 0) evaluator.rank(result, resultLen);
*/
        // saveCache(result, resultLen);
    }
    Indexes[ix].oresult = result;
    Indexes[ix].resultLen = resultLen;
    if(0 < resultLen)
        ixn++;
  }

  int ndoc = 0;
  int i;
  for(i = 0; i < nix; i++){
      setFiles(&Indexes[i]);
      ndoc += Indexes[i].Xmap->getdocNum();
  }
  int nlex = 0;
  for(i = 0; i < nix; i++){
      nlex += Indexes[i].Lex->nentry();
  }
  sprintf(resultDisp.Tidxs,"%d",nix);
  sprintf(resultDisp.Tdocs,"%d",ndoc);
  sprintf(resultDisp.Twords,"%d",nlex);

  if(param_key[0] == 0){
      int nres;
      for(int xi = 0; xi < nix; xi++){
          nres = Indexes[xi].Xmap->getdocNum();
          Result *result = new Result[nres];
          memset(result,0,nres*sizeof(Result));
          for(int i = 0; i < nres; i++){
             result[i].docid = i;
             result[i].weight = nix/(xi+1) + 1/(i+1); // docid order
             result[i].list[0] = 0x80; // EOL
          }
          Indexes[xi].oresult = result;
          Indexes[xi].resultLen = nres;
          ixn++;
      }
//dbgout << "<PLAINTEXT>" << endl;
//dbgout << "test2\n";
  }
  if((result = mergeResult(Indexes,nix,ixn,resultLen)) == NULL){
      if( resultLen <= 0 ){
          cgi_error("Not found", "nothing found. ", uquery, 0);
      }else
      cgi_error("Error", "cannot open description file.", uquery);
      return 3;
  }
  XResult **resultp = SXsortXResults(result, resultLen);
  storeCache(Indexes,nix,uquery,resultp,resultLen);
  findDocid(resultp,resultLen);

    //
    // display results
    //
    if(resultLen <= 0){
        resultLen = 0;
        if(0 == strcmp(param_pfx,"ja"))
        cgi_error("ĤޤǤ", "nothing found. ", uquery, 0);
        else
        cgi_error("Not found", "nothing found. ", uquery, 0);
        writelog(uquery, 0);
    }else{
        if(param_from <= 0 || param_from > resultLen){
            cgi_error("Error", "parameter 'from' out of range.", uquery);
            return 4;
        }
        int rto = param_from + param_n - 1;
        if(rto > resultLen) rto = resultLen;
        if(rto == 0)
            rto = resultLen;
        else
            rto = (resultLen > rto) ? rto : resultLen;

        char query[LINE_MAX];
        char title[LINE_MAX];
        Exp2string(cexp,query,sizeof(query));
        resultDisp.Nexts = setup_resultprint(title,query,uquery,resultLen,rto);

        display_html("head.html", title, query, "(no error)", resultLen);

        print_results(Indexes, resultp, uquery,
                      resultLen, param_n, param_from -1, rto - 1);

        display_html("foot.html", title, query, "(no error)", resultLen);
        sout_flush();
        writelog(uquery, resultLen);
    }
    if(result) delete []result;
    
    return 0;
}

#ifdef _MSC_VER
#include <iostream>
#include <strstream>
/*
#include <strstrea.h>
*/
void Exp2string(Expression &exp, char *query, int size){
	ostrstream squery;
	exp.print(squery);
	squery << '\0';
	query[0] = 0;
	Strncpy(query,squery.str(),size);
}
#else
#include <sstream>
#include <string>
using namespace std;
void Exp2string(Expression &exp, char *query, int size){
	stringstream squery;
	exp.print(squery);
	squery << '\0';
	query[0] = 0;
	Strncpy(query,squery.str().c_str(),size);
}
#endif
