// Copyright (c) 2004 Yutaka Sato. All rights reserved.
// created 01 Sep. 2004

#include "SX.h"

#include "Retriever.h"
#include "Unifier.h"
#include "Expression.h"
#include "Evaluator.h"
#include "Result.h"
#include "DescFile.h"
#include "ConText.h"

int url2did(const char *idx, const char *url){
	char idxbase[PATH_MAX];
	char numpath[PATH_MAX];
	SX_searchBase((char*)idx,idxbase,sizeof(idxbase));
	sprintf(numpath,"%s%s",idxbase,".num");
	NumFile nums(numpath);
	return findUrl(nums,url);
}
/*
#define o_idx	val['i']
#define o_url	val['u']
*/
#define o_idx	val[(int)'i']
#define o_key	val[(int)'k']
#define o_url	val[(int)'u']

void printusage(int ac, char *av[]){
printf("%s\n",FREYABANNER);
printf("Usage: %s command options\n",av[0]);
printf("\
command:\n\
   dump -i idx         dump documents in idx\n\
   find -i idx -u url  find document id of the url in idx\n\
   dumplex -i idx      dump lexicons\n\
   dumpctx -i idx      dump complex text\n\
   dumpxmap -i idx     dump XMap (document base array)\n\
options:\n\
   -i index\n\
   -u url\n\
");
}

int mainx(int ac, char *av[]){
	char *com,*a1,*opt;
	char *val[128];
	char *rav[256];
	int rac = 0;
	char path[PATH_MAX];

	for(int i = 0; i < 128; i++)
		val[i] = 0;

	if(ac < 2){
		printusage(ac,av);
		return 0;
	}
	com = av[1];

	for(int ai = 2; ai < ac; ai++){
		a1 = av[ai];
		if( ai+1 < ac )
			opt = av[ai+1];
		else	opt = 0;
		if( av[ai][0] == '-' ){
			int och = av[ai][1];
			if( isalpha(och) ){
				if( opt ){
					val[och] = opt;
					ai++;
				}
			}
		}else{
			rav[rac++] = a1;
		}
	}

	char idxbase[PATH_MAX];
	idxbase[0] = 0;
	if( o_idx ){
		SX_searchBase((char*)o_idx,idxbase,sizeof(idxbase));
	}

	if( strcmp(com,"info") == 0 ){
		printf("Base: %s\n",idxbase);
		sprintf(path,"%s.lex",idxbase);
		LexiconFile lex(path,"r");
		printf("Total Lexicon: %X %d\n",lex.nentry(),lex.nentry());

		sprintf(path,"%s.idx",idxbase);
		IndexFile idx(path,"r");

		sprintf(path,"%s.dsc",idxbase);
		Fstream descf(path,"r");
		DescFile dsc(descf);
		printf("Described Documents: %d\n",dsc.getsize());

		int fd;
		sprintf(path,"%s.map",idxbase);
		fd = open(path,0);
		if( 0 <= fd ){
			XMap map(fd);
			int dn = map.getdocNum();
			printf("Mapped Documents: %d (%X + %d)\n",dn-1,
				map.getdocbase(dn-2),map.getdocsize(dn-2));
			printf("Mapped Documents: %d (%X + %d)\n",dn,
				map.getdocbase(dn-1),map.getdocsize(dn-1));
		}
	}else
	if( strcmp(com,"lexid") == 0 ){
		char path[1024];
		sprintf(path,"%s.lex",idxbase);
		LexiconFile lex(path,"r");
		if( o_key ){
			printf("%d %s\n",lex.lexid(o_key),o_key);
		}
	}else
	if( strcmp(com,"dumplex") == 0 ){
		char path[1024];
		sprintf(path,"%s.lex",idxbase);
		LexiconFile lex(path,"r");
		lex.dump();
	}else
	if( strcmp(com,"dumpctx") == 0 ){
		char path[1024];
		sprintf(path,"%s.ctx",idxbase);
		ConText ctx(path,"r");
		sprintf(path,"%s.lex",idxbase);
		LexiconFile lex(path,"r");
		ctx.lexf = &lex;
		ctx.dump();
	}else
	if( strcmp(com,"dumpxmap") == 0 ){
		int fd;
		sprintf(path,"%s.map",idxbase);
		fd = open(path,0);
		printf("XMap: %d %s\n",fd,path);
		if( 0 <= fd ){
			int i;
			XMap map(fd);
			printf("documents: %d\n",map.getdocNum());
			for( i = 0; i < map.getdocNum(); i++ ){
				printf("[%4d] %8X %7X\n",i,
					map.getdocbase(i),map.getdocsize(i));
			}
		}
	}else
	if( strcmp(com,"find") == 0 ){
		if( o_url != 0 && o_idx != 0 ){
			int did = url2did(o_idx,o_url);
			printf("%d %s\n",did,o_url);
		}
	}
	else
	if( strcmp(com,"dump") == 0 ){
		if( o_idx ){
			char path[PATH_MAX];
			sprintf(path,"%s%s",idxbase,".dsc");
			Fstream descf(path,"r");
			DescFile descF(descf);
			int size = descF.getsize();
			for(int did = 0; did < size ; did++){
				DescRecord rec;
				descF.getRecord(rec,did);
				printf("%d %s\n",did,rec.getUrl());
			}
		}
	}
	else
	if( strcmp(com,"vote") == 0 ){
		if( o_url != 0 && o_idx != 0 ){
		int did = url2did(o_idx,o_url);
		if( 0 <= did ){
			int inc = 0;
			CounterElem *ce = SX_incCount(idxbase,"vote",did,inc,-1);
			if(ce != NULL){
				printf("%d %d %d %s\n",
					ce->c_count,ce->c_acount,did,o_url);
			}
		}
		}
	}
	return 0;
}

extern char **environ;
extern char ContentType[64];
extern double Start;

KeyVals *scan_query(const char *query);
int scangen_html(const char *base, KeyVals *kv);

void sout_flush();
int
main(int argc, char *argv[]){
	double now = Time();
	int rands = (int)(0x1000*(now-(int)now)) % 0x1000;
#ifdef _MSC_VER
	char *randstack = (char*)alloca(4*rands);
#else
	int randstack[rands];
#endif
	randstack[0] = 0;

	if( strstr(argv[0],".cgi") ){
		Start = now;
		KeyVals *kv = 0;
		char *query;
		if( (query = getenv("QUERY_STRING")) ){
			kv = scan_query(query);
		}else	kv = scan_query("");
		strcpy(ContentType,"text/html");
		scangen_html("fsx",kv);
		sout_flush();

		if(0)
		{
			printf("\n<HR>\n");
			printf("<PRE>\n");
			int i;
			for(i = 0; i < kv->len; i++){
				printf("param[%d] %s = %s\n",i,
					kv->keyval[i].key,kv->keyval[i].val);
			}
			for( i = 0; i < argc; i++ )
				printf("arg[%2d] %s\n",i,argv[i]);
			for( i = 0; environ[i]; i++ )
				printf("env[%2d] %s\n",i,environ[i]);
			printf("</PRE>\n");
			printf("<HR>\n");
		}
	}else{
		mainx(argc,argv);
	}
	return 0;
}

int param_from;
int param_n;
char resultFroms[16];
char resultTos[16];
void setrv(const char *nam, const char *val){
}
char *getrv(const char *nam){
	return "";
}
char *getrvx(const char *nam){
	return 0;
}
