// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// minimum kanji convert library
//  $Id: kcnvlib.cc,v 0.9 1997-12-27 07:49:58+09 harada Exp $
//
#include "SX.h"
#include "kcnvlib.h"

static char binarytbl[32] = {
    1, 1, 1, 1, 1, 1, 1, 0,  0, 0, 0, 1, 0, 0, 0, 0,
    1, 1, 1, 1, 1, 1, 1, 1,  1, 1, 0, 0, 1, 1, 1, 1,
};

int Isbinary(char c){
    if((unsigned char)c >= sizeof(binarytbl)/sizeof(binarytbl[0]))
        return 0;
    return binarytbl[(int)c];
}

//
// detect / convert
//
enum kcode
detectcode(const char *str){
    const char *cp;
    int euc_or_sjis = 0;
   
    cp = str;
    while(*cp){
        if(*cp < 32){
            if(Isbinary(*cp))
                return CODE_BINARY;
            else if(*cp == ESC){
                if(*(cp + 1) == '\0') return CODE_BINARY;
                cp++;
                if((*cp == '$')
                   && ((*(cp + 1) == '@')||(*(cp + 1) == 'B'))){
                    /* cp += 2; */
                    return CODE_JIS;
                }
                if(*cp == '('){
                    if((*(cp + 1) == 'B')||(*(cp + 1) == 'J')){
                        /* cp += 2; */
                        return CODE_JIS;
                    }
                    if(*(cp + 1) == 'I'){   /* JIS-KANA */
                        /* cp += 2; */
                        return CODE_JIS;
                    }
                }
            }else if(*cp == SI || *cp == SO){
                return CODE_JIS; /* JIS-KANA */
            }
        }
        cp++;
    }
    
    cp = str;
    while(*cp){
        if(Is_euc_here(cp)){   /* EUC? */
            if(Is_sjis_here(cp)){   /* EUC and SJIS? */
                euc_or_sjis++;
                cp += 2;
                continue;
            }
            return CODE_EUC;
        }else if(Is_sjis_here(cp)){ /* SJIS? */
            return CODE_SJIS;
        }else if(! (isalnum(*(unsigned char *)cp)
                    || isgraph(*(unsigned char *)cp)
                    || isspace(*(unsigned char *)cp))){
            return CODE_BINARY;
        }
        cp++;
    }
    
    if(euc_or_sjis) return CODE_EUC; /* (^_^; */
    return CODE_ASCII;
}


int
jis2euc(char *result, const char *jisstr){
    const char *cp;
    char *dp;
    int ctype = 0;  /* 0 ... ASCII,  1 ... KANJI,  2 ... KANA */

    cp = jisstr;
    dp = result;
    while(*cp){
        /* determine state (ctype) */
        if(*cp == ESC){
            if(*(cp + 1) == '\0') return -1;
            cp++;
            if((*cp == '$')
               && ((*(cp + 1) == '@')||(*(cp + 1) == 'B'))){
                cp += 2;
                ctype = 1;
                continue;
            }
            if(*cp == '('){
                if((*(cp + 1) == 'B')||(*(cp + 1) == 'J')){
                    cp += 2;
                    ctype = 0;
                    continue;
                }
                if(*(cp + 1) == 'I'){   /* JIS-KANA */
                    cp += 2;
                    ctype = 2;
                    continue;
                }
            }
        }else if(*cp == SO){
            cp++;
            ctype = 2;
            continue;
        }else if(*cp == SI){
            cp++;
            ctype = 0;
            continue;
        }

        /* do convert */
        if(ctype == 0){
            *dp++ = *cp++;
        }else if(ctype == 1){
            *dp++ = *cp++ | 0x80;
        }else {
            *dp++ = 0x8e;
            *dp++ = *cp++ | 0x80;
        }
    }
    *dp = '\0';

    return (int)(dp - result);
}


int
euc2jis(char *result, const char *eucstr){
    const char *cp;
    char *dp;
    int ctype = 0;

    cp = eucstr;
    dp = result;
    while(*cp){
        if(Is_kanji_here(cp)){
            if(1 != ctype){
                *dp++ = ESC; *dp++ = '$'; *dp++ = 'B';
                ctype = 1;
            }
            *dp++ = 0x7f & *cp++;
            *dp++ = 0x7f & *cp++;
            continue;
        }else if(Ishkana1(*cp)){
            if(2 != ctype){
                *dp++ = ESC; *dp++ = '('; *dp++ = 'I';
                ctype = 2;
            }
            cp++;
            *dp++ = 0x7f & *cp++;
        }else {
            if(0 != ctype){
                *dp++ = ESC; *dp++ = '('; *dp++ = 'B';
                ctype = 0;
            }
            *dp++ = *cp++;
        }
    }

    /* clean up */
    if(0 != ctype){
        *dp++ = ESC; *dp++ = '('; *dp++ = 'B';
    }
    *dp = '\0';

    return (int)(dp - result);
}


int sjis2euc(char *result, const char *sjisstr){
    const char *cp;
    char *dp;

    cp = sjisstr;
    dp = result;
    while(*cp){
        if(Is_sjis_kanji_here(cp)){
            /* shift & copy as JIS */
            *dp = (*(unsigned char *)cp <= 0x9f) ? (*cp - 0x70) : (*cp - 0xb0);
            *dp <<= 1;
            cp++;
            if(*(unsigned char *)cp < 0x9f){
                *(dp + 1) = (*(unsigned char *)cp <= 0x7e)
                    ? (*cp - 0x1f) : (*cp - 0x20);
                --*dp;
            }else{
                *(dp + 1) = *cp - 0x7e;
            }
            cp++;

            /* JIS to EUC */
            *dp++ |= 0x80;
            *dp++ |= 0x80;
            
            continue;
        }else if(Is_sjis_hkana(*cp)){
            *dp++ = 0x8e;
            *dp++ = *cp;
            cp++; continue;
        }
        *dp++ = *cp++;  /* ASCII */
    }
    *dp = '\0';

    return (int)(dp - result);
}

char
*conv2euc(char *result, const char *str, enum kcode hint){
    while(1){
        switch (hint) {
        case CODE_UNKNOWN:
            hint = detectcode(str);
            if(0 == hint) return NULL;
            continue;
        case CODE_ASCII:
        case CODE_EUC:
            return strcpy(result, str);
        case CODE_JIS:
            if(-1 == jis2euc(result, str)) return NULL;
            return result;
        case CODE_SJIS:
            if(-1 == sjis2euc(result, str)) return NULL;
            return result;
        default:
            return NULL;
        }
    }
}

int
is_valid_kanji(const char *kp){
    unsigned char valid[] = "";
    for(const unsigned char *vp = valid; *vp; vp += 2){
        if(       *vp == *(unsigned char *)kp
           && *(vp+1) == *(unsigned char *)(kp+1))
            return 1;
    }
    return Is_jis_level1(kp) || Is_jis_level2(kp)
        || Is_cyrillic(kp) || Is_greek(kp)
        || Is_hiragana_here(kp) || Is_katakana_here(kp);
}
