// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// minimum kanji convert library
//  $Id: kcnvlib.h,v 0.9 1997-12-27 07:50:15+09 harada Exp $
//
#ifndef KCNVLIB_H_INCLUDED
#define KCNVLIB_H_INCLUDED

enum kcode {CODE_UNKNOWN=0, CODE_BINARY=0,
            CODE_ASCII=1,
            CODE_JIS=2, CODE_JIS83=2, CODE_JUNET=2,
            CODE_EUC=3,
            CODE_MSKANJI=4, CODE_SJIS=4
            };
enum kcntl {SO = 0x0e, SI = 0x0f, ESC = 0x1b };

inline int Iskanji1(char c);
inline int Iskanji2(char c);
inline int Is_kanji_here(const char *p);
inline int Is_jis_level1(const char *p);
inline int Is_jis_level2(const char *p);
inline int Is_cyrillic(const char *p);
inline int Is_greek(const char *p);
inline int Is_hiragana_here(const char *p);
inline int Is_katakana_here(const char *p);
inline int Ishkana1(char c);
inline int Ishkana2(char c);
inline int Is_euc_here(const char *p);
inline int Is_dakuten_here(const char *p);
inline int Is_handakuten_here(const char *p);
inline int Is_kana_here(const char *kstr);

inline int Is_sjis_kanji1(char c);
inline int Is_sjis_kanji2(char c);
inline int Is_sjis_kanji_here(const char *p);
inline int Is_sjis_hkana(char c);
inline int Is_sjis_here(const char *p);

int Isbinary(char c);
enum kcode detectcode(const char *str);
char *conv2euc(char *result, const char *str, enum kcode hint);
int jis2euc(char *result, const char *jisstr);
int euc2jis(char *result, const char *eucstr);
int sjis2euc(char *result, const char *sjisstr);


// implementations of inline functions

//
// EUC
//
inline int Iskanji1(char c){
    return ((0xa1 <= (unsigned char)(c))&&((unsigned char)(c) <= 0xfe));
}

inline int Iskanji2(char c){
    return ((0xa1 <= (unsigned char)(c))&&((unsigned char)(c) <= 0xfe));
}

inline int Is_kanji_here(const char *p){
    return (Iskanji1(*p) && Iskanji2(*(p + 1)));
}

inline int Is_jis_level1(const char *p){
    return (    (*(unsigned char *)p >= 0xB0)
             && (*(unsigned char *)p <= 0xCF)
             && Iskanji2(*(p + 1)) );
}

inline int Is_jis_level2(const char *p){
    return (    (*(unsigned char *)p >= 0xD0)
             && (*(unsigned char *)p <= 0xF5)
             && Iskanji2(*(p + 1)) );
}

inline int Is_cyrillic(const char *p){
    return (*(unsigned char *)p == 0xA6) && Iskanji2(*p);
}

inline int Is_greek(const char *p){
    return (*(unsigned char *)p == 0xA7) && Iskanji2(*p);
}

inline int Is_hiragana_here(const char *p){
    return (*(unsigned char *)p == 0xa4) && Iskanji2(*p);
}

inline int Is_katakana_here(const char *p){
    return (*(unsigned char *)p == 0xa5) && Iskanji2(*p);
}

inline int Ishkana1(char c){
    return (0x8e == (unsigned char)(c));
}

inline int Ishkana2(char c){
    return ((0xa0 <= (unsigned char)(c))&&((unsigned char)(c) <= 0xdf));
}

inline int Is_euc_here(const char *p){
    return ( Iskanji1(*p) && Iskanji2(*(p + 1)) )
        || ( Ishkana1(*p) && Ishkana2(*(p + 1)) );
}

inline int Is_dakuten_here(const char *cp){
    return ((*(unsigned char *)cp == 0x8e
          && *(unsigned char *)(cp + 1) == 0xde)
         || (*(unsigned char *)cp == 0xa1
          && *(unsigned char *)(cp + 1) == 0xab));
}

inline int Is_handakuten_here(const char *cp){
    return ((*(unsigned char *)cp == 0x8e
          && *(unsigned char *)(cp + 1) == 0xdf)
         || (*(unsigned char *)cp == 0xa1
          && *(unsigned char *)(cp + 1) == 0xac));
}

inline int Is_kana_here(const char *kstr){
    unsigned int w;
    w = (*kstr << 8) + *(kstr + 1);

    return (0xA4A0 <= w) && ( w < 0xA5FF );   /* EUC Ҥ餬ʥ */
}

//
// SJIS
//
inline int Is_sjis_kanji1(char c){
    return ((0x81 <= (unsigned char)(c))&&((unsigned char)(c) <= 0x9f))
        || ((0xe0 <= (unsigned char)(c))&&((unsigned char)(c) <= 0xfc));
}

inline int Is_sjis_kanji2(char c){
    return ((0x40 <= (unsigned char)(c))&&((unsigned char)(c) <= 0x7e))
        || ((0x80 <= (unsigned char)(c))&&((unsigned char)(c) <= 0xfc));
}

inline int Is_sjis_kanji_here(const char *p){
    return Is_sjis_kanji1(*p) && Is_sjis_kanji2(*(p + 1));
}

inline int Is_sjis_hkana(char c){
    return ((0xa1 <= (unsigned char)(c))&&((unsigned char)(c) <= 0xdf));
}

inline int Is_sjis_here(const char *p){
    return ( Is_sjis_kanji1(*p) && Is_sjis_kanji2(*(p + 1)) )
        || Is_sjis_hkana(*p);
}

#endif
