// Copyright (c) 1997 HARADA Masanori. All rights reserved.
//
// Patricia class
// $Id: patricia.h,v 0.9 1997-12-27 07:51:14+09 harada Exp $
//
#ifndef PATRICIA_H_INCLUDE
#define PATRICIA_H_INCLUDE

class Pat_node {
public:
    char *key;
    long content;
    int checkbit;               // ˥å٤ bit
    Pat_node *left;             // bit 0 ʤ麸
    Pat_node *right;            // bit 1 ʤ鱦
    Pat_node();
    Pat_node(const char *k, long c);
    ~Pat_node();
    
    int dump_idx;
};

class Patricia {
private:
    Pat_node *root_node;
    int prefix_dump_counter;

    int dump_all_counter;

    int keyNum;
    Pat_node *down_node(const char *key, Pat_node *root);
    int make_checkbit_table(int *table, const char *key);
    void delete_subtree(Pat_node *here);
    int prefix_dump(const char *key, Pat_node *matches[], Pat_node *here);

    int dump_all(Pat_node *table[], int table_size);
    int dump_all_recursive(Pat_node *matches[], Pat_node *here);

public:
    enum { MAXKEYLENGTH = 64 };
    enum { EDUPLICATE = -1, INSERTOK = 0 };
/*
    enum { MAGIC = 0x19970924 };
*/
    enum { MAGIC = 0x20040617 };
    Patricia();
/*
    Patricia(istream &file);
*/
    Patricia(Fstream &file);
    ~Patricia();
    operator int() const { return keyNum; }
    operator bool() const { return keyNum > 0; }
    int good(){ return 0 < keyNum; }
    static int find_diffbit(const char *s1, const char *s2);
    int getbit(const char *string, int nthbit, int len);
    inline int getbit(const char *string, int nthbit);
    static const char *is_prefix(const char *s1, const char *s2);
    int size();
    long search(const char *key);
    int insert(const char *key, long content);
    int search_prefix(const char *key, Pat_node *matches[]);
    int prefix_search(const char *key, Pat_node *matches[]);

/*
    int save(ostream &file);
    int load(istream &file);
*/
    int save(Fstream &file);
    int load(Fstream &file);
};

inline int
Patricia::getbit(const char *string, int nthbit){
    return getbit(string, nthbit, strlen(string));
}

#endif
